@@ -49,7 +49,11 @@ namespace System.Windows.Forms {
 
 		private bool show_panels = false;
 		private bool sizing_grip = true;
-
+		
+		// Stuff for panel Tooltips
+		private Timer tooltip_timer;
+		private ToolTip tooltip_window;
+		private StatusBarPanel tooltip_currently_showing;
 		#endregion	// Fields
 
 		#region Public Constructors
@@ -58,6 +62,10 @@ namespace System.Windows.Forms {
 			Dock = DockStyle.Bottom;
 			this.TabStop = false;
 			this.SetStyle(ControlStyles.UserPaint | ControlStyles.Selectable, false);
+
+			// For displaying/hiding tooltips
+			MouseMove += new MouseEventHandler (StatusBar_MouseMove);
+			MouseLeave += new EventHandler (StatusBar_MouseLeave);
 		}
 		#endregion	// Public Constructors
 
@@ -97,7 +105,6 @@ namespace System.Windows.Forms {
 		}
 
 #if NET_2_0
-		[Browsable (false)]
 		[EditorBrowsable (EditorBrowsableState.Never)]
 		protected override bool DoubleBuffered {
 			get {
@@ -221,10 +228,10 @@ namespace System.Windows.Forms {
 			base.Dispose (disposing);
 		}
 
-		protected virtual void OnDrawItem (StatusBarDrawItemEventArgs e) {
+		protected virtual void OnDrawItem (StatusBarDrawItemEventArgs sbdievent) {
 			StatusBarDrawItemEventHandler eh = (StatusBarDrawItemEventHandler)(Events [DrawItemEvent]);
 			if (eh != null)
-				eh (this, e);
+				eh (this, sbdievent);
 		}
 
 		protected override void OnHandleCreated (EventArgs e) {
@@ -236,8 +243,8 @@ namespace System.Windows.Forms {
 			base.OnHandleDestroyed (e);
 		}
 
-		protected override void OnLayout (LayoutEventArgs e) {
-			base.OnLayout (e);
+		protected override void OnLayout (LayoutEventArgs levent) {
+			base.OnLayout (levent);
 		}
 
 		protected override void OnMouseDown (MouseEventArgs e) {
@@ -343,7 +350,10 @@ namespace System.Windows.Forms {
 					continue;
 				}
 				if (p.AutoSize == StatusBarPanelAutoSize.Contents) {
-					int len = (int) (DeviceContext.MeasureString (p.Text, Font).Width + 0.5F);
+					int len = (int)(TextRenderer.MeasureString (p.Text, Font).Width + 0.5F);
+					if (p.Icon != null) {
+						len += 21;
+					}
 					p.SetWidth (len + 8);
 					taken += p.Width;
 					taken += gap;
@@ -358,15 +368,14 @@ namespace System.Windows.Forms {
 				}
 			}
 
-			if (springs == null)
-				return;
-
-			int spring_total = springs.Count;
-			int total_width = Width - taken - (SizingGrip ? ThemeEngine.Current.StatusBarSizeGripWidth : 0);
-			for (int i = 0; i < spring_total; i++) {
-				StatusBarPanel p = (StatusBarPanel) springs [i];
-				int width = total_width / spring_total;
-				p.SetWidth (width >= p.MinWidth ? width : p.MinWidth);
+			if (springs != null) {
+				int spring_total = springs.Count;
+				int total_width = Width - taken - (SizingGrip ? ThemeEngine.Current.StatusBarSizeGripWidth : 0);
+				for (int i = 0; i < spring_total; i++) {
+					StatusBarPanel p = (StatusBarPanel)springs[i];
+					int width = total_width / spring_total;
+					p.SetWidth(width >= p.MinWidth ? width : p.MinWidth);
+				}
 			}
 
 			taken = border;
@@ -384,6 +393,81 @@ namespace System.Windows.Forms {
 		}
 		#endregion	// Internal Methods
 
+		#region Stuff for ToolTips
+		private void StatusBar_MouseMove (object sender, MouseEventArgs e)
+		{
+			if (!show_panels)
+				return;
+				
+			StatusBarPanel p = GetPanelAtPoint (e.Location);
+			
+			if (p != tooltip_currently_showing)
+				MouseLeftPanel (tooltip_currently_showing);
+				
+			if (p != null && tooltip_currently_showing == null)
+				MouseEnteredPanel (p);
+		}
+
+		private void StatusBar_MouseLeave (object sender, EventArgs e)
+		{
+			if (tooltip_currently_showing != null)
+				MouseLeftPanel (tooltip_currently_showing);
+		}
+
+		private StatusBarPanel GetPanelAtPoint (Point point)
+		{
+			foreach (StatusBarPanel p in Panels)
+				if (point.X >= p.X && point.X <= (p.X + p.Width))
+					return p;
+					
+			return null;
+		}
+		
+		private void MouseEnteredPanel (StatusBarPanel item)
+		{
+			tooltip_currently_showing = item;
+			ToolTipTimer.Start ();
+		}
+
+		private void MouseLeftPanel (StatusBarPanel item)
+		{
+			ToolTipTimer.Stop ();
+			ToolTipWindow.Hide (this);
+			tooltip_currently_showing = null;
+		}
+
+		private Timer ToolTipTimer {
+			get {
+				if (tooltip_timer == null) {
+					tooltip_timer = new Timer ();
+					tooltip_timer.Enabled = false;
+					tooltip_timer.Interval = 500;
+					tooltip_timer.Tick += new EventHandler (ToolTipTimer_Tick);
+				}
+
+				return tooltip_timer;
+			}
+		}
+
+		private ToolTip ToolTipWindow {
+			get {
+				if (tooltip_window == null)
+					tooltip_window = new ToolTip ();
+
+				return tooltip_window;
+			}
+		}
+
+		private void ToolTipTimer_Tick (object o, EventArgs args)
+		{
+			string tooltip = tooltip_currently_showing.ToolTipText;
+
+			if (tooltip != null && tooltip.Length > 0)
+				ToolTipWindow.Present (this, tooltip);
+
+			ToolTipTimer.Stop ();
+		}
+		#endregion
 
 		#region Events
 		[Browsable(false)]
@@ -453,12 +537,31 @@ namespace System.Windows.Forms {
 		public class StatusBarPanelCollection :	 IList, ICollection, IEnumerable {
 			#region Fields
 			private StatusBar owner;
-			private ArrayList panels;
+			private ArrayList panels = new ArrayList ();
 #if NET_2_0
 			private int last_index_by_key;
 #endif
 			#endregion	// Fields
 
+			#region UIA Framework Events
+#if NET_2_0
+			static object UIACollectionChangedEvent = new object ();
+
+			internal event CollectionChangeEventHandler UIACollectionChanged {
+				add { owner.Events.AddHandler (UIACollectionChangedEvent, value); }
+				remove { owner.Events.RemoveHandler (UIACollectionChangedEvent, value); }
+			}
+			
+			internal void OnUIACollectionChanged (CollectionChangeEventArgs e)
+			{
+				CollectionChangeEventHandler eh
+					= (CollectionChangeEventHandler) owner.Events [UIACollectionChangedEvent];
+				if (eh != null)
+					eh (owner, e);
+			}
+#endif
+			#endregion
+
 			#region Public Constructors
 			public StatusBarPanelCollection (StatusBar owner)
 			{
@@ -471,17 +574,20 @@ namespace System.Windows.Forms {
 			private int AddInternal (StatusBarPanel p, bool refresh) {
 				if (p == null)
 					throw new ArgumentNullException ("value");
-				if (panels == null)
-					panels = new ArrayList ();
 
-				int res = panels.Add (p);
 				p.SetParent (owner);
+				int res = panels.Add (p);
 
 				if (refresh) {
 					owner.CalcPanelSizes ();
 					owner.Refresh ();
 				}
 
+#if NET_2_0
+				// UIA Framework Event: Panel Added
+				OnUIACollectionChanged (new CollectionChangeEventArgs (CollectionChangeAction.Add, res));
+#endif
+
 				return res;
 			}
 
@@ -491,11 +597,7 @@ namespace System.Windows.Forms {
 			[Browsable(false)]
 			[EditorBrowsable(EditorBrowsableState.Never)]
 			public int Count {
-				get {
-					if (panels == null)
-						return 0;
-					return panels.Count;
-				}
+				get { return panels.Count; }
 			}
 
 			public bool IsReadOnly {
@@ -513,7 +615,20 @@ namespace System.Windows.Forms {
 						throw new ArgumentNullException ("index");
 					if (index < 0 || index >= Count)
 						throw new ArgumentOutOfRangeException ("index");
+
+#if NET_2_0
+					// UIA Framework Event: Panel Removed
+					OnUIACollectionChanged (new CollectionChangeEventArgs (CollectionChangeAction.Remove, index));
+#endif
+
+					value.SetParent (owner);
+
 					panels [index] = value;
+
+#if NET_2_0
+					// UIA Framework Event: Panel Added
+					OnUIACollectionChanged (new CollectionChangeEventArgs (CollectionChangeAction.Add, index));
+#endif
 				}
 			}
 			
@@ -533,8 +648,8 @@ namespace System.Windows.Forms {
 			#endregion	// Public Instance Properties
 
 			#region Public Instance Methods
-			public virtual int Add (StatusBarPanel p) {
-				return AddInternal (p, true);
+			public virtual int Add (StatusBarPanel value) {
+				return AddInternal (value, true);
 			}
 
 			public virtual StatusBarPanel Add (string text) {
@@ -544,16 +659,14 @@ namespace System.Windows.Forms {
 				return res;
 			}
 
-			public virtual void AddRange (StatusBarPanel [] range) {
-				if (range == null)
-					throw new ArgumentNullException ("panels");
-				if (range.Length == 0)
-					return;
+			public virtual void AddRange (StatusBarPanel [] panels) {
 				if (panels == null)
-					panels = new ArrayList (range.Length);
+					throw new ArgumentNullException ("panels");
+				if (panels.Length == 0)
+					return;
 
-				for (int i = 0; i < range.Length; i++)
-					AddInternal (range [i], false);
+				for (int i = 0; i < panels.Length; i++)
+					AddInternal (panels [i], false);
 				owner.Refresh ();
 			}
 
@@ -561,6 +674,11 @@ namespace System.Windows.Forms {
 				panels.Clear ();
 
 				owner.Refresh ();
+
+#if NET_2_0
+				// UIA Framework Event: Panel Cleared
+				OnUIACollectionChanged (new CollectionChangeEventArgs (CollectionChangeAction.Refresh, -1));
+#endif
 			}
 
 			public bool Contains (StatusBarPanel panel) {
@@ -615,17 +733,34 @@ namespace System.Windows.Forms {
 				// TODO: InvalidArgumentException for bad AutoSize values
 				// although it seems impossible to set it to a bad value
 				value.SetParent (owner);
-				panels [index] = value;
 
+				panels.Insert(index, value);
 				owner.Refresh ();
+
+#if NET_2_0
+				// UIA Framework Event: Panel Added
+				OnUIACollectionChanged (new CollectionChangeEventArgs (CollectionChangeAction.Add, index));
+#endif
 			}
 
-			public virtual void Remove (StatusBarPanel panel) {
-				panels.Remove (panel);
+			public virtual void Remove (StatusBarPanel value) {
+				int index = IndexOf (value);
+				panels.Remove (value);
+
+#if NET_2_0
+				// UIA Framework Event: Panel Removed
+				if (index >= 0)
+					OnUIACollectionChanged (new CollectionChangeEventArgs (CollectionChangeAction.Remove, index));
+#endif
 			}
 
 			public virtual void RemoveAt (int index) {
 				panels.RemoveAt (index);
+
+#if NET_2_0
+				// UIA Framework Event: Panel Removed
+				OnUIACollectionChanged (new CollectionChangeEventArgs (CollectionChangeAction.Remove, index));
+#endif
 			}
 
 #if NET_2_0
@@ -655,12 +790,20 @@ namespace System.Windows.Forms {
 
 
 			object IList.this [int index] {
-				get { return panels [index]; }
-				set { panels [index] = value; }
+				get { return this[index]; }
+				set { 
+					if (!(value is StatusBarPanel))
+						throw new ArgumentException ("Value must be of type StatusBarPanel.", "value");
+						
+					this[index] = (StatusBarPanel)value;
+				}
 			}
 
 			int IList.Add (object value) {
-				return panels.Add (value);
+				if (!(value is StatusBarPanel))
+					throw new ArgumentException ("Value must be of type StatusBarPanel.", "value");
+				
+				return AddInternal ((StatusBarPanel)value, true);
 			}
 
 			bool IList.Contains (object panel) {
@@ -674,7 +817,10 @@ namespace System.Windows.Forms {
 
 			void IList.Insert (int index, object value)
 			{
-				panels.Insert (index, value);
+				if (!(value is StatusBarPanel))
+					throw new ArgumentException ("Value must be of type StatusBarPanel.", "value");
+
+				Insert (index, (StatusBarPanel)value);
 			}
 
 			bool IList.IsFixedSize {
@@ -683,7 +829,8 @@ namespace System.Windows.Forms {
 
 			void IList.Remove (object value)
 			{
-				panels.Remove (value);
+				StatusBarPanel s = value as StatusBarPanel;
+				Remove (s);
 			}
 			#endregion	// IList & ICollection Interfaces
 		}
