@@ -22,6 +22,9 @@
 // Authors:
 //	Jackson Harper (jackson@ximian.com)
 //
+// NOTE: We have some tests in Test/System.Windows.Forms/DragAndDropTest.cs, which I *highly* recommend
+// to run after any change made here, since those tests are interactive, and thus are not part of
+// the common tests.
 //
 
 
@@ -29,6 +32,7 @@ using System;
 using System.IO;
 using System.Text;
 using System.Drawing;
+using System.Threading;
 using System.Collections;
 using System.Runtime.Serialization;
 using System.Runtime.InteropServices;
@@ -280,6 +284,8 @@ namespace System.Windows.Forms {
 			public IntPtr Action;
 			public IntPtr [] SupportedTypes;
 			public MouseButtons MouseState;
+			public DragDropEffects AllowedEffects;
+			public Point CurMousePos;
 			
 			public IntPtr LastWindow;
 			public IntPtr LastTopLevel;
@@ -314,7 +320,7 @@ namespace System.Windows.Forms {
 		private IntPtr XdndActionMove;
 		private IntPtr XdndActionLink;
 		//private IntPtr XdndActionPrivate;
-		//private IntPtr XdndActionList;
+		private IntPtr XdndActionList;
 		//private IntPtr XdndActionDescription;
 		//private IntPtr XdndActionAsk;
 
@@ -340,12 +346,15 @@ namespace System.Windows.Forms {
 		// check out the TODO below
 		//private IntPtr CurrentCursorHandle;
 
-		private X11Keyboard keyboard;
+		private bool tracking = false;
+		private bool dropped = false;
+		private int motion_poll;
+		//private X11Keyboard keyboard;
 
 		public X11Dnd (IntPtr display, X11Keyboard keyboard)
 		{
 			this.display = display;
-			this.keyboard = keyboard;
+			//this.keyboard = keyboard;
 
 			Init ();
 		}
@@ -384,7 +393,7 @@ namespace System.Windows.Forms {
 			drag_data.MouseState = XplatUIX11.MouseState;
 			drag_data.Data = data;
 			drag_data.SupportedTypes = DetermineSupportedTypes (data);
-
+			drag_data.AllowedEffects = allowed_effects;
 			drag_data.Action = ActionFromEffect (allowed_effects);
 
 			if (CursorNo == null) {
@@ -396,29 +405,156 @@ namespace System.Windows.Forms {
 			}
 
 			drag_data.LastTopLevel = IntPtr.Zero;
-			return DragDropEffects.Copy;
+			control = null;
+
+			System.Windows.Forms.MSG msg = new MSG();
+			object queue_id = XplatUI.StartLoop (Thread.CurrentThread);
+
+			Timer timer = new Timer ();
+			timer.Tick += new EventHandler (DndTickHandler);
+			timer.Interval = 100;
+
+			int suc;
+			drag_data.State = DragState.Dragging;
+
+			suc = XplatUIX11.XSetSelectionOwner (display, XdndSelection,
+					drag_data.Window, IntPtr.Zero);
+
+			if (suc == 0) {
+				Console.Error.WriteLine ("Could not take ownership of XdndSelection aborting drag.");
+				drag_data.Reset ();
+				return DragDropEffects.None;
+			}
+
+			drag_data.State = DragState.Dragging;
+			drag_data.CurMousePos = new Point ();
+			source = toplevel = target = IntPtr.Zero;
+			dropped = false;
+			tracking = true;
+			motion_poll = -1;
+			timer.Start ();
+
+			// Send Enter to the window initializing the dnd operation - which initializes the data
+			SendEnter (drag_data.Window, drag_data.Window, drag_data.SupportedTypes);
+			drag_data.LastTopLevel = toplevel;
+
+			while (tracking && XplatUI.GetMessage (queue_id, ref msg, IntPtr.Zero, 0, 0)) {
+
+				if (msg.message >= Msg.WM_KEYFIRST && msg.message <= Msg.WM_KEYLAST) {
+					HandleKeyMessage (msg);
+				} else {
+					switch (msg.message) {
+					case Msg.WM_LBUTTONUP:
+					case Msg.WM_RBUTTONUP:
+					case Msg.WM_MBUTTONUP:
+						if (msg.message == Msg.WM_LBUTTONDOWN && drag_data.MouseState != MouseButtons.Left)
+							break;;
+						if (msg.message == Msg.WM_RBUTTONDOWN && drag_data.MouseState != MouseButtons.Right)
+							break;
+						if (msg.message == Msg.WM_MBUTTONDOWN && drag_data.MouseState != MouseButtons.Middle)
+							break;
+						
+						HandleButtonUpMsg ();
+
+						// We don't want to dispatch button up neither (Match .Net)
+						// Thus we have to remove capture by ourselves
+						RemoveCapture (msg.hwnd);
+						continue;
+					case Msg.WM_MOUSEMOVE:
+						motion_poll = 0;
+
+						drag_data.CurMousePos.X = Control.LowOrder ((int) msg.lParam.ToInt32 ());
+						drag_data.CurMousePos.Y = Control.HighOrder ((int) msg.lParam.ToInt32 ());
+
+						HandleMouseOver ();
+						// We don't want to dispatch mouse move
+						continue;
+					}
+
+					XplatUI.DispatchMessage (ref msg);
+				}
+			}
+
+			timer.Stop ();
+
+			// If the target is a mwf control, return until DragEnter/DragLeave has been fired,
+			// which means the respective -already sent- dnd ClientMessages have been received and handled.
+			if (control != null)
+				Application.DoEvents ();
+
+			if (!dropped)
+				return DragDropEffects.None;
+			if (drag_event != null)
+				return drag_event.Effect;
+
+			// Fallback.
+			return DragDropEffects.None;
 		}
 
-		public bool HandleButtonRelease (ref XEvent xevent)
+		private void DndTickHandler (object sender, EventArgs e)
 		{
-			if (drag_data == null)
-				return false;
+			// This is to make sure we don't get stuck in a loop if another
+			// app doesn't finish the DND operation
+			if (dropped) {
+				Timer t = (Timer) sender;
+				if (t.Interval == 500)
+					tracking = false;
+				else
+					t.Interval = 500;
+			}
 
-			if (!((drag_data.MouseState == MouseButtons.Left &&
-					      xevent.ButtonEvent.button == 1) ||
-					    (drag_data.MouseState == MouseButtons.Right &&
-							    xevent.ButtonEvent.button == 3)))
-				return false;
 
+			// If motion_poll is -1, there hasn't been motion at all, so don't simulate motion yet.
+			// Otherwise if more than 100 milliseconds have lapsed, we assume the pointer is not
+			// in motion anymore, and we simulate the mouse over operation, like .Net does.
+			if (motion_poll > 1)
+				HandleMouseOver ();
+			else if (motion_poll > -1)
+				motion_poll++;
+		}
+
+		// This routines helps us to have a DndEnter/DndLeave fallback when there wasn't any mouse movement
+		// as .Net does
+		private void DefaultEnterLeave (object user_data)
+		{
+			IntPtr toplevel, window;
+			int x_root, y_root;
+
+			// The window generating the operation could be a different than the one under pointer
+			GetWindowsUnderPointer (out window, out toplevel, out x_root, out y_root);
+			Control source_control = Control.FromHandle (window);
+			if (source_control == null || !source_control.AllowDrop)
+				return;
+
+			// `data' and other members are already available
+			Point pos = Control.MousePosition;
+			DragEventArgs drag_args = new DragEventArgs (data, 0, pos.X, pos.Y, drag_data.AllowedEffects, DragDropEffects.None);
+
+			source_control.DndEnter (drag_args);
+			if ((drag_args.Effect & drag_data.AllowedEffects) != 0)
+				source_control.DndDrop (drag_args);
+			else
+				source_control.DndLeave (EventArgs.Empty);
+		}
+
+		public void HandleButtonUpMsg ()
+		{
 			if (drag_data.State == DragState.Beginning) {
 				//state = State.Accepting;
 			} else if (drag_data.State != DragState.None) {
 
 				if (drag_data.WillAccept) {
 
-					if (QueryContinue (xevent, false, DragAction.Drop))
-						return true;
-					
+					if (QueryContinue (false, DragAction.Drop))
+						return;					
+				} else {
+
+					if (QueryContinue (false, DragAction.Cancel))
+						return;
+
+					// fallback if no movement was detected, as .net does.
+					if (motion_poll == -1)
+						DefaultEnterLeave (drag_data.Data);
 				}
 
 				drag_data.State = DragState.None;
@@ -427,101 +563,87 @@ namespace System.Windows.Forms {
 				// handlers
 			}
 
-			return false;
+			return;
 		}
 
-		public bool HandleMotionNotify (ref XEvent xevent)
+		private void RemoveCapture (IntPtr handle)
 		{
-			if (drag_data == null)
-				return false;
+			Control c = MwfWindow (handle);
+			if (c.InternalCapture)
+				c.InternalCapture = false;
+		}
 
-			if (drag_data.State == DragState.Beginning) {
-				int suc;
+		public bool HandleMouseOver ()
+		{
+			IntPtr toplevel, window;
+			int x_root, y_root;
 
+			GetWindowsUnderPointer (out window, out toplevel, out x_root, out y_root);
+
+			if (window != drag_data.LastWindow && drag_data.State == DragState.Entered) {
 				drag_data.State = DragState.Dragging;
 
-				suc = XplatUIX11.XSetSelectionOwner (display, XdndSelection,
-						drag_data.Window,
-						xevent.ButtonEvent.time);
+				// TODO: Send a Leave if this is an MWF window
 
-				if (suc == 0) {
-					Console.Error.WriteLine ("Could not take ownership of XdndSelection aborting drag.");
-					drag_data.Reset ();
-					return false;
-				}
+				if (toplevel != drag_data.LastTopLevel)
+					SendLeave (drag_data.LastTopLevel, toplevel);
+			}
 
-				drag_data.State = DragState.Dragging;
-			} else if (drag_data.State != DragState.None) {
-				bool dnd_aware = false;
-				IntPtr toplevel = IntPtr.Zero;
-				IntPtr window = XplatUIX11.RootWindowHandle;
+			drag_data.State = DragState.Entered;
+			if (toplevel != drag_data.LastTopLevel) {
+				// Entering a new toplevel window
+				SendEnter (toplevel, drag_data.Window, drag_data.SupportedTypes);
+			} else {
+				// Already in a toplevel window, so send a position
+				SendPosition (toplevel, drag_data.Window,
+						drag_data.Action,
+						x_root,	y_root,
+						IntPtr.Zero);
+			}
 
-				IntPtr root, child;
-				int x_temp, y_temp;
-				int mask_return;
+			drag_data.LastTopLevel = toplevel;
+			drag_data.LastWindow = window;
+			return true;
+		}
 
-				while (XplatUIX11.XQueryPointer (display,
-						       window,
-						       out root, out child,
-						       out x_temp, out y_temp,
-						       out xevent.MotionEvent.x,
-						       out xevent.MotionEvent.y,
-						       out mask_return)) {
+		void GetWindowsUnderPointer (out IntPtr window, out IntPtr toplevel, out int x_root, out int y_root)
+		{
+			toplevel = IntPtr.Zero;
+			window = XplatUIX11.RootWindowHandle;
+
+			IntPtr root, child;
+			bool dnd_aware = false;
+			int x_temp, y_temp;
+			int mask_return;
+			int x = x_root = drag_data.CurMousePos.X;
+			int y = y_root = drag_data.CurMousePos.Y;
+
+			while (XplatUIX11.XQueryPointer (display, window, out root, out child,
+					       out x_temp, out y_temp, out x, out y, out mask_return)) {
 					
-					if (!dnd_aware) {
-						dnd_aware = IsWindowDndAware (window);
-						if (dnd_aware) {
-							toplevel = window;
-							xevent.MotionEvent.x_root = x_temp;
-							xevent.MotionEvent.y_root = y_temp;
-						}
+				if (!dnd_aware) {
+					dnd_aware = IsWindowDndAware (window);
+					if (dnd_aware) {
+						toplevel = window;
+						x_root = x_temp;
+						y_root = y_temp;
 					}
+				}
 
-					if (child == IntPtr.Zero)
-						break;
+				if (child == IntPtr.Zero)
+					break;
 					
-					window = child;
-				}
-
-				if (window != drag_data.LastWindow && drag_data.State == DragState.Entered) {
-					drag_data.State = DragState.Dragging;
-
-					// TODO: Send a Leave if this is an MWF window
-
-					if (toplevel != drag_data.LastTopLevel)
-						SendLeave (drag_data.LastTopLevel, xevent.MotionEvent.window);
-				}
-
-				drag_data.State = DragState.Entered;
-				if (toplevel != drag_data.LastTopLevel) {
-					// Entering a new toplevel window
-					SendEnter (toplevel, drag_data.Window, drag_data.SupportedTypes);
-				} else {
-					// Already in a toplevel window, so send a position
-					SendPosition (toplevel, drag_data.Window,
-							drag_data.Action,
-							xevent.MotionEvent.x_root,
-							xevent.MotionEvent.y_root,
-							xevent.MotionEvent.time);
-				}
-
-				drag_data.LastTopLevel = toplevel;
-				drag_data.LastWindow = window;
-				return true;
+				window = child;
 			}
-			return false;
 		}
 
-		public bool HandleKeyPress (ref XEvent xevent)
+		public void HandleKeyMessage (MSG msg)
 		{
-			if (VirtualKeys.VK_ESCAPE == (VirtualKeys) keyboard.EventToVkey (xevent)) {
-				if (!QueryContinue (xevent, true, DragAction.Cancel))
-					return true;
+			if (VirtualKeys.VK_ESCAPE == (VirtualKeys) msg.wParam.ToInt32()) {
+				QueryContinue (true, DragAction.Cancel);
 			}
-
-			return false;
 		}
-
+		
 		// return true if the event is handled here
 		public bool HandleClientMessage (ref XEvent xevent)
 		{
@@ -536,15 +658,14 @@ namespace System.Windows.Forms {
 				return Accepting_HandleLeaveEvent (ref xevent);
 			if (xevent.ClientMessageEvent.message_type == XdndStatus)
 				return HandleStatusEvent (ref xevent);
+			if (xevent.ClientMessageEvent.message_type == XdndFinished)
+				return HandleFinishedEvent (ref xevent);
 
 			return false;
 		}
 
 		public bool HandleSelectionNotifyEvent (ref XEvent xevent)
 		{
-			if (source != XplatUIX11.XGetSelectionOwner (display, XdndSelection))
-				return false;
-
 			MimeHandler handler = FindHandler ((IntPtr) xevent.SelectionEvent.target);
 			if (handler == null)
 				return false;
@@ -578,29 +699,54 @@ namespace System.Windows.Forms {
 			return true;
 		}
 
-		private bool QueryContinue (XEvent xevent, bool escape, DragAction action)
+		private bool QueryContinue (bool escape, DragAction action)
 		{
 			QueryContinueDragEventArgs qce = new QueryContinueDragEventArgs ((int) XplatUI.State.ModifierKeys,
 					escape, action);
 
 			Control c = MwfWindow (source);
+			
+			if (c == null) {
+				tracking = false;
+				return false;
+			}
+			
 			c.DndContinueDrag (qce);
 
 			switch (qce.Action) {
 			case DragAction.Continue:
 				return true;
 			case DragAction.Drop:
-				SendDrop (drag_data.LastTopLevel, source, xevent.ButtonEvent.time);
-				break;
+				SendDrop (drag_data.LastTopLevel, source, IntPtr.Zero);
+				tracking = false;
+				return true;
 			case DragAction.Cancel:
 				drag_data.Reset ();
 				c.InternalCapture = false;
 				break;
 			}
 
+			SendLeave (drag_data.LastTopLevel, toplevel);
+
+			RestoreDefaultCursor ();
+			tracking = false;
 			return false;
 		}
 
+		private void RestoreDefaultCursor ()
+		{
+			// Releasing the mouse buttons should automatically restore the default cursor,
+			// but canceling the operation using QueryContinue should restore it even if the
+			// mouse buttons are not released yet.
+			XplatUIX11.XChangeActivePointerGrab (display,
+					EventMask.ButtonMotionMask |
+					EventMask.PointerMotionMask |
+					EventMask.ButtonPressMask |
+					EventMask.ButtonReleaseMask,
+					Cursors.Default.Handle, IntPtr.Zero);
+
+		}
+
 		private void GiveFeedback (IntPtr action)
 		{
 			GiveFeedbackEventArgs gfe = new GiveFeedbackEventArgs (EffectFromAction (drag_data.Action), true);
@@ -692,7 +838,11 @@ namespace System.Windows.Forms {
 			pos_y = (int) xevent.ClientMessageEvent.ptr3 & 0xFFFF;
 
 			// Copy is implicitly allowed
-			allowed = EffectFromAction (xevent.ClientMessageEvent.ptr5) | DragDropEffects.Copy;
+			Control source_control = MwfWindow (source);
+			if (source_control == null)
+				allowed = EffectsFromX11Source (source, xevent.ClientMessageEvent.ptr5) | DragDropEffects.Copy;
+			else
+				allowed = drag_data.AllowedEffects;
 
 			IntPtr parent, child, new_child, last_drop_child;
 			parent = XplatUIX11.XRootWindow (display, 0);
@@ -741,16 +891,19 @@ namespace System.Windows.Forms {
 
 			if (converts_pending > 0)
 				return true;
+
 			if (!status_sent) {
 				drag_event = new DragEventArgs (data, 0, pos_x, pos_y,
 					allowed, DragDropEffects.None);
 				control.DndEnter (drag_event);
+				
 				SendStatus (source, drag_event.Effect);
 				status_sent = true;
 			} else {
 				drag_event.x = pos_x;
 				drag_event.y = pos_y;
 				control.DndOver (drag_event);
+
 				SendStatus (source, drag_event.Effect);
 			}
 			
@@ -768,6 +921,7 @@ namespace System.Windows.Forms {
 					allowed, DragDropEffects.None);
 				}
 				control.DndLeave (drag_event);
+				control = null;
 			}
 			ResetTargetData ();
 		}
@@ -777,7 +931,7 @@ namespace System.Windows.Forms {
 			if (control != null && drag_event != null) {
 				drag_event = new DragEventArgs (data,
 						0, pos_x, pos_y,
-					allowed, DragDropEffects.None);
+					allowed, drag_event.Effect);
 				control.DndDrop (drag_event);
 			}
 			SendFinished ();
@@ -796,7 +950,7 @@ namespace System.Windows.Forms {
 		{
 			if (drag_data != null && drag_data.State == DragState.Entered) {
 
-				if (!QueryContinue (xevent, false, DragAction.Continue))
+				if (!QueryContinue (false, DragAction.Continue))
 					return true;
 
 				drag_data.WillAccept = ((int) xevent.ClientMessageEvent.ptr2 & 0x1) != 0;
@@ -806,19 +960,47 @@ namespace System.Windows.Forms {
 			return true;
 		}
 
-		private DragDropEffects EffectFromAction (IntPtr action)
+		private bool HandleFinishedEvent (ref XEvent xevent)
+		{
+			return true;
+		}
+
+		private DragDropEffects EffectsFromX11Source (IntPtr source, IntPtr action_atom)
 		{
 			DragDropEffects allowed = DragDropEffects.None;
+			IntPtr type, count, remaining, data = IntPtr.Zero;
+			int format;
+
+			XplatUIX11.XGetWindowProperty (display, source, XdndActionList,
+					IntPtr.Zero, new IntPtr (32), false, (IntPtr) Atom.AnyPropertyType,
+					out type, out format, out count, out remaining, ref data);
+
+			int intptr_size = Marshal.SizeOf (typeof (IntPtr));
+			for (int i = 0; i < count.ToInt32 (); i++) {
+				IntPtr current_atom = Marshal.ReadIntPtr (data, i * intptr_size);
+				allowed |= EffectFromAction (current_atom);
+			}
+
+			// if source is not providing the action list, use the
+			// default action passed in the x11 dnd position message
+			if (allowed == DragDropEffects.None)
+				allowed = EffectFromAction (action_atom);
 
-			if (action == XdndActionCopy)
-				allowed = DragDropEffects.Copy;
-			else if (action == XdndActionMove)
-				allowed |= DragDropEffects.Move;
-			if (action == XdndActionLink)
-				allowed |= DragDropEffects.Link;
 			return allowed;
 		}
 
+		private DragDropEffects EffectFromAction (IntPtr action)
+		{
+			if (action == XdndActionCopy)
+				return DragDropEffects.Copy;
+			else if (action == XdndActionMove)
+				return DragDropEffects.Move;
+			if (action == XdndActionLink)
+				return DragDropEffects.Link;
+
+			return DragDropEffects.None;
+		}
+
 		private IntPtr ActionFromEffect (DragDropEffects effect)
 		{
 			IntPtr action = IntPtr.Zero;
@@ -838,13 +1020,14 @@ namespace System.Windows.Forms {
 		{
 			bool match = false;
 
-			if (source != XplatUIX11.XGetSelectionOwner (display, XdndSelection)) {
-				return false;
-			}
-
 			Control mwfcontrol = MwfWindow (source);
 
+			/* To take advantage of the mwfcontrol, we have to be sure
+			   that the dnd operation is still happening (since messages are asynchronous) */
 			if (mwfcontrol != null && drag_data != null) {
+				if (!tracking)
+					return false;
+
 				IDataObject dragged = drag_data.Data as IDataObject;
 				if (dragged != null) {
 					data = dragged;
@@ -910,7 +1093,7 @@ namespace System.Windows.Forms {
 			xevent.ClientMessageEvent.message_type = XdndStatus;
 			xevent.ClientMessageEvent.format = 32;
 			xevent.ClientMessageEvent.ptr1 = toplevel;
-			if (effect != DragDropEffects.None)
+			if (effect != DragDropEffects.None && (effect & allowed) != 0)
 				xevent.ClientMessageEvent.ptr2 = (IntPtr) 1;
 
 			xevent.ClientMessageEvent.ptr5 = ActionFromEffect (effect);
@@ -957,6 +1140,7 @@ namespace System.Windows.Forms {
 			xevent.ClientMessageEvent.ptr3 = time;
 			
 			XplatUIX11.XSendEvent (display, handle, false, IntPtr.Zero, ref xevent);
+			dropped = true;
 		}
 
 		private void SendPosition (IntPtr handle, IntPtr from, IntPtr action, int x, int y, IntPtr time)
@@ -1022,7 +1206,7 @@ namespace System.Windows.Forms {
 			XdndActionMove = XplatUIX11.XInternAtom (display, "XdndActionMove", false);
 			XdndActionLink = XplatUIX11.XInternAtom (display, "XdndActionLink", false);
 			//XdndActionPrivate = XplatUIX11.XInternAtom (display, "XdndActionPrivate", false);
-			//XdndActionList = XplatUIX11.XInternAtom (display, "XdndActionList", false);
+			XdndActionList = XplatUIX11.XInternAtom (display, "XdndActionList", false);
 			//XdndActionDescription = XplatUIX11.XInternAtom (display, "XdndActionDescription", false);
 			//XdndActionAsk = XplatUIX11.XInternAtom (display, "XdndActionAsk", false);
 
@@ -1143,6 +1327,10 @@ namespace System.Windows.Forms {
 				return null;
 
 			Control res = Control.FromHandle (hwnd.client_window);
+			
+			if (res == null)
+				res = Control.FromHandle (window);
+				
 			return res;
 		}
 
