@@ -76,6 +76,9 @@ namespace Microsoft.Build.BuildEngine {
 		List<string>			builtTargetKeys;
 		bool				building;
 		BuildSettings			current_settings;
+		Stack<Batch>			batches;
+		ProjectLoadSettings		project_load_settings;
+
 
 		static string extensions_path;
 		static XmlNamespaceManager	manager;
@@ -99,10 +102,14 @@ namespace Microsoft.Build.BuildEngine {
 			fullFileName = String.Empty;
 			timeOfLastDirty = DateTime.Now;
 			current_settings = BuildSettings.None;
+			project_load_settings = ProjectLoadSettings.None;
+
+			encoding = null;
 
 			builtTargetKeys = new List<string> ();
 			initialTargets = new List<string> ();
 			defaultTargets = new string [0];
+			batches = new Stack<Batch> ();
 
 			globalProperties = new BuildPropertyGroup (null, this, null, false);
 			foreach (BuildProperty bp in parentEngine.GlobalProperties)
@@ -418,6 +425,12 @@ namespace Microsoft.Build.BuildEngine {
 
 		public void Load (string projectFileName)
 		{
+			Load (projectFileName, ProjectLoadSettings.None);
+		}
+
+		public void Load (string projectFileName, ProjectLoadSettings settings)
+		{
+			project_load_settings = settings;
 			if (String.IsNullOrEmpty (projectFileName))
 				throw new ArgumentNullException ("projectFileName");
 
@@ -451,12 +464,24 @@ namespace Microsoft.Build.BuildEngine {
 		[MonoTODO ("Not tested")]
 		public void Load (TextReader textReader)
 		{
+			Load (textReader, ProjectLoadSettings.None);
+		}
+
+		public void Load (TextReader textReader, ProjectLoadSettings projectLoadSettings)
+		{
+			project_load_settings = projectLoadSettings;
 			fullFileName = String.Empty;
 			DoLoad (textReader);
 		}
 
 		public void LoadXml (string projectXml)
 		{
+			LoadXml (projectXml, ProjectLoadSettings.None);
+		}
+
+		public void LoadXml (string projectXml, ProjectLoadSettings projectLoadSettings)
+		{
+			project_load_settings = projectLoadSettings;
 			fullFileName = String.Empty;
 			DoLoad (new StringReader (projectXml));
 			MarkProjectAsDirty ();
@@ -970,7 +995,7 @@ namespace Microsoft.Build.BuildEngine {
 			}
 
 			Imports.Add (import);
-			import.Evaluate ();
+			import.Evaluate (project_load_settings == ProjectLoadSettings.IgnoreMissingImports);
 		}
 		
 		void AddItemGroup (XmlElement xmlElement, ImportedProject importedProject)
@@ -1068,7 +1093,39 @@ namespace Microsoft.Build.BuildEngine {
 		Dictionary<string, BuildItemGroup> perBatchItemsByName;
 		Dictionary<string, BuildItemGroup> commonItemsByName;
 
-		internal void SetBatchedItems (Dictionary<string, BuildItemGroup> perBatchItemsByName, Dictionary<string, BuildItemGroup> commonItemsByName)
+		struct Batch {
+			public Dictionary<string, BuildItemGroup> perBatchItemsByName;
+			public Dictionary<string, BuildItemGroup> commonItemsByName;
+
+			public Batch (Dictionary<string, BuildItemGroup> perBatchItemsByName, Dictionary<string, BuildItemGroup> commonItemsByName)
+			{
+				this.perBatchItemsByName = perBatchItemsByName;
+				this.commonItemsByName = commonItemsByName;
+			}
+		}
+
+		Stack<Batch> Batches {
+			get { return batches; }
+		}
+
+		internal void PushBatch (Dictionary<string, BuildItemGroup> perBatchItemsByName, Dictionary<string, BuildItemGroup> commonItemsByName)
+		{
+			batches.Push (new Batch (perBatchItemsByName, commonItemsByName));
+			SetBatchedItems (perBatchItemsByName, commonItemsByName);
+		}
+
+		internal void PopBatch ()
+		{
+			batches.Pop ();
+			if (batches.Count > 0) {
+				Batch b = batches.Peek ();
+				SetBatchedItems (b.perBatchItemsByName, b.commonItemsByName);
+			} else {
+				SetBatchedItems (null, null);
+			}
+		}
+
+		void SetBatchedItems (Dictionary<string, BuildItemGroup> perBatchItemsByName, Dictionary<string, BuildItemGroup> commonItemsByName)
 		{
 			this.perBatchItemsByName = perBatchItemsByName;
 			this.commonItemsByName = commonItemsByName;
@@ -1077,17 +1134,14 @@ namespace Microsoft.Build.BuildEngine {
 		// Honors batching
 		internal bool TryGetEvaluatedItemByNameBatched (string itemName, out BuildItemGroup group)
 		{
-			if (perBatchItemsByName == null && commonItemsByName == null)
-				return EvaluatedItemsByName.TryGetValue (itemName, out group);
+			if (perBatchItemsByName != null && perBatchItemsByName.TryGetValue (itemName, out group))
+				return true;
 
-			if (perBatchItemsByName != null)
-				return perBatchItemsByName.TryGetValue (itemName, out group);
-
-			if (commonItemsByName != null)
-				return commonItemsByName.TryGetValue (itemName, out group);
+			if (commonItemsByName != null && commonItemsByName.TryGetValue (itemName, out group))
+				return true;
 
 			group = null;
-			return false;
+			return EvaluatedItemsByName.TryGetValue (itemName, out group);
 		}
 
 		internal string GetMetadataBatched (string itemName, string metadataName)
@@ -1139,7 +1193,7 @@ namespace Microsoft.Build.BuildEngine {
 			return default (T);
 		}
 
-		void LogWarning (string filename, string message, params object[] messageArgs)
+		internal void LogWarning (string filename, string message, params object[] messageArgs)
 		{
 			BuildWarningEventArgs bwea = new BuildWarningEventArgs (
 				null, null, filename, 0, 0, 0, 0, String.Format (message, messageArgs),
@@ -1147,7 +1201,7 @@ namespace Microsoft.Build.BuildEngine {
 			ParentEngine.EventSource.FireWarningRaised (this, bwea);
 		}
 
-		void LogError (string filename, string message,
+		internal void LogError (string filename, string message,
 				     params object[] messageArgs)
 		{
 			BuildErrorEventArgs beea = new BuildErrorEventArgs (
@@ -1199,7 +1253,6 @@ namespace Microsoft.Build.BuildEngine {
 					throw new InvalidOperationException ("GlobalProperties can not be set to persisted property group.");
 				
 				globalProperties = value;
-				NeedToReevaluate ();
 			}
 		}
 
