@@ -7,6 +7,7 @@
 //   Sebastien Pouliot (sebastien@ximian.com)
 //   Daniel Morgan (danielmorgan@verizon.net)
 //   Gert Driesen (drieseng@users.sourceforge.net)
+//   Veerapuram Varadhan  (vvaradhan@novell.com)
 //
 // Copyright (C) 2002 Tim Coleman
 // Portions (C) 2003 Motus Technologies Inc. (http://www.motus.com)
@@ -41,11 +42,11 @@ using Mono.Security.Protocol.Ntlm;
 
 namespace Mono.Data.Tds.Protocol
 {
-	public sealed class Tds70 : Tds
+	public class Tds70 : Tds
 	{
 		#region Fields
 
-		public readonly static TdsVersion Version = TdsVersion.tds70;
+		//public readonly static TdsVersion Version = TdsVersion.tds70;
 		static readonly decimal SMALLMONEY_MIN = -214748.3648m;
 		static readonly decimal SMALLMONEY_MAX = 214748.3647m;
 
@@ -59,12 +60,24 @@ namespace Mono.Data.Tds.Protocol
 		}
 
 		public Tds70 (string server, int port, int packetSize, int timeout)
-			: base (server, port, packetSize, timeout, Version)
+			: base (server, port, packetSize, timeout, TdsVersion.tds70)
 		{
 		}
 
+		public Tds70 (string server, int port, int packetSize, int timeout, TdsVersion version)
+			: base (server, port, packetSize, timeout, version)
+		{
+		}
+		
 		#endregion // Constructors
 
+		#region Properties
+		
+		protected virtual byte[] ClientVersion {
+			get { return new byte[] {0x00, 0x0, 0x0, 0x70};}
+		}
+		#endregion // Properties
+		
 		#region Methods
 
 		private string BuildExec (string sql)
@@ -229,8 +242,9 @@ namespace Mono.Data.Tds.Protocol
 			Comm.Append (totalPacketSize);
 
 			//Comm.Append (empty, 3, pad);
-			byte[] version = {0x00, 0x0, 0x0, 0x70};
-			Comm.Append (version); // TDS Version 7
+			//byte[] version = {0x00, 0x0, 0x0, 0x71};
+			//Console.WriteLine ("Version: {0}", ClientVersion[3]);
+			Comm.Append (ClientVersion); // TDS Version 7
 			Comm.Append ((int)this.PacketSize); // Set the Block Size
 			Comm.Append (empty, 3, pad);
 			Comm.Append (magic);
@@ -371,14 +385,17 @@ namespace Mono.Data.Tds.Protocol
 
 			// Set "reset-connection" bit for the next message packet
 			Comm.ResetConnection = true;
-
+			base.Reset ();
 			return true;
 		}
 
 		public override void ExecPrepared (string commandText, TdsMetaParameterCollection parameters, int timeout, bool wantResults)
 		{
 			Parameters = parameters;
-			ExecuteQuery (BuildPreparedQuery (commandText), timeout, wantResults);
+			if (Parameters != null && Parameters.Count > 0)
+				ExecRPC (TdsRpcProcId.ExecuteSql, commandText, parameters, timeout, wantResults);
+			else
+				ExecuteQuery (BuildPreparedQuery (commandText), timeout, wantResults);
 		}
 			
 		public override void ExecProc (string commandText, TdsMetaParameterCollection parameters, int timeout, bool wantResults)
@@ -387,16 +404,8 @@ namespace Mono.Data.Tds.Protocol
 			ExecRPC (commandText, parameters, timeout, wantResults);
 		}
 
-		protected override void ExecRPC (string rpcName, TdsMetaParameterCollection parameters, 
-						 int timeout, bool wantResults)
+		private void WriteRpcParameterInfo (TdsMetaParameterCollection parameters)
 		{
-			// clean up
-			InitExec ();
-			Comm.StartPacket (TdsPacketType.RPC);
-
-			Comm.Append ( (short) rpcName.Length);
-			Comm.Append (rpcName);
-			Comm.Append ( (short) 0); //no meta data
 			if (parameters != null) {
 				foreach (TdsMetaParameter param in parameters) {
 					if (param.Direction == TdsParameterDirection.ReturnValue) 
@@ -416,6 +425,66 @@ namespace Mono.Data.Tds.Protocol
 					WriteParameterInfo (param);
 				}
 			}
+		}
+		
+		private void WritePreparedParameterInfo (TdsMetaParameterCollection parameters)
+		{
+			if (parameters == null)
+				return;
+			
+			string param = BuildPreparedParameters ();
+			Comm.Append ((byte) 0x00); // no param meta data name
+			Comm.Append ((byte) 0x00); // no status flags
+			
+			// Type_info - parameter info
+			WriteParameterInfo (new TdsMetaParameter ("prep_params", 
+			                                          param.Length > 4000 ? "ntext" : "nvarchar", 
+			                                          param));
+		}
+		
+		private void ExecRPC (TdsRpcProcId rpcId, string sql, 
+		                      TdsMetaParameterCollection parameters, 
+		                      int timeout, bool wantResults)
+		{
+			// clean up
+			InitExec ();
+			Comm.StartPacket (TdsPacketType.RPC);
+			
+			Comm.Append ((ushort) 0xFFFF);
+			Comm.Append ((ushort) rpcId);
+			Comm.Append ((short) 0x02); // no meta data
+			
+			Comm.Append ((byte) 0x00); // no param meta data name
+			Comm.Append ((byte) 0x00); // no status flags
+			
+			// Write sql as a parameter value - UCS2
+			TdsMetaParameter param = new TdsMetaParameter ("sql", 
+			                                               sql.Length > 4000 ? "ntext":"nvarchar",
+			                                               sql);		
+			WriteParameterInfo (param);
+			
+			// Write Parameter infos - name and type
+			WritePreparedParameterInfo (parameters);
+
+			// Write parameter/value info
+			WriteRpcParameterInfo (parameters);
+			Comm.SendPacket ();
+			CheckForData (timeout);
+			if (!wantResults)
+				SkipToEnd ();
+		}
+		
+		protected override void ExecRPC (string rpcName, TdsMetaParameterCollection parameters, 
+						 int timeout, bool wantResults)
+		{
+			// clean up
+			InitExec ();
+			Comm.StartPacket (TdsPacketType.RPC);
+
+			Comm.Append ( (short) rpcName.Length);
+			Comm.Append (rpcName);
+			Comm.Append ( (short) 0); //no meta data
+			WriteRpcParameterInfo (parameters);
 			Comm.SendPacket ();
 			CheckForData (timeout);
 			if (!wantResults)
@@ -432,18 +501,57 @@ namespace Mono.Data.Tds.Protocol
 			TdsColumnType colType = param.GetMetaType ();
 			param.IsNullable = false;
 
-			Comm.Append ((byte)colType); // type
-
+			bool partLenType = false;
 			int size = param.Size;
-			if (size == 0)
+			if (size < 1) {
+				if (size < 0)
+					partLenType = true;
 				size = param.GetActualSize ();
+			}
 
-			/*
-			  If column type is SqlDbType.NVarChar the size of parameter is multiplied by 2
-			  FIXME: Need to check for other types
+			// Change colType according to the following table
+			/* 
+			 * Original Type	Maxlen		New Type 
+			 * 
+			 * NVarChar		4000 UCS2	NText
+			 * BigVarChar		8000 ASCII	Text
+			 * BigVarBinary		8000 bytes	Image
+			 * 
 			 */
-			if (colType == TdsColumnType.BigNVarChar)
-				size <<= 1;
+			TdsColumnType origColType = colType;
+			if (colType == TdsColumnType.BigNVarChar) {
+				// param.GetActualSize() returns len*2
+				if (size == param.Size)
+					size <<= 1;
+				if ((size >> 1) > 4000)
+					colType = TdsColumnType.NText;
+			} else if (colType == TdsColumnType.BigVarChar) {
+				if (size > 8000)
+					colType = TdsColumnType.Text;	
+			} else if (colType == TdsColumnType.BigVarBinary) {
+				if (size > 8000)
+					colType = TdsColumnType.Image;
+			}
+			// Calculation of TypeInfo field
+			/* 
+			 * orig size value		TypeInfo field
+			 * 
+			 * >= 0 <= Maxlen		origColType + content len
+			 * > Maxlen		NewType as per above table + content len
+			 * -1		origColType + USHORTMAXLEN (0xFFFF) + content len (TDS 9)
+			 * 
+			 */
+			// Write updated colType, iff partLenType == false
+			if (TdsVersion > TdsVersion.tds81 && partLenType) {
+				Comm.Append ((byte)origColType);
+				Comm.Append ((short)-1);
+			} else if (ServerTdsVersion > TdsVersion.tds70 
+			           && origColType == TdsColumnType.Decimal) {
+				Comm.Append ((byte)TdsColumnType.Numeric);
+			} else {
+				Comm.Append ((byte)colType);
+			}
+
 			if (IsLargeType (colType))
 				Comm.Append ((short)size); // Parameter size passed in SqlParameter
 			else if (IsBlobType (colType))
@@ -453,18 +561,56 @@ namespace Mono.Data.Tds.Protocol
 
 			// Precision and Scale are non-zero for only decimal/numeric
 			if ( param.TypeName == "decimal" || param.TypeName == "numeric") {
-				Comm.Append ((param.Precision !=0 ) ? param.Precision : (byte) 28);
+				Comm.Append ((param.Precision !=0 ) ? param.Precision : (byte) 29);
 				Comm.Append (param.Scale);
+				// Convert the decimal value according to Scale
+				if (param.Value != null && param.Value != DBNull.Value &&
+				    ((decimal)param.Value) != Decimal.MaxValue && 
+				    ((decimal)param.Value) != Decimal.MinValue) {
+					decimal expo = new Decimal (System.Math.Pow (10, (double)param.Scale));
+					int pVal = (int)(((decimal)param.Value) * expo);
+					param.Value = (decimal)pVal;				
+				}
 			}
 
-			size = param.GetActualSize ();
+			
+			/* VARADHAN: TDS 8 Debugging */
+			/*
+			if (Collation != null) {
+				Console.WriteLine ("Collation is not null");
+				Console.WriteLine ("Column Type: {0}", colType);
+				Console.WriteLine ("Collation bytes: {0} {1} {2} {3} {4}", Collation[0], Collation[1], Collation[2],
+				                   Collation[3], Collation[4]);
+			} else {
+				Console.WriteLine ("Collation is null");
+			}
+			*/
+			
+			// Tds > 7.0 uses collation
+			if (Collation != null && 
+			    (colType == TdsColumnType.BigChar || colType == TdsColumnType.BigNVarChar ||
+			     colType == TdsColumnType.BigVarChar || colType == TdsColumnType.NChar ||
+			     colType == TdsColumnType.NVarChar || colType == TdsColumnType.Text ||
+			     colType == TdsColumnType.NText))
+				Comm.Append (Collation);
+
+		 	// LAMESPEC: size should be 0xFFFF for any bigvarchar, bignvarchar and bigvarbinary 
+			// types if param value is NULL
+			if ((colType == TdsColumnType.BigVarChar || 
+			     colType == TdsColumnType.BigNVarChar ||
+			     colType == TdsColumnType.BigVarBinary) && 
+			    (param.Value == null || param.Value == DBNull.Value))
+				size = -1;
+			else
+				size = param.GetActualSize ();
+
 			if (IsLargeType (colType))
-				Comm.Append ((short)size);
+				Comm.Append ((short)size); 
 			else if (IsBlobType (colType))
-				Comm.Append (size);
+				Comm.Append (size); 
 			else
 				Comm.Append ((byte)size);
-
+			
 			if (size > 0) {
 				switch (param.TypeName) {
 				case "money" : {
@@ -529,9 +675,14 @@ namespace Mono.Data.Tds.Protocol
 		{
 			Parameters = parameters;
 			string sql = commandText;
-			if (wantResults || (Parameters != null && Parameters.Count > 0))
-				sql = BuildExec (commandText);
-			ExecuteQuery (sql, timeout, wantResults);
+
+			if (Parameters != null && Parameters.Count > 0) {
+				ExecRPC (TdsRpcProcId.ExecuteSql, commandText, parameters, timeout, wantResults);
+			} else {
+				if (wantResults)
+					sql = BuildExec (commandText);
+				ExecuteQuery (sql, timeout, wantResults);
+			}
 		}
 
 		private string FormatParameter (TdsMetaParameter parameter)
@@ -622,9 +773,8 @@ namespace Mono.Data.Tds.Protocol
 			//return ColumnValues [0].ToString ();
 		}
 
-		protected override TdsDataColumnCollection ProcessColumnInfo ()
+		protected override void ProcessColumnInfo ()
 		{
-			TdsDataColumnCollection result = new TdsDataColumnCollection ();
 			int numColumns = Comm.GetTdsShort ();
 			for (int i = 0; i < numColumns; i += 1) {
 				byte[] flagData = new byte[4];
@@ -652,12 +802,13 @@ namespace Mono.Data.Tds.Protocol
 				if (IsBlobType (columnType)) {
 					columnSize = Comm.GetTdsInt ();
 					tableName = Comm.GetString (Comm.GetTdsShort ());
-				} else if (IsFixedSizeColumn (columnType))
+				} else if (IsFixedSizeColumn (columnType)) {
 					columnSize = LookupBufferSize (columnType);
-				else if (IsLargeType ((TdsColumnType) xColumnType))
+				} else if (IsLargeType ((TdsColumnType) xColumnType)) {
 					columnSize = Comm.GetTdsShort ();
-				else
+				} else {
 					columnSize = Comm.GetByte () & 0xff;
+				}
 
 				if (IsWideType ((TdsColumnType) columnType))
 					columnSize /= 2;
@@ -676,7 +827,7 @@ namespace Mono.Data.Tds.Protocol
 				string columnName = Comm.GetString (Comm.GetByte ());
 
 				TdsDataColumn col = new TdsDataColumn ();
-				result.Add (col);
+				Columns.Add (col);
 #if NET_2_0
 				col.ColumnType = columnType;
 				col.ColumnName = columnName;
@@ -688,6 +839,7 @@ namespace Mono.Data.Tds.Protocol
 				col.IsReadOnly = !writable;
 				col.AllowDBNull = nullable;
 				col.BaseTableName = tableName;
+				col.DataTypeName = Enum.GetName (typeof (TdsColumnType), xColumnType);
 #else
 				col ["ColumnType"] = columnType;
 				col ["ColumnName"] = columnName;
@@ -699,9 +851,9 @@ namespace Mono.Data.Tds.Protocol
 				col ["IsReadOnly"] = !writable;
 				col ["AllowDBNull"] = nullable;
 				col ["BaseTableName"] = tableName;
+				col ["DataTypeName"] = Enum.GetName (typeof (TdsColumnType), xColumnType);
 #endif
 			}
-			return result;
 		}
 
 		public override void Unprepare (string statementId)
