@@ -42,7 +42,7 @@ using NUnit.Framework;
 using MonoTests.SystemWeb.Framework;
 using MonoTests.stand_alone.WebHarness;
 
-namespace Tests.System.Web.UI.WebControls
+namespace MonoTests.System.Web.UI.WebControls
 {
 	public sealed class ListViewPoker : ListView
 	{
@@ -1197,12 +1197,7 @@ namespace Tests.System.Web.UI.WebControls
 			
 			EventRecorder events = new EventRecorder();
 			t.UserData = events;
-			t.Run ();
-			
-			Console.WriteLine ("recorded events:");
-			events = t.UserData as EventRecorder;
-			foreach (string e in events)
-				Console.WriteLine ("\t{0}", e);
+			t.Run ();			
 		}
 
 		public static void ListView_Edit_OnInit (Page p)
