@@ -258,13 +258,13 @@ namespace System.Web.Services.Description {
 			EndNamespace ();
 			
 			if (!found) warnings = ServiceDescriptionImportWarnings.NoCodeGenerated;
-			return true;
+			return found;
 		}
 
 		void ImportPortBinding (bool multipleBindings)
 		{
 			if (port != null) {
-				if (multipleBindings) className = port.Name;
+				if (multipleBindings) className = binding.Name;
 				else className = service.Name;
 			}
 			else
@@ -513,13 +513,14 @@ namespace System.Web.Services.Description {
 		
 		void AddIncludingSchema (XmlSchemas list, string ns)
 		{
-			XmlSchema sc = Schemas [ns];
-			if (sc == null || list.Contains (sc)) return;
-			list.Add (sc);
-			foreach (XmlSchemaObject ob in sc.Includes)
-			{
-				XmlSchemaImport import = ob as XmlSchemaImport;
-				if (import != null) AddIncludingSchema (list, import.Namespace);
+			foreach (XmlSchema sc in Schemas) {
+				if (sc.TargetNamespace == ns && !list.Contains (sc)) {
+					list.Add (sc);
+					foreach (XmlSchemaObject ob in sc.Includes) {
+						XmlSchemaImport import = ob as XmlSchemaImport;
+						if (import != null) AddIncludingSchema (list, import.Namespace);
+					}
+				}
 			}
 		}
 		
@@ -577,6 +578,10 @@ namespace System.Web.Services.Description {
 			
 			string argsClassName = classNames.AddUnique (messageName + "CompletedEventArgs", null);
 			CodeTypeDeclaration argsClass = new CodeTypeDeclaration (argsClassName);
+			argsClass.Attributes |= MemberAttributes.Public;
+#if NET_2_0
+			argsClass.IsPartial = true;
+#endif
 			argsClass.BaseTypes.Add (new CodeTypeReference ("System.ComponentModel.AsyncCompletedEventArgs"));
 
 			CodeMemberField resultsField = new CodeMemberField (typeof(object[]), "results");
@@ -616,12 +621,14 @@ namespace System.Web.Services.Description {
 			// Event delegate type
 			
 			CodeTypeDelegate delegateType = new CodeTypeDelegate (messageName + "CompletedEventHandler");
+			delegateType.Attributes |= MemberAttributes.Public;
 			delegateType.Parameters.Add (new CodeParameterDeclarationExpression (typeof(object), "sender"));
 			delegateType.Parameters.Add (new CodeParameterDeclarationExpression (argsClassName, "args"));
 			
 			// Event member
 			
 			CodeMemberEvent codeEvent = new CodeMemberEvent ();
+			codeEvent.Attributes = codeEvent.Attributes & ~MemberAttributes.AccessMask | MemberAttributes.Public;
 			codeEvent.Name = messageName + "Completed";
 			codeEvent.Type = new CodeTypeReference (delegateType.Name);
 			CodeTypeDeclaration.Members.Add (codeEvent);
@@ -790,13 +797,13 @@ namespace System.Web.Services.Description {
 		public void UnsupportedOperationBindingWarning (string text)
 		{
 			warnings |= ServiceDescriptionImportWarnings.UnsupportedOperationsIgnored;
-			AddGlobalComments ("WARNING: Could not generate operation " + OperationBinding.Name + ". " + text);
+			AddGlobalComments ("WARNING: Could not generate operation binding " + OperationBinding.Name + ". " + text);
 		}
 
 		public void UnsupportedOperationWarning (string text)
 		{
 			warnings |= ServiceDescriptionImportWarnings.UnsupportedOperationsIgnored;
-			AddGlobalComments ("WARNING: Could not generate operation " + OperationBinding.Name + ". " + text);
+			AddGlobalComments ("WARNING: Could not generate operation " + Operation.Name + ". " + text);
 		}
 
 		void AddGlobalComments (string comments)
