@@ -46,7 +46,7 @@ namespace System.Web.Compilation
 		const string DEFAULT_NAMESPACE = "ASP";
 
 #if NET_2_0
-		static Guid HashMD5 = new Guid(0x406ea660, 0x64cf, 0x4c82, 0xb6, 0xf0, 0x42, 0xd4, 0x81, 0x72, 0xa7, 0x99);
+		internal static Guid HashMD5 = new Guid(0x406ea660, 0x64cf, 0x4c82, 0xb6, 0xf0, 0x42, 0xd4, 0x81, 0x72, 0xa7, 0x99);
 		static BindingFlags replaceableFlags = BindingFlags.Public | BindingFlags.NonPublic |
 						  BindingFlags.Instance;
 #endif
@@ -67,6 +67,19 @@ namespace System.Web.Compilation
 		protected CodeTypeReferenceExpression mainClassExpr;
 		protected static CodeThisReferenceExpression thisRef = new CodeThisReferenceExpression ();
 
+#if NET_2_0
+		VirtualPath inputVirtualPath;
+		
+		public VirtualPath InputVirtualPath {
+			get {
+				if (inputVirtualPath == null)
+					inputVirtualPath = new VirtualPath (VirtualPathUtility.Combine (parser.BaseVirtualDir, Path.GetFileName (parser.InputFile)));
+
+				return inputVirtualPath;
+			}
+		}
+#endif
+		
 		protected BaseCompiler (TemplateParser parser)
 		{
 			this.parser = parser;
@@ -236,7 +249,7 @@ namespace System.Web.Compilation
 			mainClass.TypeAttributes = TypeAttributes.Public;
 			mainNS.Types.Add (mainClass);
 
-			foreach (object o in parser.Imports) {
+			foreach (object o in parser.Imports.Keys) {
 				if (o is string)
 					mainNS.Imports.Add (new CodeNamespaceImport ((string) o));
 			}
@@ -328,11 +341,7 @@ namespace System.Web.Compilation
 				return;
 			if (!baseType.IsSubclassOf (typeof (System.Web.UI.TemplateControl)))
 				return;
-
-			string arvp = Path.Combine (parser.BaseVirtualDir, Path.GetFileName (parser.InputFile));
-			if (VirtualPathUtility.IsAbsolute (arvp))
-				arvp = VirtualPathUtility.ToAppRelative (arvp);
-
+			
 			CodeTypeReference baseTypeRef = new CodeTypeReference (baseType.FullName);
 			if (parser.BaseTypeIsGlobal)
 				baseTypeRef.Options |= CodeTypeReferenceOptions.GlobalReference;
@@ -341,7 +350,7 @@ namespace System.Web.Compilation
 			CodePropertyReferenceExpression arvpProp = new CodePropertyReferenceExpression (cast, "AppRelativeVirtualPath");
 			CodeAssignStatement arvpAssign = new CodeAssignStatement ();
 			arvpAssign.Left = arvpProp;
-			arvpAssign.Right = new CodePrimitiveExpression (VirtualPathUtility.RemoveTrailingSlash (arvp));
+			arvpAssign.Right = new CodePrimitiveExpression (VirtualPathUtility.RemoveTrailingSlash (InputVirtualPath.AppRelative));
 			ctor.Statements.Add (arvpAssign);
 		}
 #endif
@@ -836,6 +845,20 @@ namespace System.Web.Compilation
 			get { return unit; }
 		}
 
+#if NET_2_0
+		internal CodeTypeDeclaration DerivedType {
+			get { return mainClass; }
+		}
+
+		internal CodeTypeDeclaration BaseType {
+			get {
+				if (partialClass == null)
+					return DerivedType;
+				return partialClass;
+			}
+		}
+#endif
+
 		internal TemplateParser Parser {
 			get { return parser; }
 		}
