@@ -47,6 +47,7 @@ using System.Configuration;
 using System.Collections.Specialized;
 using System.Collections.Generic;
 using System.Web.Configuration;
+using System.Resources;
 #endif
 
 namespace System.Web.Compilation
@@ -67,6 +68,7 @@ namespace System.Web.Compilation
 		
 #if NET_2_0
 		List <string> masterPageContentPlaceHolders;
+		static Regex startsWithBindRegex = new Regex (@"^Bind\s*\(", RegexOptions.Compiled | RegexOptions.IgnoreCase);
 		// When modifying those, make sure to look at the SanitizeBindCall to make sure it
 		// picks up correct groups.
 		static Regex bindRegex = new Regex (@"Bind\s*\(\s*[""']+(.*?)[""']+((\s*,\s*[""']+(.*?)[""']+)?)\s*\)\s*%>", RegexOptions.Compiled | RegexOptions.IgnoreCase);
@@ -522,7 +524,7 @@ namespace System.Web.Compilation
 			
 #if NET_2_0
 			bool need_if = false;
-			if (StrUtils.StartsWith (value, "Bind", true)) {
+			if (startsWithBindRegex.Match (value).Success) {
 				valueExpression = CreateEvalInvokeExpression (bindRegexInValue, value, true);
 				if (valueExpression != null)
 					need_if = true;
@@ -535,7 +537,6 @@ namespace System.Web.Compilation
 				valueExpression = new CodeSnippetExpression (value);
 			
 			method = CreateDBMethod (builder, dbMethodName, GetContainerType (builder), builder.ControlType);
-			
 			CodeVariableReferenceExpression targetExpr = new CodeVariableReferenceExpression ("target");
 
 			// This should be a CodePropertyReferenceExpression for properties... but it works anyway
@@ -658,13 +659,13 @@ namespace System.Web.Compilation
 		/*
 		static bool InvariantCompare (string a, string b)
 		{
-			return (0 == String.Compare (a, b, false, CultureInfo.InvariantCulture));
+			return (0 == String.Compare (a, b, false, Helpers.InvariantCulture));
 		}
 		*/
 
 		static bool InvariantCompareNoCase (string a, string b)
 		{
-			return (0 == String.Compare (a, b, true, CultureInfo.InvariantCulture));
+			return (0 == String.Compare (a, b, true, Helpers.InvariantCulture));
 		}
 
 		static MemberInfo GetFieldOrProperty (Type type, string name)
@@ -851,6 +852,36 @@ namespace System.Web.Compilation
 			
 			return ret;
 		}
+
+		bool ResourceProviderHasObject (string key)
+		{
+			IResourceProvider rp = HttpContext.GetResourceProvider (InputVirtualPath.Absolute, true);
+			if (rp == null)
+				return false;
+
+			IResourceReader rr = rp.ResourceReader;
+			if (rr == null)
+				return false;
+
+			try {
+				IDictionaryEnumerator ide = rr.GetEnumerator ();
+				if (ide == null)
+					return false;
+			
+				string dictKey;
+				while (ide.MoveNext ()) {
+					dictKey = ide.Key as string;
+					if (String.IsNullOrEmpty (dictKey))
+						continue;
+					if (String.Compare (key, dictKey, StringComparison.Ordinal) == 0)
+						return true;
+				}
+			} finally {
+				rr.Close ();
+			}
+			
+			return false;
+		}
 		
 		void AssignPropertyFromResources (ControlBuilder builder, MemberInfo mi, string attvalue)
 		{
@@ -860,9 +891,16 @@ namespace System.Web.Compilation
 			if (!isProperty && !isField || !IsWritablePropertyOrField (mi))
 				return;			
 
+			object[] attrs = mi.GetCustomAttributes (typeof (LocalizableAttribute), true);
+			if (attrs == null && attrs.Length == 0 && !((LocalizableAttribute)attrs [0]).IsLocalizable)
+				return;
+			
 			string memberName = mi.Name;
 			string resname = String.Concat (attvalue, ".", memberName);
 
+			if (!ResourceProviderHasObject (resname))
+				return;
+			
 			// __ctrl.Text = System.Convert.ToString(HttpContext.GetLocalResourceObject("ButtonResource1.Text"));
 			string inputFile = parser.InputFile;
 			string physPath = HttpContext.Current.Request.PhysicalApplicationPath;
@@ -939,7 +977,7 @@ namespace System.Web.Compilation
 			Type type = builder.ControlType;
 			
 			string attvalue = builder.GetAttribute (id);
-			if (id.Length > 2 && id.Substring (0, 2).ToUpper () == "ON"){
+			if (id.Length > 2 && String.Compare (id.Substring (0, 2), "ON", true, Helpers.InvariantCulture) == 0){
 				if (ev_info == null)
 					ev_info = type.GetEvents ();
 
@@ -959,7 +997,7 @@ namespace System.Web.Compilation
 			}
 
 #if NET_2_0
-			if (id.ToLower () == "meta:resourcekey") {
+			if (String.Compare (id, "meta:resourcekey", StringComparison.OrdinalIgnoreCase) == 0) {
 				AssignPropertiesFromResources (builder, attvalue);
 				return;
 			}
@@ -989,7 +1027,7 @@ namespace System.Web.Compilation
 				string value = attvalue.Substring (3, attvalue.Length - 5).Trim ();
 				CodeExpression valueExpression = null;
 #if NET_2_0
-				if (StrUtils.StartsWith (value, "Bind", true))
+				if (startsWithBindRegex.Match (value).Success)
 					valueExpression = CreateEvalInvokeExpression (bindRegexInValue, value, true);
 				else
 #endif
@@ -1061,6 +1099,9 @@ namespace System.Web.Compilation
 			AddEventAssign (method, builder, "DataBinding", typeof (EventHandler), dbMethodName);
 
 			method = CreateDBMethod (builder, dbMethodName, GetContainerType (builder), builder.ControlType);
+#if NET_2_0
+			builder.DataBindingMethod = method;
+#endif
 			CodeCastExpression cast;
 			CodeMethodReferenceExpression methodExpr;
 			CodeMethodInvokeExpression expr;
@@ -1284,6 +1325,7 @@ namespace System.Web.Compilation
 			parent.RenderMethod.Statements.Add (AddLinePragma (expr, cr));
 		}
 
+#if !NET_2_0
 		static PropertyInfo GetContainerProperty (Type type, string[] propNames)
 		{
 			PropertyInfo prop;
@@ -1297,7 +1339,6 @@ namespace System.Web.Compilation
 			return null;
 		}
 
-#if !NET_2_0
 		static string[] containerPropNames = {"Items", "Rows"};
 #endif
 		
@@ -1378,7 +1419,9 @@ namespace System.Web.Compilation
 
 			// Add the DataBind handler
 			method = CreateDBMethod (builder, dbMethodName, GetContainerType (builder), typeof (DataBoundLiteralControl));
-
+#if NET_2_0
+			builder.DataBindingMethod = method;
+#endif
 			CodeVariableReferenceExpression targetExpr = new CodeVariableReferenceExpression ("target");
 			CodeMethodInvokeExpression invoke = new CodeMethodInvokeExpression ();
 			invoke.Method = new CodeMethodReferenceExpression (targetExpr, "SetDataBoundString");
@@ -1393,7 +1436,7 @@ namespace System.Web.Compilation
 			method.Statements.Add (AddLinePragma (invoke, builder));
 			
 			mainClass.Members.Add (method);
-
+			
 			AddChildCall (builder, db);
 		}
 
@@ -1456,11 +1499,8 @@ namespace System.Web.Compilation
 					FlushText (builder, sb);
 					if (b is ObjectTagBuilder) {
 						ProcessObjectTag ((ObjectTagBuilder) b);
-						continue;
-					}
-
-					StringPropertyBuilder pb = b as StringPropertyBuilder;
-					if (pb != null){
+					} else if (b is StringPropertyBuilder) {
+						StringPropertyBuilder pb = b as StringPropertyBuilder;
 						if (pb.Children != null && pb.Children.Count > 0) {
 							StringBuilder asb = new StringBuilder ();
 							foreach (string s in pb.Children)
@@ -1471,11 +1511,9 @@ namespace System.Web.Compilation
 							assign.Right = new CodePrimitiveExpression (asb.ToString ());
 							method.Statements.Add (AddLinePragma (assign, builder));
 						}
-						continue;
 					}
-
 #if NET_2_0
-					if (b is ContentBuilderInternal) {
+					else if (b is ContentBuilderInternal) {
 						ContentBuilderInternal cb = (ContentBuilderInternal) b;
 						CreateControlTree (cb, false, true);
 						AddContentTemplateInvocation (cb, builder.Method, cb.Method.Name);
@@ -1483,33 +1521,28 @@ namespace System.Web.Compilation
 					}
 #endif
 					// Ignore TemplateBuilders - they are processed in InitMethod
-					if (b is TemplateBuilder)
-						continue;
-
-					if (b is CodeRenderBuilder) {
+					else if (b is TemplateBuilder) {
+					} else if (b is CodeRenderBuilder) {
 						AddCodeRender (builder, (CodeRenderBuilder) b);
-						continue;
-					}
-
-					if (b is DataBindingBuilder) {
+					} else if (b is DataBindingBuilder) {
 						AddDataBindingLiteral (builder, (DataBindingBuilder) b);
-						continue;
-					}
-					
-					if (b is ControlBuilder) {
+					} else if (b is ControlBuilder) {
 						ControlBuilder child = (ControlBuilder) b;
 						CreateControlTree (child, inTemplate, builder.ChildrenAsProperties);
 						AddChildCall (builder, child);
 						continue;
-					}
+					} else
+						throw new Exception ("???");
 
-					throw new Exception ("???");
+#if NET_2_0
+					ControlBuilder bldr = b as ControlBuilder;
+					bldr.ProcessGeneratedCode (CompileUnit, BaseType, DerivedType, bldr.Method, bldr.DataBindingMethod);
+#endif
 				}
 
 				FlushText (builder, sb);
 			}
 
-
 			ControlBuilder defaultPropertyBuilder = builder.DefaultPropertyBuilder;
 			if (defaultPropertyBuilder != null) {
 				CreateControlTree (defaultPropertyBuilder, false, true);
@@ -1545,6 +1578,10 @@ namespace System.Web.Compilation
 
 			if (!childrenAsProperties && typeof (Control).IsAssignableFrom (builder.ControlType))
 				builder.Method.Statements.Add (new CodeMethodReturnStatement (ctrlVar));
+
+#if NET_2_0
+			builder.ProcessGeneratedCode (CompileUnit, BaseType, DerivedType, builder.Method, builder.DataBindingMethod);
+#endif
 		}
 
 #if NET_2_0
@@ -1795,9 +1832,21 @@ namespace System.Web.Compilation
 			object[] attrs;
 			
 #if NET_2_0
-			attrs = member.GetCustomAttributes (typeof (TypeConverterAttribute), true);
-			if (attrs.Length > 0)
-				tca = attrs [0] as TypeConverterAttribute;
+			TypeDescriptionProvider prov = TypeDescriptor.GetProvider (member.ReflectedType);
+			if (prov == null)
+				return null;
+
+			ICustomTypeDescriptor desc = prov.GetTypeDescriptor (member.ReflectedType);
+			PropertyDescriptorCollection coll = desc != null ? desc.GetProperties () : null;
+
+			if (coll == null || coll.Count == 0)
+				return null;
+
+			PropertyDescriptor pd = coll.Find (member.Name, false);
+			if (pd == null)
+				return null;
+
+			return pd.Converter;
 #else
 			attrs = member.GetCustomAttributes (true);
 			
@@ -1807,8 +1856,10 @@ namespace System.Web.Compilation
 				if (tca != null)
 					break;
 			}
-#endif
 
+			if (tca == TypeConverterAttribute.Default)
+				tca = null;
+			
 			if (tca == null)
 				return null;
 
@@ -1827,6 +1878,7 @@ namespace System.Web.Compilation
 				return null;
 
 			return (TypeConverter) Activator.CreateInstance (t);
+#endif
 		}
 		
 		CodeExpression CreateNullableExpression (Type type, CodeExpression inst, bool nullable)
@@ -1889,16 +1941,17 @@ namespace System.Web.Compilation
 			}
 #endif
 			if (type == typeof (string)) {
+#if NET_2_0
+				object[] urlAttr = member.GetCustomAttributes (typeof (UrlPropertyAttribute), true);
+				if (urlAttr.Length != 0)
+					str = HandleUrlProperty ((preConverted && convertedFromAttr is string) ? (string) convertedFromAttr : str, member);
+				else
+#endif
 				if (preConverted)
 					return CreateNullableExpression (originalType,
 									 new CodePrimitiveExpression ((string) convertedFromAttr),
 									 wasNullable);
 				
-#if NET_2_0
-				object[] urlAttr = member.GetCustomAttributes (typeof (UrlPropertyAttribute), true);
-				if (urlAttr.Length != 0)
-					str = HandleUrlProperty (str, member);
-#endif
 				return CreateNullableExpression (originalType, new CodePrimitiveExpression (str), wasNullable);
 			} else if (type == typeof (bool)) {
 				if (preConverted)
@@ -1927,7 +1980,7 @@ namespace System.Web.Compilation
 				return CreateNullableExpression (originalType,
 								 new CodePrimitiveExpression (
 									 Convert.ChangeType (preConverted ? convertedFromAttr : str,
-											     type, CultureInfo.InvariantCulture)),
+											     type, Helpers.InvariantCulture)),
 								 wasNullable);
 
 			if (type == typeof (string [])) {
@@ -2012,10 +2065,34 @@ namespace System.Web.Compilation
 				}
 			}
 
-			TypeConverter converter = preConverted ? cvt :
-				wasNullable ? TypeDescriptor.GetConverter (type) :
-				TypeDescriptor.GetProperties (member.DeclaringType) [member.Name].Converter;
+			TypeConverter converter = preConverted ? cvt : wasNullable ? TypeDescriptor.GetConverter (type) : null;
+			if (converter == null) {
+				PropertyDescriptor pdesc = TypeDescriptor.GetProperties (member.DeclaringType) [member.Name];
+				if (pdesc != null)
+					converter = pdesc.Converter;
+				else {
+					Type memberType;
+					switch (member.MemberType) {
+						case MemberTypes.Field:
+							memberType = ((FieldInfo)member).FieldType;
+							break;
 
+						case MemberTypes.Property:
+							memberType = ((PropertyInfo)member).PropertyType;
+							break;
+
+						default:
+							memberType = null;
+							break;
+					}
+
+					if (memberType == null)
+						return null;
+
+					converter = TypeDescriptor.GetConverter (memberType);
+				}
+			}
+			
 			if (preConverted || (converter != null && SafeCanConvertFrom (typeof (string), converter))) {
 				object value = preConverted ? convertedFromAttr : converter.ConvertFromInvariantString (str);
 
@@ -2128,6 +2205,10 @@ namespace System.Web.Compilation
 		{
 			if (value is System.Web.UI.WebControls.Unit) {
 				System.Web.UI.WebControls.Unit s = (System.Web.UI.WebControls.Unit) value;
+				if (s.IsEmpty) {
+					FieldInfo f = typeof (Unit).GetField ("Empty");
+					return new InstanceDescriptor (f, null);
+				}
 				ConstructorInfo c = typeof(System.Web.UI.WebControls.Unit).GetConstructor (
 					BindingFlags.Instance | BindingFlags.Public,
 					null,
@@ -2139,6 +2220,10 @@ namespace System.Web.Compilation
 			
 			if (value is System.Web.UI.WebControls.FontUnit) {
 				System.Web.UI.WebControls.FontUnit s = (System.Web.UI.WebControls.FontUnit) value;
+				if (s.IsEmpty) {
+					FieldInfo f = typeof (FontUnit).GetField ("Empty");
+					return new InstanceDescriptor (f, null);
+				}
 
 				Type cParamType = null;
 				object cParam = null;
