@@ -36,6 +36,7 @@ using System.Drawing;
 using System.Security.Permissions;
 using System.Threading;
 using System.Text;
+using System.Web.Util;
 
 namespace System.Web.UI.WebControls {
 	// CAS
@@ -50,26 +51,26 @@ namespace System.Web.UI.WebControls {
 	[ControlValueProperty ("SelectedDate", "1/1/0001 12:00:00 AM")]
 	[SupportsEventValidation]
 #endif		
-	public class Calendar : WebControl, IPostBackEventHandler {
-
-		private TableItemStyle dayHeaderStyle;
-		private TableItemStyle dayStyle;
-		private TableItemStyle nextPrevStyle;
-		private TableItemStyle otherMonthDayStyle;
-		private TableItemStyle selectedDayStyle;
-		private TableItemStyle titleStyle;
-		private TableItemStyle todayDayStyle;
-		private TableItemStyle selectorStyle;
-		private TableItemStyle weekendDayStyle;
-		private DateTimeFormatInfo dateInfo;
-		private SelectedDatesCollection selectedDatesCollection;
-		private ArrayList dateList;
-		private DateTime today = DateTime.Today;
-		private static DateTime dateZenith  = new DateTime (2000, 1,1);
-		private const int daysInAWeek = 7;
-		private static readonly object DayRenderEvent = new object ();
-		private static readonly object SelectionChangedEvent = new object ();
-		private static readonly object VisibleMonthChangedEvent = new object ();
+	public class Calendar : WebControl, IPostBackEventHandler
+	{
+		TableItemStyle dayHeaderStyle;
+		TableItemStyle dayStyle;
+		TableItemStyle nextPrevStyle;
+		TableItemStyle otherMonthDayStyle;
+		TableItemStyle selectedDayStyle;
+		TableItemStyle titleStyle;
+		TableItemStyle todayDayStyle;
+		TableItemStyle selectorStyle;
+		TableItemStyle weekendDayStyle;
+		DateTimeFormatInfo dateInfo;
+		SelectedDatesCollection selectedDatesCollection;
+		ArrayList dateList;
+		DateTime today = DateTime.Today;
+		static DateTime dateZenith  = new DateTime (2000, 1,1);
+		const int daysInAWeek = 7;
+		static readonly object DayRenderEvent = new object ();
+		static readonly object SelectionChangedEvent = new object ();
+		static readonly object VisibleMonthChangedEvent = new object ();
 
 		public Calendar ()
 		{
@@ -644,7 +645,7 @@ namespace System.Web.UI.WebControls {
 			}
 		}
 		
-		private DateTime DisplayDate {
+		DateTime DisplayDate {
 			get {
 				DateTime dateTime = VisibleDate;
 				if (dateTime == DateTime.MinValue) // If visibledate is still the default value
@@ -654,7 +655,7 @@ namespace System.Web.UI.WebControls {
 			}
 		}
 
-		private DayOfWeek DisplayFirstDayOfWeek {
+		DayOfWeek DisplayFirstDayOfWeek {
 			get {
 				if (FirstDayOfWeek != FirstDayOfWeek.Default)
 					return (DayOfWeek)  FirstDayOfWeek;
@@ -687,6 +688,9 @@ namespace System.Web.UI.WebControls {
 		void IPostBackEventHandler.RaisePostBackEvent (string arg)
 #endif
 		{
+#if NET_2_0
+			ValidateEvent (UniqueID, arg);
+#endif
 			if (arg.Length < 1)
 				return;
 
@@ -910,7 +914,7 @@ namespace System.Web.UI.WebControls {
 		//
 		// Private methods
 		//
-		private void WriteDayHeader (HtmlTextWriter writer)
+		void WriteDayHeader (HtmlTextWriter writer)
 		{
 			int i, first;
 			string dayName;
@@ -999,7 +1003,7 @@ namespace System.Web.UI.WebControls {
 			writer.RenderEndTag ();
 		}
 
-		private void WriteDay (DateTime date, HtmlTextWriter writer)
+		void WriteDay (DateTime date, HtmlTextWriter writer)
 		{			
 			TableItemStyle style = new TableItemStyle ();
 			TableCell cell = new TableCell ();
@@ -1051,7 +1055,7 @@ namespace System.Web.UI.WebControls {
 			cell.RenderControl (writer);
 		}
 
-		private void WriteDays (HtmlTextWriter writer)
+		void WriteDays (HtmlTextWriter writer)
 		{
 			DateTime date = new DateTime (DisplayDate.Year, DisplayDate.Month, 1); // first date
 			DateTime lastDate;
@@ -1098,7 +1102,7 @@ namespace System.Web.UI.WebControls {
 			}
 		}
 
-		private string BuildLink (string arg, string text, Color foreColor, bool hasLink)
+		string BuildLink (string arg, string text, Color foreColor, bool hasLink)
 		{
 			StringBuilder str = new StringBuilder ();
 			Color clr;
@@ -1134,7 +1138,7 @@ namespace System.Web.UI.WebControls {
 			return str.ToString ();
 		}
 
-		private int GetDaysFromZenith (DateTime date)
+		int GetDaysFromZenith (DateTime date)
 		{
 			TimeSpan span =  date.Subtract (dateZenith);
 			return span.Days;
@@ -1144,7 +1148,7 @@ namespace System.Web.UI.WebControls {
 		void WriteCaption (HtmlTextWriter writer)
 		{
 			if (CaptionAlign != TableCaptionAlign.NotSet)
-				writer.AddAttribute (HtmlTextWriterAttribute.Align, CaptionAlign.ToString (CultureInfo.InvariantCulture));
+				writer.AddAttribute (HtmlTextWriterAttribute.Align, CaptionAlign.ToString (Helpers.InvariantCulture));
 
 			writer.RenderBeginTag (HtmlTextWriterTag.Caption);
 			writer.Write (Caption);
@@ -1152,7 +1156,7 @@ namespace System.Web.UI.WebControls {
 		}
 #endif
 
-		private void WriteTitle (HtmlTextWriter writer)
+		void WriteTitle (HtmlTextWriter writer)
 		{
 			TableCell cellNextPrev = null;
 			TableCell titleCell = new TableCell ();
@@ -1228,7 +1232,7 @@ namespace System.Web.UI.WebControls {
 		}
 
 
-		private string GetNextPrevFormatText (DateTime date, bool next)
+		string GetNextPrevFormatText (DateTime date, bool next)
 		{
 			string text;
 			DateTimeFormatInfo dti = DateInfo;
@@ -1248,17 +1252,17 @@ namespace System.Web.UI.WebControls {
 			return text;
 		}
 
-		private bool IsWeekEnd (DayOfWeek day)
+		bool IsWeekEnd (DayOfWeek day)
 		{
 			return (day == DayOfWeek.Saturday || day == DayOfWeek.Sunday);
 		}
 
-		private double GetCellWidth ()
+		double GetCellWidth ()
 		{
 			return HasWeekSelectors (SelectionMode) ? 100/8 : 100/7;
 		}
 
-		private System.Globalization.Calendar GetGlobalCalendar ()
+		System.Globalization.Calendar GetGlobalCalendar ()
 		{
 			return DateTimeFormatInfo.CurrentInfo.Calendar;
 		}
