@@ -45,7 +45,8 @@ namespace System.Web.UI
 	public sealed class PageParser : TemplateControlParser
 	{
 		PagesEnableSessionState enableSessionState = PagesEnableSessionState.True;
-		bool enableViewStateMac = true;
+		bool enableViewStateMac;
+		bool enableViewStateMacSet;
 		bool smartNavigation;
 		bool haveTrace;
 		bool trace;
@@ -97,22 +98,20 @@ namespace System.Web.UI
 		}
 
 #if NET_2_0
-		internal PageParser (string virtualPath, TextReader reader, HttpContext context)
+		internal PageParser (VirtualPath virtualPath, TextReader reader, HttpContext context)
 			: this (virtualPath, null, reader, context)
 		{
 		}
 		
-		internal PageParser (string virtualPath, string inputFile, TextReader reader, HttpContext context)
+		internal PageParser (VirtualPath virtualPath, string inputFile, TextReader reader, HttpContext context)
 		{
-			this.VirtualPath = new VirtualPath (virtualPath);
+			this.VirtualPath = virtualPath;
 			Context = context;
-			BaseVirtualDir = VirtualPathUtility.GetDirectory (virtualPath, false);
+			BaseVirtualDir = virtualPath.DirectoryNoNormalize;
 			Reader = reader;
-			if (String.IsNullOrEmpty (inputFile)) {
-				HttpRequest req = context != null ? context.Request : null;
-				if (req != null)
-					InputFile = req.MapPath (virtualPath);
-			} else
+			if (String.IsNullOrEmpty (inputFile))
+				InputFile = virtualPath.PhysicalPath;
+			else
 				InputFile = inputFile;
 			SetBaseType (null);
 			AddApplicationAssembly ();
@@ -155,7 +154,12 @@ namespace System.Web.UI
 								    HttpContext context)
 		{
 #if NET_2_0
-			return BuildManager.CreateInstanceFromVirtualPath (virtualPath, typeof (IHttpHandler)) as IHttpHandler;
+			bool isFake = false;
+
+			if (!String.IsNullOrEmpty (inputFile))
+				isFake = !inputFile.StartsWith (HttpRuntime.AppDomainAppPath);
+			
+			return BuildManager.CreateInstanceFromVirtualPath (new VirtualPath (virtualPath, inputFile, isFake), typeof (IHttpHandler)) as IHttpHandler;
 #else
 			PageParser pp = new PageParser (virtualPath, inputFile, context);
 			IHttpHandler h = (IHttpHandler) pp.GetCompiledInstance ();
@@ -170,11 +174,11 @@ namespace System.Web.UI
 			// attribute is processed case-insensitive
 			string enabless = GetString (atts, "EnableSessionState", null);
 			if (enabless != null) {
-				if (String.Compare (enabless, "readonly", true) == 0)
+				if (String.Compare (enabless, "readonly", true, Helpers.InvariantCulture) == 0)
 					enableSessionState = PagesEnableSessionState.ReadOnly;
-				else if (String.Compare (enabless, "true", true) == 0)
+				else if (String.Compare (enabless, "true", true, Helpers.InvariantCulture) == 0)
 					enableSessionState = PagesEnableSessionState.True;
-				else if (String.Compare (enabless, "false", true) == 0)
+				else if (String.Compare (enabless, "false", true, Helpers.InvariantCulture) == 0)
 					enableSessionState = PagesEnableSessionState.False;
 				else
 					ThrowParseException ("Invalid value for enableSessionState: " + enabless);
@@ -318,11 +322,11 @@ namespace System.Web.UI
 			if (clientTarget != null) {
 				clientTarget = clientTarget.Trim ();
 #if NET_2_0
-				ClientTargetSection sec = (ClientTargetSection)WebConfigurationManager.GetWebApplicationSection ("system.web/clientTarget");
+				ClientTargetSection sec = GetConfigSection <ClientTargetSection> ("system.web/clientTarget");
 				ClientTarget ct = null;
 				
 				if ((ct = sec.ClientTargets [clientTarget]) == null)
-					clientTarget = clientTarget.ToLower (CultureInfo.InvariantCulture);
+					clientTarget = clientTarget.ToLowerInvariant ();
 				
 				if (ct == null && (ct = sec.ClientTargets [clientTarget]) == null) {
 					ThrowParseException (String.Format (
@@ -339,7 +343,7 @@ namespace System.Web.UI
 				if (coll != null) {
 					ct = coll [clientTarget];
 					if (ct == null)
-						ct = coll [clientTarget.ToLower ()];
+						ct = coll [clientTarget.ToLower (Helpers.InvariantCulture)];
 				}
 				
 				if (ct == null) {
@@ -380,8 +384,12 @@ namespace System.Web.UI
 			enable_event_validation = GetBool (atts, "EnableEventValidation", enable_event_validation);
 			maintainScrollPositionOnPostBack = GetBool (atts, "MaintainScrollPositionOnPostBack", maintainScrollPositionOnPostBack);
 #endif
+			if (atts.ContainsKey ("EnableViewStateMac")) {
+				enableViewStateMac = GetBool (atts, "EnableViewStateMac", enableViewStateMac);
+				enableViewStateMacSet = true;
+			}
+			
 			// Ignored by now
-			GetString (atts, "EnableViewStateMac", null);
 			GetString (atts, "SmartNavigation", null);
 
 			base.ProcessMainAttributes (atts);
@@ -389,11 +397,11 @@ namespace System.Web.UI
 		
 #if NET_2_0
 		internal override void AddDirective (string directive, Hashtable atts)
-		{			
+		{
 			bool isMasterType = String.Compare ("MasterType", directive, StringComparison.OrdinalIgnoreCase) == 0;
 			bool isPreviousPageType = isMasterType ? false : String.Compare ("PreviousPageType", directive,
 											 StringComparison.OrdinalIgnoreCase) == 0;
-			
+
 			string typeName = null;
 			string virtualPath = null;
 			Type type = null;
@@ -401,7 +409,7 @@ namespace System.Web.UI
 			if (isMasterType || isPreviousPageType) {
 				PageParserFilter pfilter = PageParserFilter;
 				if (pfilter != null)
-					pfilter.PreprocessDirective (directive.ToLower (CultureInfo.InvariantCulture), atts);
+					pfilter.PreprocessDirective (directive.ToLowerInvariant (), atts);
 				
 				typeName = GetString (atts, "TypeName", null);
 				virtualPath = GetString (atts, "VirtualPath", null);
@@ -472,6 +480,10 @@ namespace System.Web.UI
 		internal bool EnableViewStateMac {
 			get { return enableViewStateMac; }
 		}
+
+		internal bool EnableViewStateMacSet {
+			get { return enableViewStateMacSet; }
+		}
 		
 		internal bool SmartNavigation {
 			get { return smartNavigation; }
