@@ -73,9 +73,11 @@ using System.Security.Permissions;
 using System.Security.Principal;
 using System.Threading;
 using System.Web.Caching;
+using System.Web.Compilation;
 using System.Web.Configuration;
 using System.Web.SessionState;
 using System.Web.UI;
+using System.Web.Util;
 
 #if TARGET_J2EE
 using Mainsoft.Web;
@@ -856,12 +858,20 @@ namespace System.Web {
 			if (custom == null) // Sigh
 				throw new NullReferenceException ();
 
-			if (0 == String.Compare (custom, "browser", true, CultureInfo.InvariantCulture))
+			if (0 == String.Compare (custom, "browser", true, Helpers.InvariantCulture))
 				return context.Request.Browser.Type;
 
 			return null;
 		}
 
+		bool ShouldHandleException (Exception e)
+		{
+			if (e is ParseException)
+				return false;
+
+			return true;
+		}
+		
 		//
 		// If we catch an error, queue this error
 		//
@@ -869,15 +879,18 @@ namespace System.Web {
 		{
 			bool first = context.Error == null;
 			context.AddError (e);
-			if (first) {
+			if (first && ShouldHandleException (e)) {
 				EventHandler eh = nonApplicationEvents [errorEvent] as EventHandler;
 				if (eh != null){
 					try {
 						eh (this, EventArgs.Empty);
+						if (stop_processing)
+							context.ClearError ();
 					} catch (ThreadAbortException taex){
 						context.ClearError ();
-						if (FlagEnd.Value == taex.ExceptionState)
-							// This happens on Redirect() or End()
+						if (FlagEnd.Value == taex.ExceptionState || HttpRuntime.DomainUnloading)
+							// This happens on Redirect(), End() and
+							// when unloading the AppDomain
 							Thread.ResetAbort ();
 						else
 							// This happens on Thread.Abort()
@@ -934,19 +947,19 @@ namespace System.Web {
 			} catch (ThreadAbortException taex) {
 				object obj = taex.ExceptionState;
 				Thread.ResetAbort ();
-				stop_processing = true;
 				if (obj is StepTimeout)
 					ProcessError (new HttpException ("The request timed out."));
 				else {
 					context.ClearError ();
-					if (FlagEnd.Value != obj)
+					if (FlagEnd.Value != obj && !HttpRuntime.DomainUnloading)
 						context.AddError (taex);
 				}
-
+				
+				stop_processing = true;
 				PipelineDone ();
 			} catch (Exception e) {
-				stop_processing = true;
 				ProcessError (e);
+				stop_processing = true;
 				PipelineDone ();
 			}
 		}
@@ -1177,6 +1190,9 @@ namespace System.Web {
 			if (stop_processing)
 				yield return true;
 
+#if NET_2_0
+			context.MapRequestHandlerDone = false;
+#endif
 			StartTimer ("BeginRequest");
 			eventHandler = Events [BeginRequestEvent];
 			if (eventHandler != null) {
@@ -1243,6 +1259,7 @@ namespace System.Web {
 				foreach (bool stop in RunHooks (eventHandler))
 					yield return stop;
 			StopTimer ();
+			context.MapRequestHandlerDone = true;
 #endif
 			
 			StartTimer ("GetHandler");
@@ -1481,7 +1498,9 @@ namespace System.Web {
 			cfg = GlobalizationConfiguration.GetInstance (null);
 			if (cfg != null) {
 				app_culture = cfg.Culture;
+				autoCulture = false; // to hush the warning
 				appui_culture = cfg.UICulture;
+				autoUICulture = false; // to hush the warning
 			}
 #endif
 
@@ -1492,14 +1511,14 @@ namespace System.Web {
 			if (app_culture != null) {
 				prev_app_culture = th.CurrentCulture;
 				CultureInfo new_app_culture = GetThreadCulture (Request, app_culture, autoCulture);
-				if (!new_app_culture.Equals (CultureInfo.InvariantCulture))
+				if (!new_app_culture.Equals (Helpers.InvariantCulture))
 					th.CurrentCulture = new_app_culture;
 			}
 
 			if (appui_culture != null) {
 				prev_appui_culture = th.CurrentUICulture;
 				CultureInfo new_app_culture = GetThreadCulture (Request, appui_culture, autoUICulture);
-				if (!new_app_culture.Equals (CultureInfo.InvariantCulture))
+				if (!new_app_culture.Equals (Helpers.InvariantCulture))
 					th.CurrentUICulture = new_app_culture;
 			}
 
@@ -1596,8 +1615,7 @@ namespace System.Web {
 
 			bool allowCache;
 #if NET_2_0
-			global::System.Configuration.Configuration cfg = WebConfigurationManager.OpenWebConfiguration (req.Path, null, req.FilePath);
-			HttpHandlersSection httpHandlersSection = cfg.GetSection ("system.web/httpHandlers") as HttpHandlersSection;
+			HttpHandlersSection httpHandlersSection = WebConfigurationManager.GetSection ("system.web/httpHandlers", req.Path, req.Context) as HttpHandlersSection;
 			ret = httpHandlersSection.LocateHandler (verb, url, out allowCache);
 #else
 			HandlerFactoryConfiguration factory_config = (HandlerFactoryConfiguration) HttpContext.GetAppConfig ("system.web/httpHandlers");
