@@ -28,12 +28,14 @@
 //
 using System;
 using System.Collections;
+using System.Globalization;
 using System.IO;
 using System.Reflection;
 using System.Web.UI;
 using System.Web.SessionState;
 using System.Web.Configuration;
 using System.Threading;
+using System.Web.Util;
 
 using System.Web.Compilation;
 #if TARGET_J2EE
@@ -108,7 +110,7 @@ namespace System.Web {
 				return false;
 
 			if (pi [0].ParameterType != typeof (object) ||
-			    pi [1].ParameterType != typeof (EventArgs))
+			    !typeof (EventArgs).IsAssignableFrom (pi [1].ParameterType))
 				return false;
 			
 			return true;
@@ -479,11 +481,21 @@ namespace System.Web {
 						app_state = new HttpApplicationState ();
 					}
 
-					WatchLocationForRestart("Global.asax");
-					WatchLocationForRestart("global.asax");
-					WatchLocationForRestart(String.Empty, "Web.config", true);
-					WatchLocationForRestart(String.Empty, "web.config", true);
-					WatchLocationForRestart(String.Empty, "Web.Config", true);
+					WatchLocationForRestart ("?lobal.asax");
+#if CODE_DISABLED_UNTIL_SYSTEM_CONFIGURATION_IS_FIXED
+					// This is the correct behavior, but until
+					// System.Configuration is fixed to properly reload
+					// configuration when it is modified on disk, we need to use
+					// the recursive watchers below.
+					WatchLocationForRestart ("?eb.?onfig");
+#else
+					// This is to avoid startup delays. Inotify/FAM code looks
+					// recursively for all subdirectories and adds them to the
+					// watch set. This can take a lot of time for deep directory
+					// trees (see bug #490497)
+					ThreadPool.QueueUserWorkItem (new WaitCallback (SetUpWebConfigWatchers), null);
+#endif
+					
 					needs_init = false;
 #if NET_2_0
 				} catch (Exception) {
@@ -503,6 +515,11 @@ namespace System.Web {
 
 			}
 		}
+
+		static void SetUpWebConfigWatchers (object state)
+		{
+			WatchLocationForRestart (String.Empty, "?eb.?onfig", true);
+		}
 		
 		//
 		// Multiple-threads might hit this one on startup, and we have
@@ -610,7 +627,7 @@ namespace System.Web {
 
                 internal static bool WatchLocationForRestart (string filter)
 	        {
-			return WatchLocationForRestart ("", filter, false);
+			return WatchLocationForRestart (String.Empty, filter, false);
 	        }
 
 		internal static bool WatchLocationForRestart (string virtualPath, string filter)
@@ -685,6 +702,34 @@ namespace System.Web {
 			}
 		}
 
+		internal static void DisableWatcher (string virtualPath, string filter)
+		{
+			EnableWatcherEvents (virtualPath, filter, false);
+		}
+
+		internal static void EnableWatcher (string virtualPath, string filter)
+		{
+			EnableWatcherEvents (virtualPath, filter, true);
+		}
+		
+		static void EnableWatcherEvents (string virtualPath, string filter, bool enable)
+		{
+			lock (watchers_lock) {
+				foreach (FileSystemWatcher watcher in watchers) {
+					if (
+#if NET_2_0
+						String.Compare (watcher.Path, virtualPath, StringComparison.Ordinal) != 0 || String.Compare (watcher.Filter, filter, StringComparison.Ordinal) != 0
+#else
+						String.Compare (watcher.Path, virtualPath) != 0 || String.Compare (watcher.Filter, filter) != 0
+#endif
+					)
+						continue;
+					
+					watcher.EnableRaisingEvents = enable;
+				}
+			}
+		}
+		
 		internal static void EnableWatchers ()
 		{
 			lock (watchers_lock) {
@@ -700,6 +745,36 @@ namespace System.Web {
 
 	        static void OnFileChanged(object sender, FileSystemEventArgs args)
 	        {
+			string name = args.Name;
+			bool isConfig = false;
+			
+			if (StrUtils.EndsWith (name, "onfig", true)) {
+				if (String.Compare (Path.GetFileName (name), "web.config", true, Helpers.InvariantCulture) != 0)
+					return;
+				isConfig = true;
+			} else if (StrUtils.EndsWith (name, "lobal.asax", true) && String.Compare (name, "global.asax", true, Helpers.InvariantCulture) != 0)
+				return;
+
+			// {Inotify,FAM}Watcher will notify about events for a directory regardless
+			// of the filter pattern. This might be a bug in the watchers code, but
+			// since I couldn't find any rationale for the code in there I'd opted for
+			// not removing it and instead working around the issue here. Fix for bug
+			// #495011
+			FileSystemWatcher watcher = sender as FileSystemWatcher;
+			if (watcher != null && String.Compare (watcher.Filter, "?eb.?onfig", true, Helpers.InvariantCulture) == 0 && Directory.Exists (name))
+				return;
+
+#if NET_2_0
+			// We re-enable suppression here since WebConfigurationManager will disable
+			// it after save is done. WebConfigurationManager is called twice by
+			// Configuration - just after opening the target file and just after closing
+			// it. For that reason we will receive two change notifications and if we
+			// disabled suppression here, it would reload the application on the second
+			// change notification.
+			if (isConfig && WebConfigurationManager.SuppressAppReload (true))
+				return;
+#endif
+			
 	        	lock (watchers_lock) {
 				if(app_shutdown)
 					return;
