@@ -188,7 +188,13 @@ namespace System.Web {
 				
 				EnsureCapacity (position + count);
 				byte *src = (byte *) ptr.ToPointer ();
-				memcpy (data + position, src, count);
+				if (count < 32) {
+					byte *dest = (data + position);
+					for (int i = 0; i < count; i++)
+						*dest++ = *src++;
+				} else {
+					memcpy (data + position, src, count);
+				}
 				position += count;
 			}
 
@@ -266,12 +272,16 @@ namespace System.Web {
 				get { return length; }
 			}
 
-			public int Write (byte [] buf, int offset, int count)
+			public unsafe int Write (byte [] buf, int offset, int count)
 			{
 				if (Expandable == false)
 					throw new Exception ("This should not happen.");
 
-				blocks.Write (buf, offset, count);
+				fixed (byte *p = &buf[0]) {
+					IntPtr p2 = new IntPtr (p + offset);
+					blocks.Write (p2, count);
+				}
+
 				length += count;
 				return count;
 			}
@@ -509,10 +519,31 @@ namespace System.Web {
 
 		}
 
-	        public override void Write (byte [] buffer, int offset, int count)
+		public override void Write (byte [] buffer, int offset, int count)
 		{
 			bool buffering = response.BufferOutput;
 
+#if NET_2_0
+			if (buffer == null)
+				throw new ArgumentNullException ("buffer");
+#endif
+
+			int max_count = buffer.Length - offset;
+#if NET_2_0
+			if (offset < 0 || max_count <= 0)
+#else
+			if (offset < 0)
+#endif
+				throw new ArgumentOutOfRangeException ("offset");
+			if (count < 0)
+				throw new ArgumentOutOfRangeException ("count");
+			if (count > max_count)
+				count = max_count;
+#if ONLY_1_1
+			if (max_count <= 0)
+				return;
+#endif
+
 			if (buffering) {
 				// It does not matter whether we're in ApplyFilter or not
 				AppendBuffer (buffer, offset, count);
@@ -717,54 +748,54 @@ namespace System.Web {
 			memcpy4 (dest, src, size);
 		}
 
-	        public override bool CanRead {
-	                get {
-				return false;
-			}	
-	        }
-			
-	        public override bool CanSeek {
-	                get {
+		public override bool CanRead {
+			get {
 				return false;
 			}
-	        }
-	        public override bool CanWrite {
-	                get {
+		}
+
+		public override bool CanSeek {
+			get {
+				return false;
+			}
+		}
+
+		public override bool CanWrite {
+			get {
 				return true;
 			}
-	        }
-		
+		}
+
 		const string notsupported = "HttpResponseStream is a forward, write-only stream";
-		
-	        public override long Length {
-	                get {
-				throw new InvalidOperationException (notsupported);
+
+		public override long Length {
+			get {
+				throw new NotSupportedException (notsupported);
 			}
-	        }
+		}
 	
 		public override long Position {
-	                get {
-				throw new InvalidOperationException (notsupported);
+			get {
+				throw new NotSupportedException (notsupported);
 			}
-	                set {
-				throw new InvalidOperationException (notsupported);
+			set {
+				throw new NotSupportedException (notsupported);
 			}
-	        }
-		
-	        public override long Seek (long offset, SeekOrigin origin)
-		{
-			throw new InvalidOperationException (notsupported);
 		}
 		
-	        public override void SetLength (long value) 
+		public override long Seek (long offset, SeekOrigin origin)
 		{
-			throw new InvalidOperationException (notsupported);
+			throw new NotSupportedException (notsupported);
+		}
+
+		public override void SetLength (long value) 
+		{
+			throw new NotSupportedException (notsupported);
 		}
 	
 		public override int Read (byte [] buffer, int offset, int count)
 		{
-			throw new InvalidOperationException (notsupported);
+			throw new NotSupportedException (notsupported);
 		}
 	}
 }
-
