@@ -223,7 +223,7 @@ namespace System.Xml.Serialization
 					return ReadTypedPrimitive (AnyType);
             }
 
-			object ob = Activator.CreateInstance (typeMap.TypeData.Type);
+			object ob = Activator.CreateInstance (typeMap.TypeData.Type, true);
 
 			Reader.MoveToElement();
 			bool isEmpty = Reader.IsEmptyElement;
@@ -291,16 +291,6 @@ namespace System.Xml.Serialization
 
 		void ReadMembers (ClassMap map, object ob, bool isValueList, bool readByOrder)
 		{
-			// Set the default values of the members
-			if (map.MembersWithDefault != null)
-			{
-				ArrayList members = map.MembersWithDefault;
-				for (int n=0; n<members.Count; n++) {
-					XmlTypeMapMember mem = (XmlTypeMapMember) members[n];
-					SetMemberValueFromAttr (mem, ob, mem.DefaultValue, isValueList);
-				}
-			}
-			
 			// Reads attributes
 			ReadAttributeMembers (map, ob, isValueList);
 
@@ -623,7 +613,7 @@ namespace System.Xml.Serialization
 					return ReadObject (elem.MappedType, elem.IsNullable, true);
 
 				case SchemaTypes.XmlSerializable:
-					object ob = Activator.CreateInstance (elem.TypeData.Type);
+					object ob = Activator.CreateInstance (elem.TypeData.Type, true);
 					return ReadSerializable ((IXmlSerializable)ob);
 
 				default:
@@ -734,7 +724,7 @@ namespace System.Xml.Serialization
 			else	// Must be IEnumerable
 			{
 				if (list == null) {
-					if (canCreateInstance) list = Activator.CreateInstance (type);
+					if (canCreateInstance) list = Activator.CreateInstance (type, true);
 					else throw CreateReadOnlyCollectionException (type.FullName);
 				}
 
@@ -753,7 +743,7 @@ namespace System.Xml.Serialization
 			if (listType.IsArray)
 				return EnsureArrayIndex (null, 0, listType.GetElementType());
 			else
-				return Activator.CreateInstance (listType);
+				return Activator.CreateInstance (listType, true);
 		}
 		
 		object InitializeList (TypeData listType)
@@ -761,7 +751,7 @@ namespace System.Xml.Serialization
 			if (listType.Type.IsArray)
 				return null;
 			else
-				return Activator.CreateInstance (listType.Type);
+				return Activator.CreateInstance (listType.Type, true);
 		}
 
 		void FillList (object list, object items)
@@ -812,6 +802,8 @@ namespace System.Xml.Serialization
 
 		object GetEnumValue (XmlTypeMapping typeMap, string val)
 		{
+			if (val == null)
+				return null;
 			EnumMap map = (EnumMap) typeMap.ObjectMap;
 			string ev = map.GetEnumName (typeMap.TypeFullName, val);
 			if (ev == null) throw CreateUnknownConstantException (val, typeMap.TypeData.Type);
@@ -825,7 +817,7 @@ namespace System.Xml.Serialization
 			{
 				if (Reader.LocalName == typeMap.ElementName && Reader.NamespaceURI == typeMap.Namespace)
 				{
-					object ob = Activator.CreateInstance (typeMap.TypeData.Type);
+					object ob = Activator.CreateInstance (typeMap.TypeData.Type, true);
 					return ReadSerializable ((IXmlSerializable)ob);
 				}
 				else
