@@ -51,11 +51,29 @@ namespace System.Xml
 		{
 		}
 
+		static XmlException invalidDataException = new XmlException ("invalid data.");
+
 		public override void Close ()
 		{
 			this.input.Close ();
 		}
 
+		public override int Read ([In, Out] char[] dest_buffer, int index, int count)
+		{
+			try {
+				return base.Read (dest_buffer, index, count);
+			}
+#if NET_1_1
+			catch (System.ArgumentException) {
+				throw invalidDataException;
+			}
+#else
+			catch (System.Text.DecoderFallbackException) {
+				throw invalidDataException;
+			}
+#endif
+		}
+
 		protected override void Dispose (bool disposing)
 		{
 			base.Dispose (disposing);
@@ -343,7 +361,7 @@ namespace System.Xml
 
 		static string GetStringFromBytes (byte [] bytes, int index, int count)
 		{
-#if NET_2_1
+#if NET_2_1 && !MONOTOUCH
 			char [] chars = new char [count];
 			for (int i = index; i < count; i++)
 				chars [i] = (char) bytes [i];
@@ -489,7 +507,6 @@ namespace System.Xml
 					return c;
 				}
 			}
-			throw new InvalidOperationException ();
 		}
 
 		public Encoding ActualEncoding {
