@@ -2188,6 +2188,45 @@ World";
 			Assert.AreEqual (expected, sw.ToString ());
 		}
 
+		[Test] // bug #556619
+		public void RejectCurrencySymbolAsNumber ()
+		{
+			XslTransform xslt = new XslTransform ();
+			StringWriter sw = new StringWriter ();
+			string s =
+@"<?xml version=""1.0""?>
+
+<xsl:stylesheet
+  version=""1.0""
+  xmlns:xsl=""http://www.w3.org/1999/XSL/Transform""
+  >
+
+  <xsl:template name=""F"">
+    <xsl:param name=""p""/>
+    <xsl:choose>
+      <xsl:when test=""number($p)"">
+YES <xsl:value-of select=""number($p)""/>
+      </xsl:when>
+      <xsl:otherwise>
+NO
+      </xsl:otherwise>
+    </xsl:choose>
+  </xsl:template>
+  <xsl:template match=""/"">
+    <xsl:call-template name=""F"">
+      <xsl:with-param name=""p"">$4$2</xsl:with-param>
+    </xsl:call-template>
+  </xsl:template>
+
+</xsl:stylesheet>";
+			xslt.Load (new XmlTextReader (new StringReader (s)));
+			xslt.Transform (new XPathDocument (new StringReader (
+@"<?xml version=""1.0""?>
+<root>
+</root>")), null, sw);
+			Assert.IsTrue (sw.ToString ().IndexOf ("NO") > 0);
+		}
+
 #if NET_2_0
 		[Test] // bug #349375
 		public void PreserveWhitespace ()
@@ -2243,6 +2282,7 @@ World";
 
 		[Test] // reverse case of #349375
 		[Category ("NotWorking")]
+//		[Category ("NotDotNet")]
 		public void PreserveWhitespace2 ()
 		{
 			XslCompiledTransform xslt = new XslCompiledTransform ();
@@ -2291,7 +2331,8 @@ World";
     </p>
   </l0>
 </root>")), null, sw);
-			Assert.AreEqual (@"<y><t yes-one-node="""">1</t><t not-node=""""></t></y>", sw.ToString ());
+//			Assert.AreEqual (@"<y><t yes-one-node="""">1</t><t not-node=""""></t></y>", sw.ToString ());
+			Assert.AreEqual ("\r\n  \r\n    <y><t yes-one-node=\"\">1</t><t not-node=\"\"></t></y>\r\n  \r\n", sw.ToString ());
 		}
 
 		[Test]
@@ -2305,6 +2346,77 @@ World";
 			xslTransform.Transform (new XPathDocument ("Test/XmlFiles/xsl/391424.xml", XmlSpace.Preserve), null, sw);
 			Assert.AreEqual ("<?xml version=\"1.0\" encoding=\"utf-16\"?>Document found", sw.ToString ());
 		}
+
+		[Test]
+		public void XslTextElement_PreservesWhitespace () // bug 450797
+		{
+			XslCompiledTransform xslt = new XslCompiledTransform ();
+			xslt.Load (new XmlTextReader (new StringReader (@"
+<xsl:stylesheet
+  xmlns:xsl=""http://www.w3.org/1999/XSL/Transform""
+  version=""1.0"">
+  <xsl:output method='text' omit-xml-declaration='yes'/>
+  <xsl:template match='foo'>
+    <xsl:text> </xsl:text>
+  </xsl:template>
+</xsl:stylesheet>")));
+			StringWriter sw = new StringWriter ();
+			xslt.Transform (new XmlTextReader (new StringReader (@"<foo>bar</foo>")), null, sw);
+			Assert.AreEqual (" ", sw.ToString ());
+		}
+
+		[Test]
+		public void Bug487065 ()
+		{
+			using (XmlReader input = GetInput ()) {
+				using (XmlWriter output = XmlWriter.Create (Path.Combine(Path.GetTempPath(), "out.xml"))) {
+					PreProcess (input, output);
+					Assert.IsTrue (valueHasBeenSet, "#1");
+				}
+			}
+		}
+
+		private static XmlReader GetInput ()
+		{
+			return XmlReader.Create (new StringReader ("<cb:config-template xmlns:cb=\"urn:ccnet.config.builder\"><cb:define name=\"var1\">value1</cb:define></cb:config-template>"));
+		}
+
+		private void PreProcess (XmlReader input, XmlWriter output)
+		{
+			var xsltsettings = new XsltSettings (true, true);
+			var xsltpreprocess = new XslCompiledTransform (true);
+
+			var stream =
+				new StringReader (@"
+<xsl:stylesheet version='1.0' xmlns:xsl='http://www.w3.org/1999/XSL/Transform'
+ xmlns:msxsl='urn:schemas-microsoft-com:xslt'
+ xmlns:cb='urn:ccnet.config.builder'    
+ xmlns:env='environment'
+ xmlns:exsl='http://exslt.org/common'
+ exclude-result-prefixes='msxsl cb env exsl'>
+  <xsl:template match='cb:define[not(*) and @name]'>
+    <xsl:for-each select='env:define_text_constant(@name,.)'/>
+  </xsl:template>
+</xsl:stylesheet>");
+
+			using (XmlReader xsltreader = XmlReader.Create (stream))
+				xsltpreprocess.Load (xsltreader, xsltsettings, new XmlUrlResolver());
+
+			var xsltargs = new XsltArgumentList ();
+
+			// The XSLT calls extension functions in _current_env.
+			xsltargs.AddExtensionObject ("environment", this);
+			xsltpreprocess.Transform (input, xsltargs, output, null);
+		}
+
+		public void define_text_constant (string name, string value)
+		{
+			Assert.AreEqual("var1", name, "#a1");
+			Assert.AreEqual("value1", value, "#a2");
+			valueHasBeenSet = true;
+		}
+
+		private bool valueHasBeenSet;
 #endif
 	}
 }
