@@ -34,6 +34,7 @@ namespace Mono.CSharp
 	using System;
 	using System.CodeDom;
 	using System.CodeDom.Compiler;
+	using System.ComponentModel;
 	using System.IO;
 	using System.Text;
 	using System.Reflection;
@@ -216,6 +217,22 @@ namespace Mono.CSharp
 			mcsOutput = new StringCollection ();
 			mcsOutMutex = new Mutex ();
 #endif
+
+			string monoPath = Environment.GetEnvironmentVariable ("MONO_PATH");
+			if (monoPath == null)
+				monoPath = String.Empty;
+			
+			string privateBinPath = AppDomain.CurrentDomain.SetupInformation.PrivateBinPath;
+			if (privateBinPath != null && privateBinPath.Length > 0)
+				monoPath = String.Format ("{0}:{1}", privateBinPath, monoPath);
+
+			if (monoPath.Length > 0) {
+				StringDictionary dict = mcs.StartInfo.EnvironmentVariables;
+				if (dict.ContainsKey ("MONO_PATH"))
+					dict ["MONO_PATH"] = monoPath;
+				else
+					dict.Add ("MONO_PATH", monoPath);
+			}
 			
 			mcs.StartInfo.CreateNoWindow=true;
 			mcs.StartInfo.UseShellExecute=false;
@@ -227,7 +244,16 @@ namespace Mono.CSharp
 			
 			try {
 				mcs.Start();
+			} catch (Exception e) {
+				Win32Exception exc = e as Win32Exception;
+				if (exc != null) {
+					throw new SystemException (String.Format ("Error running {0}: {1}", mcs.StartInfo.FileName,
+									Win32Exception.W32ErrorMessage (exc.NativeErrorCode)));
+				}
+				throw;
+			}
 
+			try {
 #if NET_2_0
 				mcs.BeginOutputReadLine ();
 				mcs.BeginErrorReadLine ();
@@ -321,6 +347,10 @@ namespace Mono.CSharp
 			else
 				args.Append("/target:library ");
 
+			string privateBinPath = AppDomain.CurrentDomain.SetupInformation.PrivateBinPath;
+			if (privateBinPath != null && privateBinPath.Length > 0)
+				args.AppendFormat ("/lib:\"{0}\" ", privateBinPath);
+			
 			if (options.Win32Resource != null)
 				args.AppendFormat("/win32res:\"{0}\" ",
 					options.Win32Resource);
