@@ -204,13 +204,14 @@ namespace System.Net
 			}
 		}
 
+		[MonoTODO ("We don't support KeepAlive = true")]
 		public bool KeepAlive {
 			get {
 				return keepAlive;
 			}
 			set {
 				CheckRequestStarted ();
-				keepAlive = value;
+				//keepAlive = value;
 			}
 		}
 
@@ -366,7 +367,7 @@ namespace System.Net
 
 				if (!InFinalState ()) {
 					State = RequestState.Aborted;
-					ftpResponse = new FtpWebResponse (requestUri, method, FtpStatusCode.FileActionAborted, "Aborted by request");
+					ftpResponse = new FtpWebResponse (this, requestUri, method, FtpStatusCode.FileActionAborted, "Aborted by request");
 				}
 			}
 		}
@@ -500,7 +501,7 @@ namespace System.Net
 		void ProcessRequest () {
 
 			if (State == RequestState.Scheduled) {
-				ftpResponse = new FtpWebResponse (requestUri, method, keepAlive);
+				ftpResponse = new FtpWebResponse (this, requestUri, method, keepAlive);
 
 				try {
 					ProcessMethod ();
@@ -635,14 +636,14 @@ namespace System.Net
 			FtpStatus status;
 			
 			if (method == WebRequestMethods.Ftp.PrintWorkingDirectory)
-				method = ChangeDir;
+				method = "PWD";
 
 			if (method == WebRequestMethods.Ftp.Rename)
 				method = RenameFromCommand;
 			
 			status = SendCommand (method, file_name);
 
-			ftpResponse.Stream = new EmptyStream ();
+			ftpResponse.Stream = Stream.Null;
 			
 			string desc = status.StatusDescription;
 
@@ -719,14 +720,6 @@ namespace System.Net
 		{
 			State = RequestState.OpeningData;
 
-			// Handle content offset
-			if (offset > 0) {
-				FtpStatus status = SendCommand (RestCommand, offset.ToString ());
-
-				if (status.StatusCode != FtpStatusCode.FileCommandPending)
-					throw CreateExceptionFromResponse (status);
-			}
-
 			OpenDataConnection ();
 
 			State = RequestState.TransferInProgress;
@@ -863,7 +856,7 @@ namespace System.Net
 
 		Exception CreateExceptionFromResponse (FtpStatus status)
 		{
-			FtpWebResponse ftpResponse = new FtpWebResponse (requestUri, method, status);
+			FtpWebResponse ftpResponse = new FtpWebResponse (this, requestUri, method, status);
 			
 			WebException exc = new WebException ("Server returned an error: " + status.StatusDescription, 
 				null, WebExceptionStatus.ProtocolError, ftpResponse);
@@ -883,6 +876,12 @@ namespace System.Net
 				CloseConnection ();
 		}
 
+		internal void OperationCompleted ()
+		{
+			if(!keepAlive)
+				CloseConnection ();
+		}
+
 		void SetCompleteWithError (Exception exc)
 		{
 			if (asyncResult != null) {
@@ -937,6 +936,13 @@ namespace System.Net
 			
 			Socket s = InitDataConnection ();
 
+			// Handle content offset
+			if (offset > 0) {
+				status = SendCommand (RestCommand, offset.ToString ());
+				if (status.StatusCode != FtpStatusCode.FileCommandPending)
+					throw CreateExceptionFromResponse (status);
+			}
+
 			if (method != WebRequestMethods.Ftp.ListDirectory && method != WebRequestMethods.Ftp.ListDirectoryDetails &&
 			    method != WebRequestMethods.Ftp.UploadFileWithUniqueName) {
 				status = SendCommand (method, file_name);
@@ -1077,7 +1083,7 @@ namespace System.Net
 				if (!Int32.TryParse (response.Substring (0, 3), out code))
 					return ServiceNotAvailable ();
 
-				if (response [3] == '-'){
+				if (response.Length > 3 && response [3] == '-'){
 					string line = null;
 					string find = code.ToString() + ' ';
 					while (true){
@@ -1134,13 +1140,6 @@ namespace System.Net
 			if (InFinalState ())
 				throw new InvalidOperationException ("Cannot change final state");
 		}
-
-		class EmptyStream : MemoryStream
-		{
-			internal EmptyStream ()
-				: base (new byte [0], false) {
-			}
-		}
 	}
 }
 
