@@ -8,7 +8,7 @@
 //	Miguel de Icaza (miguel@ximian.com)
 //
 // Copyright 2003 Ximian, Inc. (http://www.ximian.com)
-// Copyright 2006, 2007 Novell, Inc. (http://www.novell.com)
+// Copyright 2006, 2010 Novell, Inc. (http://www.novell.com)
 //
 //
 // Permission is hereby granted, free of charge, to any person obtaining
@@ -103,7 +103,7 @@ namespace System.Net
 			for (int i = '0'; i <= '9'; i++, index++)
 				hexBytes [index] = (byte) i;
 
-			for (int i = 'A'; i <= 'F'; i++, index++)
+			for (int i = 'a'; i <= 'f'; i++, index++)
 				hexBytes [index] = (byte) i;
 		}
 		
@@ -117,7 +117,7 @@ namespace System.Net
 			get {
 				if (baseString == null) {
 					if (baseAddress == null)
-						return "";
+						return string.Empty;
 				}
 
 				baseString = baseAddress.ToString ();
@@ -125,7 +125,7 @@ namespace System.Net
 			}
 			
 			set {
-				if (value == null || value == "") {
+				if (value == null || value.Length == 0) {
 					baseAddress = null;
 				} else {
 					baseAddress = new Uri (value);
@@ -196,7 +196,7 @@ namespace System.Net
 			get { return encoding; }
 			set {
 				if (value == null)
-					throw new ArgumentNullException ("value");
+					throw new ArgumentNullException ("Encoding");
 				encoding = value;
 			}
 		}
@@ -272,13 +272,15 @@ namespace System.Net
 			
 			try {
 				request = SetupRequest (address);
-				WebResponse response = request.GetResponse ();
-				Stream st = ProcessResponse (response);
+				WebResponse response = GetWebResponse (request);
+				Stream st = response.GetResponseStream ();
 				return ReadAll (st, (int) response.ContentLength, userToken);
 			} catch (ThreadInterruptedException){
 				if (request != null)
 					request.Abort ();
 				throw;
+			} catch (WebException wexc) {
+				throw;
 			} catch (Exception ex) {
 				throw new WebException ("An error occurred " +
 					"performing a WebClient request.", ex);
@@ -315,6 +317,8 @@ namespace System.Net
 				async = false;
 #endif				
 				DownloadFileCore (address, fileName, null);
+			} catch (WebException wexc) {
+				throw;
 			} catch (Exception ex) {
 				throw new WebException ("An error occurred " +
 					"performing a WebClient request.", ex);
@@ -330,8 +334,8 @@ namespace System.Net
 			using (FileStream f = new FileStream (fileName, FileMode.Create)) {
 				try {
 					request = SetupRequest (address);
-					WebResponse response = request.GetResponse ();
-					Stream st = ProcessResponse (response);
+					WebResponse response = GetWebResponse (request);
+					Stream st = response.GetResponseStream ();
 					
 					int cLength = (int) response.ContentLength;
 					int length = (cLength <= -1 || cLength > 32*1024) ? 32*1024 : cLength;
@@ -389,8 +393,10 @@ namespace System.Net
 				async = false;
 #endif				
 				request = SetupRequest (address);
-				WebResponse response = request.GetResponse ();
-				return ProcessResponse (response);
+				WebResponse response = GetWebResponse (request);
+				return response.GetResponseStream ();
+			} catch (WebException wexc) {
+				throw;
 			} catch (Exception ex) {
 				throw new WebException ("An error occurred " +
 					"performing a WebClient request.", ex);
@@ -446,6 +452,8 @@ namespace System.Net
 #endif				
 				WebRequest request = SetupRequest (address, method, true);
 				return request.GetRequestStream ();
+			} catch (WebException wexc) {
+				throw;
 			} catch (Exception ex) {
 				throw new WebException ("An error occurred " +
 					"performing a WebClient request.", ex);
@@ -541,8 +549,8 @@ namespace System.Net
 					stream.Write (data, 0, contentLength);
 				}
 				
-				WebResponse response = request.GetResponse ();
-				Stream st = ProcessResponse (response);
+				WebResponse response = GetWebResponse (request);
+				Stream st = response.GetResponseStream ();
 				return ReadAll (st, (int) response.ContentLength, userToken);
 			} catch (ThreadInterruptedException){
 				if (request != null)
@@ -594,6 +602,8 @@ namespace System.Net
 				async = false;
 #endif				
 				return UploadFileCore (address, method, fileName, null);
+			} catch (WebException wexc) {
+				throw;
 			} catch (Exception ex) {
 				throw new WebException ("An error occurred " +
 					"performing a WebClient request.", ex);
@@ -651,8 +661,8 @@ namespace System.Net
 				reqStream.Write (realBoundary, 0, realBoundary.Length);
 				reqStream.Close ();
 				reqStream = null;
-				WebResponse response = request.GetResponse ();
-				Stream st = ProcessResponse (response);
+				WebResponse response = GetWebResponse (request);
+				Stream st = response.GetResponseStream ();
 				resultBytes = ReadAll (st, (int) response.ContentLength, userToken);
 			} catch (ThreadInterruptedException){
 				if (request != null)
@@ -715,6 +725,8 @@ namespace System.Net
 				async = false;
 #endif				
 				return UploadValuesCore (address, method, data, null);
+			} catch (WebException wexc) {
+				throw;
 			} catch (Exception ex) {
 				throw new WebException ("An error occurred " +
 					"performing a WebClient request.", ex);
@@ -738,13 +750,12 @@ namespace System.Net
 			Headers ["Content-Type"] = urlEncodedCType;
 			WebRequest request = SetupRequest (uri, method, true);
 			try {
-				Stream rqStream = request.GetRequestStream ();
 				MemoryStream tmpStream = new MemoryStream ();
 				foreach (string key in data) {
-					byte [] bytes = Encoding.ASCII.GetBytes (key);
+					byte [] bytes = Encoding.UTF8.GetBytes (key);
 					UrlEncodeAndWrite (tmpStream, bytes);
 					tmpStream.WriteByte ((byte) '=');
-					bytes = Encoding.ASCII.GetBytes (data [key]);
+					bytes = Encoding.UTF8.GetBytes (data [key]);
 					UrlEncodeAndWrite (tmpStream, bytes);
 					tmpStream.WriteByte ((byte) '&');
 				}
@@ -754,12 +765,14 @@ namespace System.Net
 					tmpStream.SetLength (--length); // remove trailing '&'
 				
 				byte [] buf = tmpStream.GetBuffer ();
-				rqStream.Write (buf, 0, length);
-				rqStream.Close ();
+				request.ContentLength = length;
+				using (Stream rqStream = request.GetRequestStream ()) {
+					rqStream.Write (buf, 0, length);
+				}
 				tmpStream.Close ();
 				
-				WebResponse response = request.GetResponse ();
-				Stream st = ProcessResponse (response);
+				WebResponse response = GetWebResponse (request);
+				Stream st = response.GetResponseStream ();
 				return ReadAll (st, (int) response.ContentLength, userToken);
 			} catch (ThreadInterruptedException) {
 				request.Abort ();
@@ -925,7 +938,7 @@ namespace System.Net
 		
 		WebRequest SetupRequest (Uri uri)
 		{
-			WebRequest request = WebRequest.Create (uri);
+			WebRequest request = GetWebRequest (uri);
 #if NET_2_0
 			if (Proxy != null)
 				request.Proxy = Proxy;
@@ -980,12 +993,6 @@ namespace System.Net
 			return request;
 		}
 
-		Stream ProcessResponse (WebResponse response)
-		{
-			responseHeaders = response.Headers;
-			return response.GetResponseStream ();
-		}
-
 		byte [] ReadAll (Stream stream, int length, object userToken)
 		{
 			MemoryStream ms = null;
@@ -1228,8 +1235,8 @@ namespace System.Net
 					WebRequest request = null;
 					try {
 						request = SetupRequest ((Uri) args [0]);
-						WebResponse response = request.GetResponse ();
-						Stream stream = ProcessResponse (response);
+						WebResponse response = GetWebResponse (request);
+						Stream stream = response.GetResponseStream ();
 						OnOpenReadCompleted (
 							new OpenReadCompletedEventArgs (stream, null, false, args [1]));
 					} catch (ThreadInterruptedException){
@@ -1395,7 +1402,7 @@ namespace System.Net
 				throw new ArgumentNullException ("data");
 			
 			lock (this) {
-				SetBusy ();
+				CheckBusy ();
 				async = true;
 				
 				async_thread = new Thread (delegate (object state) {
@@ -1533,21 +1540,32 @@ namespace System.Net
 				UploadValuesCompleted (this, args);
 		}
 
-		[MonoNotSupported("")]
-		protected virtual WebRequest GetWebRequest (Uri address)
-		{
-			throw new NotImplementedException ();
-		}
-
-		protected virtual WebResponse GetWebResponse (WebRequest request)
-		{
-			return request.GetResponse ();
-		}
-
 		protected virtual WebResponse GetWebResponse (WebRequest request, IAsyncResult result)
 		{
-			return request.EndGetResponse (result);
+			WebResponse response = request.EndGetResponse (result);
+			responseHeaders = response.Headers;
+			return response;
 		}
 #endif
+
+#if NET_2_0
+		protected virtual
+#endif
+		WebRequest GetWebRequest (Uri address)
+		{
+			return WebRequest.Create (address);
+		}
+
+#if NET_2_0
+		protected virtual
+#endif
+		WebResponse GetWebResponse (WebRequest request)
+		{
+			WebResponse response = request.GetResponse ();
+			responseHeaders = response.Headers;
+			return response;
+		}
+
 	}
 }
+
