@@ -112,10 +112,15 @@ namespace System {
 
 		// Constructors		
 
+#if NET_2_1 && !MONOTOUCH
+		public Uri (string uriString) : this (uriString, UriKind.Absolute) 
+		{
+		}
+#else
 		public Uri (string uriString) : this (uriString, false) 
 		{
 		}
-
+#endif
 		protected Uri (SerializationInfo serializationInfo, 
 			       StreamingContext streamingContext) :
 			this (serializationInfo.GetString ("AbsoluteUri"), true)
@@ -143,13 +148,55 @@ namespace System {
 				break;
 			default:
 				string msg = Locale.GetText ("Invalid UriKind value '{0}'.", uriKind);
-				throw new ArgumentException ("uriKind", msg);
+				throw new ArgumentException (msg);
+			}
+		}
+
+		//
+		// An exception-less constructor, returns success
+		// condition on the out parameter `success'.
+		//
+		Uri (string uriString, UriKind uriKind, out bool success)
+		{
+			if (uriString == null) {
+				success = false;
+				return;
+			}
+
+			if (uriKind != UriKind.RelativeOrAbsolute &&
+				uriKind != UriKind.Absolute &&
+				uriKind != UriKind.Relative) {
+				string msg = Locale.GetText ("Invalid UriKind value '{0}'.", uriKind);
+				throw new ArgumentException (msg);
+			}
+
+			source = uriString;
+			if (ParseNoExceptions (uriKind, uriString) != null)
+				success = false;
+			else {
+				success = true;
+				
+				switch (uriKind) {
+				case UriKind.Absolute:
+					if (!IsAbsoluteUri)
+						success = false;
+					break;
+				case UriKind.Relative:
+					if (IsAbsoluteUri)
+						success = false;
+					break;
+				case UriKind.RelativeOrAbsolute:
+					break;
+				default:
+					success = false;
+					break;
+				}
 			}
 		}
 
 		public Uri (Uri baseUri, Uri relativeUri)
-			: this (baseUri, relativeUri.OriginalString, false)
 		{
+			Merge (baseUri, relativeUri == null ? String.Empty : relativeUri.OriginalString);
 			// FIXME: this should call UriParser.Resolve
 		}
 
@@ -177,8 +224,8 @@ namespace System {
 #endif
 
 		public Uri (Uri baseUri, string relativeUri) 
-			: this (baseUri, relativeUri, false) 
 		{
+			Merge (baseUri, relativeUri);
 			// FIXME: this should call UriParser.Resolve
 		}
 
@@ -187,9 +234,17 @@ namespace System {
 #endif
 		public Uri (Uri baseUri, string relativeUri, bool dontEscape) 
 		{
+			userEscaped = dontEscape;
+			Merge (baseUri, relativeUri);
+		}
+
+		private void Merge (Uri baseUri, string relativeUri)
+		{
 #if NET_2_0
 			if (baseUri == null)
 				throw new ArgumentNullException ("baseUri");
+			if (!baseUri.IsAbsoluteUri)
+				throw new ArgumentOutOfRangeException ("baseUri");
 			if (relativeUri == null)
 				relativeUri = String.Empty;
 #else
@@ -198,8 +253,6 @@ namespace System {
 #endif
 			// See RFC 2396 Par 5.2 and Appendix C
 
-			userEscaped = dontEscape;
-
 			// Check Windows UNC (for // it is scheme/host separator)
 			if (relativeUri.Length >= 2 && relativeUri [0] == '\\' && relativeUri [1] == '\\') {
 				source = relativeUri;
@@ -425,7 +478,7 @@ namespace System {
 				return host; 
 			} 
 		}
-
+#if !NET_2_1 || MONOTOUCH
 		public UriHostNameType HostNameType { 
 			get {
 				EnsureAbsoluteUri ();
@@ -446,6 +499,8 @@ namespace System {
 			} 
 		}
 
+#endif // NET_2_1
+
 		public bool IsDefaultPort { 
 			get {
 				EnsureAbsoluteUri ();
@@ -460,6 +515,7 @@ namespace System {
 			}
 		}
 
+#if !NET_2_1 || MONOTOUCH
 		public bool IsLoopback { 
 			get {
 				EnsureAbsoluteUri ();
@@ -474,20 +530,24 @@ namespace System {
 
 				if (host == "loopback" || host == "localhost") 
 					return true;
-					
-				try {
-					if (IPAddress.Loopback.Equals (IPAddress.Parse (host)))
-						return true;
-				} catch (FormatException) {}
 
-				try {
-					return IPv6Address.IsLoopback (IPv6Address.Parse (host));
-				} catch (FormatException) {}
-				
+				IPAddress result;
+				if (IPAddress.TryParse (host, out result))
+					if (IPAddress.Loopback.Equals (result))
+						return true;
+
+				IPv6Address result6;
+				if (IPv6Address.TryParse (host, out result6)){
+					if (IPv6Address.IsLoopback (result6))
+						return true;
+				}
+
 				return false;
 			} 
 		}
 
+#endif // NET_2_1
+
 		public bool IsUnc {
 			// rule: This should be true only if
 			//   - uri string starts from "\\", or
@@ -640,16 +700,20 @@ namespace System {
 		public bool IsAbsoluteUri {
 			get { return isAbsoluteUri; }
 		}
-
+#endif
 		// LAMESPEC: source field is supplied in such case that this
 		// property makes sense. For such case that source field is
 		// not supplied (i.e. .ctor(Uri, string), this property
 		// makes no sense. To avoid silly regression it just returns
 		// ToString() value now. See bug #78374.
-		public string OriginalString {
+#if NET_2_0
+		public
+#else
+		internal
+#endif
+		string OriginalString {
 			get { return source != null ? source : ToString (); }
 		}
-#endif
 
 		// Methods		
 
@@ -664,27 +728,38 @@ namespace System {
 			if (IsDomainAddress (name))
 				return UriHostNameType.Dns;				
 				
-			try {
-				IPv6Address.Parse (name);
+			IPv6Address addr;
+			if (IPv6Address.TryParse (name, out addr))
 				return UriHostNameType.IPv6;
-			} catch (FormatException) {}
 			
 			return UriHostNameType.Unknown;
 		}
 		
 		internal static bool IsIPv4Address (string name)
-		{		
+		{
 			string [] captures = name.Split (new char [] {'.'});
 			if (captures.Length != 4)
 				return false;
+
 			for (int i = 0; i < 4; i++) {
+				int length;
+
+				length = captures [i].Length;
+				if (length == 0)
+					return false;
+				uint number;
+#if NET_2_0
+				if (!UInt32.TryParse (captures [i], out number))
+					return false;
+#else
 				try {
-					int d = Int32.Parse (captures [i], CultureInfo.InvariantCulture);
-					if (d < 0 || d > 255)
-						return false;
+					number = UInt32.Parse (captures [i]);
 				} catch (Exception) {
 					return false;
 				}
+#endif
+				if (number > 255)
+					return false;
 			}
 			return true;
 		}			
@@ -710,6 +785,7 @@ namespace System {
 			
 			return true;
 		}
+#if !NET_2_1
 
 #if NET_2_0
 		[Obsolete("This method does nothing, it has been obsoleted")]
@@ -722,6 +798,16 @@ namespace System {
 			//
 		}
 
+		[MonoTODO ("Find out what this should do")]
+#if NET_2_0
+		[Obsolete]
+#endif
+		protected virtual void CheckSecurity ()
+		{
+		}
+
+#endif // NET_2_1
+
 		// defined in RFC3986 as = ALPHA *( ALPHA / DIGIT / "+" / "-" / ".")
 		public static bool CheckSchemeName (string schemeName) 
 		{
@@ -754,14 +840,6 @@ namespace System {
 #endif
 		}
 
-		[MonoTODO ("Find out what this should do")]
-#if NET_2_0
-		[Obsolete]
-#endif
-		protected virtual void CheckSecurity ()
-		{
-		}
-
 		public override bool Equals (object comparant) 
 		{
 			if (comparant == null) 
@@ -1062,10 +1140,14 @@ namespace System {
 			path = EscapeString (path);
 		}
 
-#if NET_2_0
+#if NET_2_1 && !MONOTOUCH
+		static string EscapeString (string str)
+#else
+	#if NET_2_0
 		[Obsolete]
-#endif
+	#endif
 		protected static string EscapeString (string str) 
+#endif
 		{
 			return EscapeString (str, false, true, true);
 		}
@@ -1146,10 +1228,14 @@ namespace System {
 			}
 		}
 
-#if NET_2_0
+#if NET_2_1 && !MONOTOUCH
+		string Unescape (string str)
+#else
+	#if NET_2_0
 		[Obsolete]
-#endif
+	#endif
 		protected virtual string Unescape (string str)
+#endif
 		{
 			return Unescape (str, false);
 		}
@@ -1206,17 +1292,21 @@ namespace System {
 			path = path.Replace ("\\", "/");
 		}
 
-		private void ParseAsWindowsAbsoluteFilePath (string uriString)
+		//
+		// Returns null on success, string with error on failure
+		//
+		private string ParseAsWindowsAbsoluteFilePath (string uriString)
 		{
-			if (uriString.Length > 2 && uriString [2] != '\\'
-					&& uriString [2] != '/')
-				throw new UriFormatException ("Relative file path is not allowed.");
+			if (uriString.Length > 2 && uriString [2] != '\\' && uriString [2] != '/')
+				return "Relative file path is not allowed.";
 			scheme = UriSchemeFile;
 			host = String.Empty;
 			port = -1;
 			path = uriString.Replace ("\\", "/");
 			fragment = String.Empty;
 			query = String.Empty;
+
+			return null;
 		}
 
 		private void ParseAsUnixAbsoluteFilePath (string uriString)
@@ -1248,10 +1338,26 @@ namespace System {
 				path = uriString;
 		}
 
-		// this parse method is as relaxed as possible about the format
-		// it will hardly ever throw a UriFormatException
+		//
+		// This parse method will throw exceptions on failure
+		//  
 		private void Parse (UriKind kind, string uriString)
 		{			
+			if (uriString == null)
+				throw new ArgumentNullException ("uriString");
+
+			string s = ParseNoExceptions (kind, uriString);
+			if (s != null)
+				throw new UriFormatException (s);
+		}
+
+		//
+		// This parse method will not throw exceptions on failure
+		//
+		// Returns null on success, or a description of the error in the parsing
+		//
+		private string ParseNoExceptions (UriKind kind, string uriString)
+		{
 			//
 			// From RFC 2396 :
 			//
@@ -1259,21 +1365,18 @@ namespace System {
 			//       12            3  4          5       6  7        8 9
 			//			
 			
-			if (uriString == null)
-				throw new ArgumentNullException ("uriString");
-
 			uriString = uriString.Trim();
 			int len = uriString.Length;
 
 			if (len == 0){
 				if (kind == UriKind.Relative || kind == UriKind.RelativeOrAbsolute){
 					isAbsoluteUri = false;
-					return;
+					return null;
 				}
 			}
 			
 			if (len <= 1 && (kind != UriKind.Relative))
-				throw new UriFormatException ();
+				return "Absolute URI is too short";
 
 			int pos = 0;
 
@@ -1281,13 +1384,17 @@ namespace System {
 			// Identify Windows path, unix path, or standard URI.
 			pos = uriString.IndexOf (':');
 			if (pos == 0) {
-				throw new UriFormatException("Invalid URI: The format of the URI could not be determined.");
+				return "Invalid URI: The format of the URI could not be determined.";
 			} else if (pos < 0) {
 				// It must be Unix file path or Windows UNC
 				if (uriString [0] == '/' && Path.DirectorySeparatorChar == '/'){
 					ParseAsUnixAbsoluteFilePath (uriString);
+#if NET_2_1 && !MONOTOUCH
+					isAbsoluteUri = false;
+#else
 					if (kind == UriKind.Relative)
 						isAbsoluteUri = false;
+#endif
 					
 				} else if (uriString.Length >= 2 && uriString [0] == '\\' && uriString [1] == '\\')
 					ParseAsWindowsUNC (uriString);
@@ -1296,13 +1403,15 @@ namespace System {
 					isAbsoluteUri = false;
 					path = uriString;
 				}
-				return;
+				return null;
 			} else if (pos == 1) {
 				if (!IsAlpha (uriString [0]))
-					throw new UriFormatException ("URI scheme must start with a letter.");
+					return "URI scheme must start with a letter.";
 				// This means 'a:' == windows full path.
-				ParseAsWindowsAbsoluteFilePath (uriString);
-				return;
+				string msg = ParseAsWindowsAbsoluteFilePath (uriString);
+				if (msg != null)
+					return msg;
+				return null;
 			} 
 
 			// scheme
@@ -1310,10 +1419,8 @@ namespace System {
 
 			// Check scheme name characters as specified in RFC2396.
 			// Note: different checks in 1.x and 2.0
-			if (!CheckSchemeName (scheme)) {
-				string msg = Locale.GetText ("URI scheme must start with a letter and must consist of one of alphabet, digits, '+', '-' or '.' character.");
-				throw new UriFormatException (msg);
-			}
+			if (!CheckSchemeName (scheme)) 
+				return Locale.GetText ("URI scheme must start with a letter and must consist of one of alphabet, digits, '+', '-' or '.' character.");
 
 			// from here we're practically working on uriString.Substring(startpos,endpos-startpos)
 			int startpos = pos + 1;
@@ -1343,14 +1450,15 @@ namespace System {
 			if (IsPredefinedScheme (scheme) && scheme != UriSchemeMailto && scheme != UriSchemeNews && (
 				(endpos-startpos < 2) ||
 				(endpos-startpos >= 2 && uriString [startpos] == '/' && uriString [startpos+1] != '/')))				
-				throw new UriFormatException ("Invalid URI: The Authority/Host could not be parsed.");
+				return "Invalid URI: The Authority/Host could not be parsed.";
 			
 			
 			bool startsWithSlashSlash = endpos-startpos >= 2 && uriString [startpos] == '/' && uriString [startpos+1] == '/';
 			bool unixAbsPath = scheme == UriSchemeFile && startsWithSlashSlash && (endpos-startpos == 2 || uriString [startpos+2] == '/');
+			bool windowsFilePath = false;
 			if (startsWithSlashSlash) {
 				if (kind == UriKind.Relative)
-					throw new UriFormatException ("Absolute URI when we expected a relative one");
+					return "Absolute URI when we expected a relative one";
 				
 				if (scheme != UriSchemeMailto && scheme != UriSchemeNews)
 					startpos += 2;
@@ -1372,19 +1480,26 @@ namespace System {
 					}
 				}
 				
-				if (endpos - startpos > 1 && uriString [startpos + 1] == ':')
+				if (endpos - startpos > 1 && uriString [startpos + 1] == ':') {
 					unixAbsPath = false;
+					windowsFilePath = true;
+				}
 
 			} else if (!IsPredefinedScheme (scheme)) {
 				path = uriString.Substring(startpos, endpos-startpos);
 				isOpaquePart = true;
-				return;
+				return null;
 			}
 
 			// 5 path
-			pos = uriString.IndexOf ('/', startpos, endpos-startpos);
-			if (unixAbsPath)
+			if (unixAbsPath) {
 				pos = -1;
+			} else {
+				pos = uriString.IndexOf ('/', startpos, endpos-startpos);
+				if (pos == -1 && windowsFilePath)
+					pos = uriString.IndexOf ('\\', startpos, endpos-startpos);
+			}
+
 			if (pos == -1) {
 				if ((scheme != Uri.UriSchemeMailto) &&
 #if ONLY_1_1
@@ -1398,7 +1513,10 @@ namespace System {
 			}
 
 			// 4.a user info
-			pos = uriString.IndexOf ('@', startpos, endpos-startpos);
+			if (unixAbsPath)
+				pos = -1;
+			else
+				pos = uriString.IndexOf ('@', startpos, endpos-startpos);
 			if (pos != -1) {
 				userinfo = uriString.Substring (startpos, pos-startpos);
 				startpos = pos + 1;
@@ -1406,22 +1524,26 @@ namespace System {
 
 			// 4.b port
 			port = -1;
-			pos = uriString.LastIndexOf (':', endpos-1, endpos-startpos);
 			if (unixAbsPath)
 				pos = -1;
+			else
+				pos = uriString.LastIndexOf (':', endpos-1, endpos-startpos);
 			if (pos != -1 && pos != endpos - 1) {
 				string portStr = uriString.Substring(pos + 1, endpos - (pos + 1));
 				if (portStr.Length > 0 && portStr[portStr.Length - 1] != ']') {
-					try {
 #if NET_2_0
-						port = (int) UInt16.Parse (portStr, NumberStyles.Integer, CultureInfo.InvariantCulture);
+					if (!Int32.TryParse (portStr, NumberStyles.Integer, CultureInfo.InvariantCulture, out port) ||
+					    port < 0 || port > UInt16.MaxValue)
+						return "Invalid URI: Invalid port number";
+					endpos = pos;
 #else
+					try {
 						port = (int) UInt32.Parse (portStr, CultureInfo.InvariantCulture);
-#endif
 						endpos = pos;
 					} catch (Exception) {
-						throw new UriFormatException ("Invalid URI: Invalid port number");
+						return "Invalid URI: Invalid port number";
 					}
+#endif
 				} else {
 					if (port == -1) {
 						port = GetDefaultPort (scheme);
@@ -1438,7 +1560,7 @@ namespace System {
 			host = uriString;
 
 			if (unixAbsPath) {
-				path = '/' + uriString;
+				path = Reduce ('/' + uriString);
 				host = String.Empty;
 			} else if (host.Length == 2 && host [1] == ':') {
 				// windows filepath
@@ -1456,35 +1578,78 @@ namespace System {
 					host = String.Empty;
 				}
 			} else if (host.Length == 0 &&
-				(scheme == UriSchemeHttp || scheme == UriSchemeGopher || scheme == UriSchemeNntp
-				 || scheme == UriSchemeHttps || scheme == UriSchemeFtp)) {
-				throw new UriFormatException ("Invalid URI: The hostname could not be parsed");
+				   (scheme == UriSchemeHttp || scheme == UriSchemeGopher || scheme == UriSchemeNntp ||
+				    scheme == UriSchemeHttps || scheme == UriSchemeFtp)) {
+				return "Invalid URI: The hostname could not be parsed";
 			}
 
 			bool badhost = ((host.Length > 0) && (CheckHostName (host) == UriHostNameType.Unknown));
 			if (!badhost && (host.Length > 1) && (host[0] == '[') && (host[host.Length - 1] == ']')) {
-				try {
-					host = "[" + IPv6Address.Parse (host).ToString (true) + "]";
-				}
-				catch (Exception) {
+				IPv6Address ipv6addr;
+				
+				if (IPv6Address.TryParse (host, out ipv6addr))
+					host = "[" + ipv6addr.ToString (true) + "]";
+				else
 					badhost = true;
-				}
-			}
-			if (badhost) {
-				string msg = Locale.GetText ("Invalid URI: The hostname could not be parsed.");
-				throw new UriFormatException (msg);
 			}
+#if NET_2_0
+			if (badhost && (Parser is DefaultUriParser || Parser == null))
+				return Locale.GetText ("Invalid URI: The hostname could not be parsed. (" + host + ")");
+
+			UriFormatException ex = null;
+			if (Parser != null)
+				Parser.InitializeAndValidate (this, out ex);
+			if (ex != null)
+				return ex.Message;
+#else
+			if (badhost)
+				return Locale.GetText ("Invalid URI: The hostname could not be parsed. (" + host + ")");
+#endif
 
 			if ((scheme != Uri.UriSchemeMailto) &&
 					(scheme != Uri.UriSchemeNews) &&
 					(scheme != Uri.UriSchemeFile)) {
 				path = Reduce (path);
 			}
+
+			return null;
 		}
 
 		private static string Reduce (string path)
 		{
-			path = path.Replace ('\\','/');
+			// quick out, allocation-free, for a common case
+			if (path == "/")
+				return path;
+
+			// replace '\', %5C ('\') and %2f ('/') into '/'
+			// other escaped values seems to survive this step
+			StringBuilder res = new StringBuilder();
+			for (int i=0; i < path.Length; i++) {
+				char c = path [i];
+				switch (c) {
+				case '\\':
+					res.Append ('/');
+					break;
+				case '%':
+					if (i < path.Length - 2) {
+						char c1 = path [i + 1];
+						char c2 = Char.ToUpper (path [i + 2]);
+						if (((c1 == '2') && (c2 == 'F')) || ((c1 == '5') && (c2 == 'C'))) {
+							res.Append ('/');
+							i += 2;
+						} else {
+							res.Append (c);
+						}
+					} else {
+						res.Append (c);
+					}
+					break;
+				default:
+					res.Append (c);
+					break;
+				}
+			}
+			path = res.ToString ();
 			ArrayList result = new ArrayList ();
 
 			for (int startpos = 0; startpos < path.Length; ) {
@@ -1523,7 +1688,7 @@ namespace System {
 			if (result.Count == 0)
 				return "/";
 
-			StringBuilder res = new StringBuilder();
+			res.Length = 0;
 			if (path [0] == '/')
 				res.Append ('/');
 
@@ -1729,6 +1894,18 @@ namespace System {
 			return false;
 		}
 
+		internal static bool MaybeUri (string s)
+		{
+			int p = s.IndexOf (':');
+			if (p == -1)
+				return false;
+
+			if (p >= 10)
+				return false;
+
+			return IsPredefinedScheme (s.Substring (0, p));
+		}
+		
 		private static bool IsPredefinedScheme (string scheme)
 		{
 			switch (scheme) {
@@ -1767,8 +1944,12 @@ namespace System {
 
 		private UriParser Parser {
 			get {
-				if (parser == null)
+				if (parser == null) {
 					parser = UriParser.GetParser (Scheme);
+					// no specific parser ? then use a default one
+					if (parser == null)
+						parser = new DefaultUriParser ("*");
+				}
 				return parser;
 			}
 			set { parser = value; }
@@ -1899,21 +2080,24 @@ namespace System {
 		{
 			if (uriString == null)
 				return false;
-			Uri uri = new Uri (uriString, uriKind);
-			return uri.IsWellFormedOriginalString ();
+
+			Uri uri;
+			if (Uri.TryCreate (uriString, uriKind, out uri))
+				return uri.IsWellFormedOriginalString ();
+			return false;
 		}
 
-		// [MonoTODO ("rework code to avoid exception catching")]
 		public static bool TryCreate (string uriString, UriKind uriKind, out Uri result)
 		{
-			try {
-				result = new Uri (uriString, uriKind);
+			bool success;
+
+			Uri r = new Uri (uriString, uriKind, out success);
+			if (success) {
+				result = r;
 				return true;
 			}
-			catch (UriFormatException) {
-				result = null;
-				return false;
-			}
+			result = null;
+			return false;
 		}
 
 		// [MonoTODO ("rework code to avoid exception catching")]
@@ -1923,8 +2107,7 @@ namespace System {
 				// FIXME: this should call UriParser.Resolve
 				result = new Uri (baseUri, relativeUri);
 				return true;
-			}
-			catch (UriFormatException) {
+			} catch (UriFormatException) {
 				result = null;
 				return false;
 			}
@@ -1935,10 +2118,9 @@ namespace System {
 		{
 			try {
 				// FIXME: this should call UriParser.Resolve
-				result = new Uri (baseUri, relativeUri);
+				result = new Uri (baseUri, relativeUri.OriginalString);
 				return true;
-			}
-			catch (UriFormatException) {
+			} catch (UriFormatException) {
 				result = null;
 				return false;
 			}
