@@ -1,4 +1,4 @@
-// JapaneseCalendar.cs
+// System.Globalization.JapaneseCalendar.cs
 //
 // (C) Ulrich Kunitz 2002
 //
@@ -29,6 +29,7 @@
 namespace System.Globalization {
 
 using System;
+using System.Runtime.InteropServices;
 
 /// <summary>
 /// This is the Japanese calendar. It differs from the Gregorian calendar
@@ -68,6 +69,9 @@ using System;
 /// </para>
 /// </remarks>
 [Serializable]
+#if NET_2_0
+[ComVisible (true)]
+#endif
 [MonoTODO ("Serialization format not compatible with .NET")]
 public class JapaneseCalendar : Calendar {
 	/// <summary>
@@ -238,226 +242,6 @@ public class JapaneseCalendar : Calendar {
 		M_ArgumentInRange("day", day, 1, GetDaysInMonth(year, month, era));
 		return gregorianYear;
 	}
-
-#if false
-
-	// Ifdefed out because this is not on the .NET Framewokr.
-	
-	/// <summary>
-	/// Overridden. Adds days to a given date.
-	/// </summary>
-	/// <param name="time">The
-	/// <see cref="T:System.DateTime"/> to which to add
-	/// days.
-	/// </param>
-	/// <param name="days">The number of days to add.</param>
-	/// <returns>A new <see cref="T:System.DateTime"/> value, that
-	/// results from adding <paramref name="days"/> to the specified
-	/// DateTime.</returns>
-	/// <exception cref="T:System.ArgumentOutOfRangeException">
-	/// The exception is thrown if the
-	/// <see cref="T:System.DateTime"/> return value is outside all
-	/// supported eras.
-	/// </exception>
-	public override DateTime AddDays(DateTime time, int days) {
-		DateTime t = base.AddDays(time, days);
-		M_CheckDateTime(t);
-		return t;
-	}
-
-	/// <summary>
-	/// Overridden. Adds hours to a given date.
-	/// </summary>
-	/// <param name="time">The
-	/// <see cref="T:System.DateTime"/> to which to add
-	/// hours.
-	/// </param>
-	/// <param name="hours">The number of hours to add.</param>
-	/// <returns>A new <see cref="T:System.DateTime"/> value, that
-	/// results from adding <paramref name="hours"/> to the specified
-	/// DateTime.</returns>
-	/// <exception cref="T:System.ArgumentOutOfRangeException">
-	/// The exception is thrown if the
-	/// <see cref="T:System.DateTime"/> return value is outside all
-	/// supported eras.
-	/// </exception>
-	public override DateTime AddHours(DateTime time, int hours) {
-		DateTime t = base.AddHours(time, hours);
-		M_CheckDateTime(t);
-		return t;
-	}
-
-	/// <summary>
-	/// Overridden. Adds milliseconds to a given date.
-	/// </summary>
-	/// <param name="time">The
-	/// <see cref="T:System.DateTime"/> to which to add
-	/// milliseconds.
-	/// </param>
-	/// <param name="milliseconds">The number of milliseconds given as
-	/// double to add. Keep in mind the 100 nanosecond resolution of 
-	/// <see cref="T:System.DateTime"/>.
-	/// </param>
-	/// <returns>A new <see cref="T:System.DateTime"/> value, that
-	/// results from adding <paramref name="milliseconds"/> to the specified
-	/// DateTime.</returns>
-	/// <exception cref="T:System.ArgumentOutOfRangeException">
-	/// The exception is thrown if the
-	/// <see cref="T:System.DateTime"/> return value is outside all
-	/// supported eras.
-	/// </exception>
-	public override DateTime AddMilliseconds(DateTime time,
-		double milliseconds)
-	{
-		DateTime t = base.AddMilliseconds(time, milliseconds);
-		M_CheckDateTime(t);
-		return t;
-	}
-
-	/// <summary>
-	/// Overridden. Adds minutes to a given date.
-	/// </summary>
-	/// <param name="time">The
-	/// <see cref="T:System.DateTime"/> to which to add
-	/// minutes.
-	/// </param>
-	/// <param name="minutes">The number of minutes to add.</param>
-	/// <returns>A new <see cref="T:System.DateTime"/> value, that
-	/// results from adding <paramref name="minutes"/> to the specified
-	/// DateTime.</returns>
-	/// <exception cref="T:System.ArgumentOutOfRangeException">
-	/// The exception is thrown if the
-	/// <see cref="T:System.DateTime"/> return value is outside all
-	/// supported eras.
-	/// </exception>
-	public override DateTime AddMinutes(DateTime time, int minutes) {
-		DateTime t = base.AddMinutes(time, minutes);
-		M_CheckDateTime(t);
-		return t;
-	}
-
-	/// <summary>
-	/// Overridden. Adds seconds to a given date.
-	/// </summary>
-	/// <param name="time">The
-	/// <see cref="T:System.DateTime"/> to which to add
-	/// seconds.
-	/// </param>
-	/// <param name="seconds">The number of seconds to add.</param>
-	/// <returns>A new <see cref="T:System.DateTime"/> value, that
-	/// results from adding <paramref name="seconds"/> to the specified
-	/// DateTime.</returns>
-	/// <exception cref="T:System.ArgumentOutOfRangeException">
-	/// The exception is thrown if the
-	/// <see cref="T:System.DateTime"/> return value is outside all
-	/// supported eras.
-	/// </exception>
-	public override DateTime AddSeconds(DateTime time, int seconds) {
-		DateTime t = base.AddSeconds(time, seconds);
-		M_CheckDateTime(t);
-		return t;
-	}
-
-
-	/// <summary>
-	/// Overridden. Adds weeks to a given date.
-	/// </summary>
-	/// <param name="time">The
-	/// <see cref="T:System.DateTime"/> to which to add
-	/// weeks.
-	/// </param>
-	/// <param name="weeks">The number of weeks to add.</param>
-	/// <returns>A new <see cref="T:System.DateTime"/> value, that
-	/// results from adding <paramref name="weeks"/> to the specified
-	/// DateTime.</returns>
-	/// <exception cref="T:System.ArgumentOutOfRangeException">
-	/// The exception is thrown if the
-	/// <see cref="T:System.DateTime"/> return value is outside all
-	/// supported eras.
-	/// </exception>
-	public override DateTime AddWeeks(DateTime time, int weeks) {
-		DateTime t = base.AddWeeks(time, weeks);
-		M_CheckDateTime(t);
-		return t;
-	}
-
-	/// <summary>
-	/// Overridden. Gives the hour of the specified time.
-	/// </summary>
-	/// <param name="time">The
-	/// <see cref="T:System.DateTime"/> that specifies the
-	/// time.
-	/// </param>
-	/// <returns>An integer that gives the hour of the specified time,
-	/// starting with 0.</returns>
-	/// <exception cref="T:System.ArgumentOutOfRangeException">
-	/// The exception is thrown if the
-	/// <see cref="T:System.DateTime"/> parameter is outside all
-	/// supported eras.
-	/// </exception>
-	public override int GetHour(DateTime time) {
-		M_CheckDateTime(time);
-		return base.GetHour(time);
-	}
-
-	/// <summary>
-	/// Overridden. Gives the milliseconds in the current second
-	/// of the specified time.
-	/// </summary>
-	/// <param name="time">The
-	/// <see cref="T:System.DateTime"/> that specifies the
-	/// time.
-	/// </param>
-	/// <returns>An integer that gives the milliseconds in the seconds
-	/// of the specified time, starting with 0.</returns>
-	/// <exception cref="T:System.ArgumentOutOfRangeException">
-	/// The exception is thrown if the
-	/// <see cref="T:System.DateTime"/> parameter is outside all
-	/// supported eras.
-	/// </exception>
-	public override double GetMilliseconds(DateTime time) {
-		M_CheckDateTime(time);
-		return base.GetMilliseconds(time);
-	}
-
-	/// <summary>
-	/// Overridden. Gives the minute of the specified time.
-	/// </summary>
-	/// <param name="time">The
-	/// <see cref="T:System.DateTime"/> that specifies the
-	/// time.
-	/// </param>
-	/// <returns>An integer that gives the minute of the specified time,
-	/// starting with 0.</returns>
-	/// <exception cref="T:System.ArgumentOutOfRangeException">
-	/// The exception is thrown if the
-	/// <see cref="T:System.DateTime"/> parameter is outside all
-	/// supported eras.
-	/// </exception>
-	public override int GetMinute(DateTime time) {
-		M_CheckDateTime(time);
-		return base.GetMinute(time);
-	}
-
-	/// <summary>
-	/// Overridden. Gives the second of the specified time.
-	/// </summary>
-	/// <param name="time">The
-	/// <see cref="T:System.DateTime"/> that specifies the
-	/// time.
-	/// </param>
-	/// <returns>An integer that gives the second of the specified time,
-	/// starting with 0.</returns>
-	/// <exception cref="T:System.ArgumentOutOfRangeException">
-	/// The exception is thrown if the
-	/// <see cref="T:System.DateTime"/> parameter is outside all
-	/// supported eras.
-	/// </exception>
-	public override int GetSecond(DateTime time) {
-		M_CheckDateTime(time);
-		return base.GetMinute(time);
-	}
-#endif
 	
 	/// <summary>
 	/// Overrideden. Adds months to a given date.
@@ -628,6 +412,14 @@ public class JapaneseCalendar : Calendar {
 		return era;
 	}
 
+#if NET_2_0
+	[ComVisible (false)]
+	public override int GetLeapMonth (int year, int era)
+	{
+		return 0;
+	}
+#endif
+
 	/// <summary>
 	/// Overridden. Gives the number of the month of the specified
 	/// date.
@@ -666,6 +458,14 @@ public class JapaneseCalendar : Calendar {
 		return 12;
 	}
 
+#if NET_2_0
+	[ComVisible (false)]
+	public override int GetWeekOfYear (DateTime time, CalendarWeekRule rule, DayOfWeek firstDayOfWeek)
+	{
+		return base.GetWeekOfYear (time, rule, firstDayOfWeek);
+	}
+#endif
+
 	/// <summary>
 	/// Overridden. Gives the number of the year of the specified
 	/// date.
@@ -789,14 +589,14 @@ public class JapaneseCalendar : Calendar {
 	/// is out of range.
 	/// </exception>
 	public override DateTime ToDateTime(int year, int month, int day,
-		int hour, int minute, int second, int milliseconds,
+		int hour, int minute, int second, int millisecond,
 		int era)
 	{
 		int gregorianYear = M_CheckYMDEG(year, month, day, ref era);
-		M_CheckHMSM(hour, minute, second, milliseconds);
+		M_CheckHMSM(hour, minute, second, millisecond);
 		return CCGregorianCalendar.ToDateTime(
 			gregorianYear, month, day,
-			hour, minute, second, milliseconds);
+			hour, minute, second, millisecond);
 	}
 
 
@@ -821,21 +621,27 @@ public class JapaneseCalendar : Calendar {
 	}
 
 #if NET_2_0
+
+#if !NET_2_1
+	[ComVisible (false)]
 	public override CalendarAlgorithmType AlgorithmType {
 		get {
 			return CalendarAlgorithmType.SolarCalendar;
 		}
 	}
+#endif
 
 	static DateTime JapanMin = new DateTime (1868, 9, 8, 0, 0, 0);
 	static DateTime JapanMax = new DateTime (9999, 12, 31, 11, 59, 59);
 		
+	[ComVisible (false)]
 	public override DateTime MinSupportedDateTime {
 		get {
 			return JapanMin;
 		}
 	}
 
+	[ComVisible (false)]
 	public override DateTime MaxSupportedDateTime {
 		get {
 			return JapanMax;
