@@ -39,10 +39,12 @@ using System.Threading;
 
 #if NET_2_0
 using Microsoft.Win32.SafeHandles;
-using System.Security.AccessControl;
-#endif
 #if NET_2_1
 using System.IO.IsolatedStorage;
+using System.Security;
+#else
+using System.Security.AccessControl;
+#endif
 #endif
 
 namespace System.IO
@@ -99,7 +101,13 @@ namespace System.IO
 			this.access = access;
 			this.owner = ownsHandle;
 			this.async = isAsync;
+#if NET_2_1 && !MONOTOUCH
+			// default the browser to 'all' anonymous files and let other usage (like smcs) with 'normal'
+			// (i.e. non-anonymous except for isolated storage) files and paths
+			this.anonymous = SecurityManager.SecurityEnabled;
+#else
 			this.anonymous = false;
+#endif
 
 			InitBuffer (bufferSize, noBuffering);
 
@@ -141,7 +149,7 @@ namespace System.IO
 		{
 		}
 
-#if NET_2_0
+#if NET_2_0 && !NET_2_1
 		public FileStream (string path, FileMode mode, FileAccess access, FileShare share, int bufferSize, FileOptions options)
 			: this (path, mode, access, share, bufferSize, false, options)
 		{
@@ -163,6 +171,7 @@ namespace System.IO
 				   int bufferSize, bool isAsync)
 			:this (handle.DangerousGetHandle (), access, false, bufferSize, isAsync)
 		{
+			this.safeHandle = handle;
 		}
 
 		public FileStream (string path, FileMode mode,
@@ -243,8 +252,7 @@ namespace System.IO
 			if (Directory.Exists (path)) {
 				// don't leak the path information for isolated storage
 				string msg = Locale.GetText ("Access to the path '{0}' is denied.");
-				string fname = (anonymous) ? Path.GetFileName (path) : Path.GetFullPath (path);
-				throw new UnauthorizedAccessException (String.Format (msg, fname));
+				throw new UnauthorizedAccessException (String.Format (msg, GetSecureFileName (path, false)));
 			}
 
 			/* Append streams can't be read (see FileMode
@@ -270,6 +278,7 @@ namespace System.IO
 					string msg = Locale.GetText ("Could not find a part of the path \"{0}\".");
 					string fname = (anonymous) ? dname : Path.GetFullPath (path);
 #if NET_2_1
+					// don't use GetSecureFileName for the directory name
 					throw new IsolatedStorageException (String.Format (msg, fname));
 #else
 					throw new DirectoryNotFoundException (String.Format (msg, fname));
@@ -281,7 +290,7 @@ namespace System.IO
 					mode != FileMode.CreateNew && !File.Exists (path)) {
 				// don't leak the path information for isolated storage
 				string msg = Locale.GetText ("Could not find file \"{0}\".");
-				string fname = (anonymous) ? Path.GetFileName (path) : Path.GetFullPath (path);
+				string fname = GetSecureFileName (path);
 #if NET_2_1
 				throw new IsolatedStorageException (String.Format (msg, fname));
 #else
@@ -300,8 +309,7 @@ namespace System.IO
 			this.handle = MonoIO.Open (path, mode, access, share, options, out error);
 			if (handle == MonoIO.InvalidHandle) {
 				// don't leak the path information for isolated storage
-				string fname = (anonymous) ? Path.GetFileName (path) : Path.GetFullPath (path);
-				throw MonoIO.GetException (fname, error);
+				throw MonoIO.GetException (GetSecureFileName (path), error);
 			}
 
 			this.access = access;
@@ -388,8 +396,7 @@ namespace System.IO
 				length = MonoIO.GetLength (handle, out error);
 				if (error != MonoIOError.ERROR_SUCCESS) {
 					// don't leak the path information for isolated storage
-					string fname = (anonymous) ? Path.GetFileName (name) : name;
-					throw MonoIO.GetException (fname, error);
+					throw MonoIO.GetException (GetSecureFileName (name), error);
 				}
 
 				return(length);
@@ -437,7 +444,17 @@ namespace System.IO
 		public virtual SafeFileHandle SafeFileHandle {
 			[SecurityPermission (SecurityAction.LinkDemand, UnmanagedCode = true)]
 			[SecurityPermission (SecurityAction.InheritanceDemand, UnmanagedCode = true)]
-			get { throw new NotImplementedException (); }
+			get {
+				SafeFileHandle ret;
+
+				if (safeHandle != null)
+					ret = safeHandle;
+				else
+					ret = new SafeFileHandle (handle, owner);
+
+				FlushBuffer ();
+				return ret;
+			}
 		}
 #endif
 
@@ -646,8 +663,7 @@ namespace System.IO
 				MonoIO.Write (handle, src, offset, count, out error);
 				if (error != MonoIOError.ERROR_SUCCESS) {
 					// don't leak the path information for isolated storage
-					string fname = (anonymous) ? Path.GetFileName (name) : name;
-					throw MonoIO.GetException (fname, error);
+					throw MonoIO.GetException (GetSecureFileName (name), error);
 				}
 				
 				buf_start += count;
@@ -703,7 +719,7 @@ namespace System.IO
 
 			if (buf_dirty) {
 				MemoryStream ms = new MemoryStream ();
-				FlushBufferToStream (ms);
+				FlushBuffer (ms);
 				ms.Write (array, offset, numBytes);
 				offset = 0;
 				numBytes = (int) ms.Length;
@@ -787,8 +803,7 @@ namespace System.IO
 
 			if (error != MonoIOError.ERROR_SUCCESS) {
 				// don't leak the path information for isolated storage
-				string fname = (anonymous) ? Path.GetFileName (name) : name;
-				throw MonoIO.GetException (fname, error);
+				throw MonoIO.GetException (GetSecureFileName (name), error);
 			}
 			
 			return(buf_start);
@@ -815,8 +830,7 @@ namespace System.IO
 			MonoIO.SetLength (handle, value, out error);
 			if (error != MonoIOError.ERROR_SUCCESS) {
 				// don't leak the path information for isolated storage
-				string fname = (anonymous) ? Path.GetFileName (name) : name;
-				throw MonoIO.GetException (fname, error);
+				throw MonoIO.GetException (GetSecureFileName (name), error);
 			}
 
 			if (Position > value)
@@ -863,8 +877,7 @@ namespace System.IO
 			MonoIO.Lock (handle, position, length, out error);
 			if (error != MonoIOError.ERROR_SUCCESS) {
 				// don't leak the path information for isolated storage
-				string fname = (anonymous) ? Path.GetFileName (name) : name;
-				throw MonoIO.GetException (fname, error);
+				throw MonoIO.GetException (GetSecureFileName (name), error);
 			}
 		}
 
@@ -884,8 +897,7 @@ namespace System.IO
 			MonoIO.Unlock (handle, position, length, out error);
 			if (error != MonoIOError.ERROR_SUCCESS) {
 				// don't leak the path information for isolated storage
-				string fname = (anonymous) ? Path.GetFileName (name) : name;
-				throw MonoIO.GetException (fname, error);
+				throw MonoIO.GetException (GetSecureFileName (name), error);
 			}
 		}
 
@@ -902,8 +914,13 @@ namespace System.IO
 		protected virtual void Dispose (bool disposing)
 #endif
 		{
+			Exception exc = null;
 			if (handle != MonoIO.InvalidHandle) {
-				FlushBuffer ();
+				try {
+					FlushBuffer ();
+				} catch (Exception e) {
+					exc = e;
+				}
 
 				if (owner) {
 					MonoIOError error;
@@ -911,8 +928,7 @@ namespace System.IO
 					MonoIO.Close (handle, out error);
 					if (error != MonoIOError.ERROR_SUCCESS) {
 						// don't leak the path information for isolated storage
-						string fname = (anonymous) ? Path.GetFileName (name) : name;
-						throw MonoIO.GetException (fname, error);
+						throw MonoIO.GetException (GetSecureFileName (name), error);
 					}
 
 					handle = MonoIO.InvalidHandle;
@@ -926,9 +942,11 @@ namespace System.IO
 			}
 			if (disposing)
 				GC.SuppressFinalize (this);
+ 			if (exc != null)
+ 				throw exc;
 		}
 
-#if NET_2_0
+#if NET_2_0 && !NET_2_1
 		public FileSecurity GetAccessControl ()
 		{
 			throw new NotImplementedException ();
@@ -985,21 +1003,31 @@ namespace System.IO
 			return(count);
 		}
 
-		void FlushBufferToStream (Stream st)
+		void FlushBuffer (Stream st)
 		{
 			if (buf_dirty) {
+				MonoIOError error;
+
 				if (CanSeek == true) {
-					MonoIOError error;
 					MonoIO.Seek (handle, buf_start,
 						     SeekOrigin.Begin,
 						     out error);
 					if (error != MonoIOError.ERROR_SUCCESS) {
 						// don't leak the path information for isolated storage
-						string fname = (anonymous) ? Path.GetFileName (name) : name;
-						throw MonoIO.GetException (fname, error);
+						throw MonoIO.GetException (GetSecureFileName (name), error);
 					}
 				}
-				st.Write (buf, 0, buf_length);
+				if (st == null) {
+					MonoIO.Write (handle, buf, 0,
+						      buf_length, out error);
+
+					if (error != MonoIOError.ERROR_SUCCESS) {
+						// don't leak the path information for isolated storage
+						throw MonoIO.GetException (GetSecureFileName (name), error);
+					}
+				} else {
+					st.Write (buf, 0, buf_length);
+				}
 			}
 
 			buf_start += buf_offset;
@@ -1009,43 +1037,18 @@ namespace System.IO
 
 		private void FlushBuffer ()
 		{
-			if (buf_dirty) {
-				MonoIOError error;
-				
-				if (CanSeek == true) {
-					MonoIO.Seek (handle, buf_start,
-						     SeekOrigin.Begin,
-						     out error);
-					if (error != MonoIOError.ERROR_SUCCESS) {
-						// don't leak the path information for isolated storage
-						string fname = (anonymous) ? Path.GetFileName (name) : name;
-						throw MonoIO.GetException (fname, error);
-					}
-				}
-				MonoIO.Write (handle, buf, 0,
-					      buf_length, out error);
-
-				if (error != MonoIOError.ERROR_SUCCESS) {
-					// don't leak the path information for isolated storage
-					string fname = (anonymous) ? Path.GetFileName (name) : name;
-					throw MonoIO.GetException (fname, error);
-				}
-			}
-
-			buf_start += buf_offset;
-			buf_offset = buf_length = 0;
-			buf_dirty = false;
+			FlushBuffer (null);
 		}
 
 		private void FlushBufferIfDirty ()
 		{
 			if (buf_dirty)
-				FlushBuffer ();
+				FlushBuffer (null);
 		}
 
 		private void RefillBuffer ()
 		{
-			FlushBuffer();
+			FlushBuffer (null);
 			
 			buf_length = ReadData (handle, buf, 0,
 					       buf_size);
@@ -1064,8 +1067,7 @@ namespace System.IO
 				amount = 0; // might not be needed, but well...
 			} else if (error != MonoIOError.ERROR_SUCCESS) {
 				// don't leak the path information for isolated storage
-				string fname = (anonymous) ? Path.GetFileName (name) : name;
-				throw MonoIO.GetException (fname, error);
+				throw MonoIO.GetException (GetSecureFileName (name), error);
 			}
 			
 			/* Check for read error */
@@ -1098,6 +1100,17 @@ namespace System.IO
 			buf_dirty = false;
 		}
 
+		private string GetSecureFileName (string filename)
+		{
+			return (anonymous) ? Path.GetFileName (filename) : Path.GetFullPath (filename);
+		}
+
+		private string GetSecureFileName (string filename, bool full)
+		{
+			return (anonymous) ? Path.GetFileName (filename) : 
+				(full) ? Path.GetFullPath (filename) : filename;
+		}
+
 		// fields
 
 		internal const int DefaultBufferSize = 8192;
@@ -1118,5 +1131,9 @@ namespace System.IO
 		private string name = "[Unknown]";	// name of file.
 
 		IntPtr handle;				// handle to underlying file
+#if NET_2_0
+		SafeFileHandle safeHandle;              // set only when using one of the
+							// constructors taking SafeFileHandle
+#endif
 	}
 }
