@@ -47,20 +47,23 @@ namespace System.Reflection.Emit {
 #endif
 	[ClassInterface (ClassInterfaceType.None)]
 	public sealed class ConstructorBuilder : ConstructorInfo, _ConstructorBuilder {
+	
+#pragma warning disable 169, 414
 		private RuntimeMethodHandle mhandle;
 		private ILGenerator ilgen;
-		private Type[] parameters;
+		internal Type[] parameters;
 		private MethodAttributes attrs;
 		private MethodImplAttributes iattrs;
 		private int table_idx;
 		private CallingConventions call_conv;
 		private TypeBuilder type;
-		private ParameterBuilder[] pinfo;
+		internal ParameterBuilder[] pinfo;
 		private CustomAttributeBuilder[] cattrs;
 		private bool init_locals = true;
 		private Type[][] paramModReq;
 		private Type[][] paramModOpt;
 		private RefEmitPermissionSet[] permissions;
+#pragma warning restore 169, 414
 
 		internal ConstructorBuilder (TypeBuilder tb, MethodAttributes attributes, CallingConventions callingConvention, Type[] parameterTypes, Type[][] paramModReq, Type[][] paramModOpt)
 		{
@@ -193,6 +196,7 @@ namespace System.Reflection.Emit {
 
 		public void AddDeclarativeSecurity (SecurityAction action, PermissionSet pset)
 		{
+#if !NET_2_1
 			if (pset == null)
 				throw new ArgumentNullException ("pset");
 			if ((action == SecurityAction.RequestMinimum) ||
@@ -217,6 +221,7 @@ namespace System.Reflection.Emit {
 
 			permissions [permissions.Length - 1] = new RefEmitPermissionSet (action, pset.ToXml ().ToString ());
 			attrs |= MethodAttributes.HasSecurity;
+#endif
 		}
 
 		public ParameterBuilder DefineParameter (int iSequence, ParameterAttributes attributes, string strParamName)
