@@ -53,18 +53,28 @@ using System.Configuration.Assemblies;
 #if NET_2_0
 using System.Collections.Generic;
 using System.Runtime.ConstrainedExecution;
+using System.Text;
 #endif
 
 namespace System {
 
 #if NET_2_0
 	[ComVisible (true)]
+#if !NET_2_1
 	[ComDefaultInterface (typeof (_AppDomain))]
+#endif
 #endif
 	[ClassInterface(ClassInterfaceType.None)]
-	public sealed class AppDomain : MarshalByRefObject , _AppDomain , IEvidenceFactory
-	{
+#if NET_2_1
+	public sealed class AppDomain : MarshalByRefObject {
+#else
+	public sealed class AppDomain : MarshalByRefObject, _AppDomain, IEvidenceFactory {
+#endif
+        #pragma warning disable 169
+        #region Sync with object-internals.h
 		IntPtr _mono_app_domain;
+		#endregion
+        #pragma warning restore 169
 		static string _process_guid;
 
 		[ThreadStatic]
@@ -73,6 +83,9 @@ namespace System {
 		[ThreadStatic]
 		static Hashtable assembly_resolve_in_progress;
 
+		[ThreadStatic]
+		static Hashtable assembly_resolve_in_progress_refonly;
+#if !NET_2_1 || MONOTOUCH
 		// CAS
 		private Evidence _evidence;
 		private PermissionSet _granted;
@@ -82,7 +95,7 @@ namespace System {
 
 		[ThreadStatic]
 		private static IPrincipal _principal;
-		
+#endif
 		static AppDomain default_domain;
 
 		private AppDomain ()
@@ -92,6 +105,9 @@ namespace System {
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private extern AppDomainSetup getSetup ();
 
+#if NET_2_1
+		internal
+#endif
 		AppDomainSetup SetupInformationNoCopy {
 			get { return getSetup (); }
 		}
@@ -103,13 +119,13 @@ namespace System {
 			}
 		}
 
-#if NET_2_0
+#if NET_2_0 && !NET_2_1
 		[MonoTODO]
 		public ApplicationTrust ApplicationTrust {
 			get { throw new NotImplementedException (); }
 		}
 #endif
-
+#if !NET_2_1 || MONOTOUCH
 		public string BaseDirectory {
 			get {
 				string path = SetupInformationNoCopy.ApplicationBase;
@@ -152,6 +168,7 @@ namespace System {
 				return (SetupInformationNoCopy.ShadowCopyFiles == "true");
 			}
 		}
+#endif
 
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private extern string getFriendlyName ();
@@ -161,7 +178,7 @@ namespace System {
 				return getFriendlyName ();
 			}
 		}
-
+#if !NET_2_1 || MONOTOUCH
 		public Evidence Evidence {
 			get {
 				// if the host (runtime) hasn't provided it's own evidence...
@@ -209,7 +226,7 @@ namespace System {
 		internal PermissionSet GrantedPermissionSet {
 			get { return _granted; }
 		}
-
+#endif
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private static extern AppDomain getCurDomain ();
 		
@@ -235,6 +252,8 @@ namespace System {
 			}
 		}
 
+#if !NET_2_1 || MONOTOUCH
+
 #if NET_2_0
 		[Obsolete ("AppDomain.AppendPrivatePath has been deprecated. Please investigate the use of AppDomainSetup.PrivateBinPath instead.")]
 #endif
@@ -277,6 +296,7 @@ namespace System {
 			SetupInformationNoCopy.ShadowCopyDirectories = String.Empty;
 		}
 
+#if !NET_2_1
 		public ObjectHandle CreateComInstanceFrom (string assemblyName, string typeName)
 		{
 			return Activator.CreateComInstanceFrom (assemblyName, typeName);
@@ -288,6 +308,7 @@ namespace System {
 		{
 			return Activator.CreateComInstanceFrom (assemblyFile, typeName, hashValue ,hashAlgorithm);
 		}
+#endif
 #endif
 
 		public ObjectHandle CreateInstance (string assemblyName, string typeName)
@@ -388,6 +409,8 @@ namespace System {
 			return (oh != null) ? oh.Unwrap () : null;
 		}
 
+#endif // !NET_2_1
+
 		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access)
 		{
 			return DefineDynamicAssembly (name, access, null, null, null, null, null, false);
@@ -485,14 +508,6 @@ namespace System {
 		}
 #endif
 
-#if NET_2_1
-		// TODO: the last parameter is ignored for now
-		public AssemblyBuilder DefineDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access, bool emitSymbolInfo)
-		{
-			return DefineDynamicAssembly (name, access, null, null, null, null, null, false);
-		}
-#endif
-
 		internal AssemblyBuilder DefineInternalDynamicAssembly (AssemblyName name, AssemblyBuilderAccess access)
 		{
 			return new AssemblyBuilder (name, null, access, true);
@@ -573,13 +588,13 @@ namespace System {
 			return Load (assemblyRef, null);
 		}
 
-		internal Assembly LoadSatellite (AssemblyName assemblyRef)
+		internal Assembly LoadSatellite (AssemblyName assemblyRef, bool throwOnError)
 		{
 			if (assemblyRef == null)
 				throw new ArgumentNullException ("assemblyRef");
 
 			Assembly result = LoadAssembly (assemblyRef.FullName, null, false);
-			if (result == null)
+			if (result == null && throwOnError)
 				throw new FileNotFoundException (null, assemblyRef.Name);
 			return result;
 		}
@@ -686,7 +701,7 @@ namespace System {
 			assembly.FromByteArray = true;
 			return assembly;
 		}
-
+#if !NET_2_1 || MONOTOUCH
 		[SecurityPermission (SecurityAction.Demand, ControlPolicy = true)]
 		public void SetAppDomainPolicy (PolicyLevel domainPolicy)
 		{
@@ -752,7 +767,7 @@ namespace System {
 
 			_principal = principal;
 		}
-
+#endif
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private static extern AppDomain InternalSetDomainByID (int domain_id);
  
@@ -841,6 +856,8 @@ namespace System {
 			return _process_guid;
 		}
 
+#if !NET_2_1 || MONOTOUCH
+
 		public static AppDomain CreateDomain (string friendlyName)
 		{
 			return CreateDomain (friendlyName, null, null);
@@ -854,7 +871,7 @@ namespace System {
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private static extern AppDomain createDomain (string friendlyName, AppDomainSetup info);
 
-		[MonoTODO ("Currently it does not allow the setup in the other domain")]
+		[MonoLimitationAttribute ("Currently it does not allow the setup in the other domain")]
 		[SecurityPermission (SecurityAction.Demand, ControlAppDomain = true)]
 		public static AppDomain CreateDomain (string friendlyName, Evidence securityInfo, AppDomainSetup info)
 		{
@@ -895,7 +912,7 @@ namespace System {
 			else
 				ad._evidence = new Evidence (securityInfo);	// copy
 
-#if NET_2_0
+#if NET_2_0 && !NET_2_1
 			if (info.AppDomainInitializer != null) {
 				if (!info.AppDomainInitializer.Method.IsStatic)
 					throw new ArgumentException ("Non-static methods cannot be invoked as an appdomain initializer");
@@ -914,7 +931,7 @@ namespace System {
 			return ad;
 		}
 
-#if NET_2_0
+#if NET_2_0 && !NET_2_1
 		[Serializable]
 		class Loader {
 
@@ -955,6 +972,18 @@ namespace System {
 		{
 			return CreateDomain (friendlyName, securityInfo, CreateDomainSetup (appBasePath, appRelativeSearchPath, shadowCopyFiles));
 		}
+		
+#if NET_2_0 && !NET_2_1
+		public static AppDomain CreateDomain (string friendlyName, Evidence securityInfo, AppDomainSetup info,
+		                                      PermissionSet grantSet, params StrongName [] fullTrustAssemblies)
+		{
+			if (info == null)
+				throw new ArgumentNullException ("info");
+
+			info.ApplicationTrust = new ApplicationTrust (grantSet, fullTrustAssemblies ?? new StrongName [0]);
+			return CreateDomain (friendlyName, securityInfo, info);		
+		}
+#endif
 
 		static AppDomainSetup CreateDomainSetup (string appBasePath, string appRelativeSearchPath, bool shadowCopyFiles)
 		{
@@ -972,17 +1001,10 @@ namespace System {
 				info.ShadowCopyFiles = null;
 #endif
 
+
 			return info;
 		}
-
-#if NET_2_0
-		[MonoTODO]
-		public static AppDomain CreateDomain (string friendlyName, Evidence securityInfo, AppDomainSetup info,
-		                                      PermissionSet grantSet, params StrongName [] fullTrustAssemblies)
-		{
-			throw new NotImplementedException ();
-		}
-#endif
+#endif // !NET_2_1
 
 		[MethodImplAttribute (MethodImplOptions.InternalCall)]
 		private static extern bool InternalIsFinalizingForUnload (int domain_id);
@@ -1026,6 +1048,7 @@ namespace System {
 		}
 #endif
 
+#if !NET_2_1
 #if NET_2_0
 		[Obsolete ("Use AppDomainSetup.DynamicBase")]
 #endif
@@ -1034,6 +1057,7 @@ namespace System {
 		{
 			SetupInformationNoCopy.DynamicBase = path;
 		}
+#endif // !NET_2_1
 
 #if NET_2_0
 		[Obsolete ("AppDomain.GetCurrentThreadId has been deprecated"
@@ -1049,7 +1073,14 @@ namespace System {
 
 		public override string ToString ()
 		{
+#if !NET_2_1 || MONOTOUCH
 			return getFriendlyName ();
+#else
+			StringBuilder sb = new StringBuilder ("Name:");
+			sb.AppendLine (FriendlyName);
+			sb.AppendLine ("There are no context policies.");
+			return sb.ToString ();
+#endif
 		}
 
 		private static void ValidateAssemblyName (string name)
@@ -1097,7 +1128,7 @@ namespace System {
 		private Assembly DoAssemblyResolve (string name, bool refonly)
 		{
 			ResolveEventHandler del;
-#if NET_2_0
+#if NET_2_0 && !NET_2_1
 			if (refonly)
 				del = ReflectionOnlyAssemblyResolve;
 			else
@@ -1109,20 +1140,24 @@ namespace System {
 				return null;
 			
 			/* Prevent infinite recursion */
-			Hashtable ht = assembly_resolve_in_progress;
-			if (ht == null) {
-				ht = new Hashtable ();
-				assembly_resolve_in_progress = ht;
+			Hashtable ht;
+			if (refonly) {
+				ht = assembly_resolve_in_progress_refonly;
+				if (ht == null) {
+					ht = new Hashtable ();
+					assembly_resolve_in_progress_refonly = ht;
+				}
+			} else {
+				ht = assembly_resolve_in_progress;
+				if (ht == null) {
+					ht = new Hashtable ();
+					assembly_resolve_in_progress = ht;
+				}
 			}
 
-			Assembly ass = (Assembly) ht [name];
-#if NET_2_0
-			if (ass != null && (ass.ReflectionOnly == refonly))
+			string s = (string) ht [name];
+			if (s != null)
 				return null;
-#else
-			if (ass != null)
-				return null;
-#endif
 			ht [name] = name;
 			try {
 				Delegate[] invocation_list = del.GetInvocationList ();
@@ -1184,12 +1219,13 @@ namespace System {
 				DomainUnload(this, null);
 		}
 
+#if !NET_2_1
 		internal byte[] GetMarshalledDomainObjRef ()
 		{
 			ObjRef oref = RemotingServices.Marshal (AppDomain.CurrentDomain, null, typeof (AppDomain));
 			return CADSerializer.SerializeObject (oref).GetBuffer();
 		}
-
+#endif
 		internal void ProcessMessageInDomain (byte[] arrRequest, CADMethodCallMessage cadMsg,
 		                                      out byte[] arrResponse, out CADMethodReturnMessage cadMrm)
 		{
@@ -1209,6 +1245,7 @@ namespace System {
 			else
 				arrResponse = null;
 		}
+
 #pragma warning restore 169
 
 		// End of methods called from the runtime
@@ -1253,13 +1290,32 @@ namespace System {
 		public event UnhandledExceptionEventHandler UnhandledException;
 #endif
 
+#if NET_4_0 || BOOTSTRAP_NET_4_0
+		[MonoTODO]
+		public bool IsHomogenous {
+			get { return true; }
+		}
+#endif
+
 #if NET_2_0
+        #pragma warning disable 649
+		private AppDomainManager _domain_manager;
+        #pragma warning restore 649
+
+		// default is null
+		public AppDomainManager DomainManager {
+			get { return _domain_manager; }
+		}
+#endif
+
+#if NET_2_0 && (!NET_2_1 || MONOTOUCH)
 
 		public event ResolveEventHandler ReflectionOnlyAssemblyResolve;
-		
+
+        #pragma warning disable 649
 		private ActivationContext _activation;
 		private ApplicationIdentity _applicationIdentity;
-		private AppDomainManager _domain_manager;
+        #pragma warning restore 649
 
 		// properties
 
@@ -1271,11 +1327,6 @@ namespace System {
 			get { return _applicationIdentity; }
 		}
 
-		// default is null
-		public AppDomainManager DomainManager {
-			get { return _domain_manager; }
-		}
-
 		public int Id {
 			[ReliabilityContract (Consistency.WillNotCorruptState, Cer.Success)]
 			get { return getDomainID (); }
@@ -1342,7 +1393,15 @@ namespace System {
 		}
 #endif
 
-#if NET_1_1
+#if NET_2_1 && !MONOTOUCH
+		public int ExecuteAssemblyByName (string assemblyName)
+		{
+			// critical code in SL that we're not calling in ML
+			throw new NotImplementedException ();
+		}
+#endif
+
+#if NET_1_1 && !NET_2_1
 		void _AppDomain.GetIDsOfNames ([In] ref Guid riid, IntPtr rgszNames, uint cNames, uint lcid, IntPtr rgDispId)
 		{
 			throw new NotImplementedException ();
