@@ -49,11 +49,39 @@ namespace System
 		, IComparable<DateTime>, IEquatable <DateTime>
 #endif
 	{
+#if MONOTOUCH
+		static DateTime () {
+			if (MonoTouchAOTHelper.FalseFlag) {
+				var comparer = new System.Collections.Generic.GenericComparer <DateTime> ();
+				var eqcomparer = new System.Collections.Generic.GenericEqualityComparer <DateTime> ();
+			}
+		}
+#endif
 		private TimeSpan ticks;
 
 #if NET_2_0
 		DateTimeKind kind;
 #endif
+#if !NET_2_0
+		internal struct DateTimeOffset {
+			public DateTime DateTime;
+			public TimeSpan Offset;
+
+			public DateTimeOffset (DateTime dt) : this (dt, TimeSpan.Zero)
+			{
+			}
+
+			public DateTimeOffset (long ticks, TimeSpan offset) : this (new DateTime (ticks), offset)
+			{
+			}
+
+			public DateTimeOffset (DateTime dt, TimeSpan offset)
+			{
+				DateTime = dt;
+				Offset = offset;
+			}
+		}
+#endif
 
 		private const int dp400 = 146097;
 		private const int dp100 = 36524;
@@ -88,6 +116,7 @@ namespace System
 		private static readonly string[] ParseTimeFormats = new string [] {
 			"H:m:s.fffffffzzz",
 			"H:m:s.fffffff",
+			"H:m:s tt zzz",
 			"H:m:szzz",
 			"H:m:s",
 			"H:mzzz",
@@ -579,7 +608,7 @@ namespace System
 					(value * TimeSpan.TicksPerMillisecond) < long.MinValue) {
 				throw new ArgumentOutOfRangeException();
 			}
-			long msticks = (long) (value * TimeSpan.TicksPerMillisecond);
+			long msticks = (long) Math.Round (value * TimeSpan.TicksPerMillisecond);
 
 			return AddTicks (msticks);
 		}
@@ -885,77 +914,115 @@ namespace System
 
 		public static DateTime Parse (string s, IFormatProvider provider, DateTimeStyles styles)
 		{
-			
-			const string formatExceptionMessage = "String was not recognized as a valid DateTime.";
-#if !NET_2_0
-			const string argumentYearRangeExceptionMessage = "Valid values are between 1 and 9999, inclusive.";
-#endif
-			
 			if (s == null)
 				throw new ArgumentNullException ("s");
+
+			DateTime res;
+			DateTimeOffset dto;
+			Exception exception = null;
+			if (!CoreParse (s, provider, styles, out res, out dto, true, ref exception))
+				throw exception;
+			
+			return res;
+		}
+
+		const string formatExceptionMessage = "String was not recognized as a valid DateTime.";
+		
+		internal static bool CoreParse (string s, IFormatProvider provider, DateTimeStyles styles,
+					      out DateTime result, out DateTimeOffset dto, bool setExceptionOnError, ref Exception exception)
+		{
+			dto = new DateTimeOffset (0, TimeSpan.Zero);
+			if (s == null || s.Length == 0) {
+				if (setExceptionOnError)
+					exception = new FormatException (formatExceptionMessage);
+				result = MinValue;
+				return false;
+			}
+
 			if (provider == null)
 				provider = CultureInfo.CurrentCulture;
 			DateTimeFormatInfo dfi = DateTimeFormatInfo.GetInstance (provider);
 
-			bool longYear = false;
-			DateTime result;
 			// Try first all the combinations of ParseAllDateFormats & ParseTimeFormats
-			string[] allDateFormats = YearMonthDayFormats (dfi);
+			string[] allDateFormats = YearMonthDayFormats (dfi, setExceptionOnError, ref exception);
+			if (allDateFormats == null){
+				result = MinValue;
+				return false;
+			}
+
+			bool longYear = false;
 			for (int i = 0; i < allDateFormats.Length; i++) {
 				string firstPart = allDateFormats [i];
 				bool incompleteFormat = false;
-				if (_DoParse (s, firstPart, "", false, out result, dfi, styles, true, ref incompleteFormat, ref longYear))
-					return result;
+				if (_DoParse (s, firstPart, "", false, out result, out dto, dfi, styles, true, ref incompleteFormat, ref longYear))
+					return true;
+
 				if (!incompleteFormat)
 					continue;
 
 				for (int j = 0; j < ParseTimeFormats.Length; j++) {
-					if (_DoParse (s, firstPart, ParseTimeFormats [j], false, out result, dfi, styles, true, ref incompleteFormat, ref longYear))
-						return result;
+					if (_DoParse (s, firstPart, ParseTimeFormats [j], false, out result, out dto, dfi, styles, true, ref incompleteFormat, ref longYear))
+						return true;
 				}
 			}
-			string[] monthDayFormats = IsDayBeforeMonth (dfi) ? DayMonthShortFormats : MonthDayShortFormats;
+
+			//
+			// Month day formats
+			//
+			int dayIndex = dfi.MonthDayPattern.IndexOf('d');
+			int monthIndex = dfi.MonthDayPattern.IndexOf('M');
+			if (dayIndex == -1 || monthIndex == -1){
+				result = MinValue;
+				if (setExceptionOnError)
+					exception = new FormatException (Locale.GetText("Order of month and date is not defined by {0}", dfi.MonthDayPattern));
+				return false;
+			}
+			bool is_day_before_month = dayIndex < monthIndex;
+			string[] monthDayFormats = is_day_before_month ? DayMonthShortFormats : MonthDayShortFormats;
 			for (int i = 0; i < monthDayFormats.Length; i++) {
 				bool incompleteFormat = false;
-				if (_DoParse (s, monthDayFormats[i], "", false, out result, dfi, styles, true, ref incompleteFormat, ref longYear))
-					return result;
+				if (_DoParse (s, monthDayFormats[i], "", false, out result, out dto, dfi, styles, true, ref incompleteFormat, ref longYear))
+					return true;
 			}
+			
 			for (int j = 0; j < ParseTimeFormats.Length; j++) {
 				string firstPart = ParseTimeFormats [j];
 				bool incompleteFormat = false;
-				if (_DoParse (s, firstPart, "", false, out result, dfi, styles, false, ref incompleteFormat, ref longYear))
-					return result;
+				if (_DoParse (s, firstPart, "", false, out result, out dto, dfi, styles, false, ref incompleteFormat, ref longYear))
+					return true;
 				if (!incompleteFormat)
 					continue;
 
 				for (int i = 0; i < monthDayFormats.Length; i++) {
-					if (_DoParse (s, firstPart, monthDayFormats [i], false, out result, dfi, styles, false, ref incompleteFormat, ref longYear))
-						return result;
+					if (_DoParse (s, firstPart, monthDayFormats [i], false, out result, out dto, dfi, styles, false, ref incompleteFormat, ref longYear))
+						return true;
 				}
 				for (int i = 0; i < allDateFormats.Length; i++) {
 					string dateFormat = allDateFormats [i];
 					if (dateFormat[dateFormat.Length - 1] == 'T')
 						continue; // T formats must be before the time part
-					if (_DoParse (s, firstPart, dateFormat, false, out result, dfi, styles, false, ref incompleteFormat, ref longYear))
-						return result;
+					if (_DoParse (s, firstPart, dateFormat, false, out result, out dto, dfi, styles, false, ref incompleteFormat, ref longYear))
+						return true;
 				}
 			}
 
 			// Try as a last resort all the patterns
-			if (ParseExact (s, dfi.GetAllDateTimePatternsInternal (), dfi, styles, out result, false, ref longYear))
-				return result;
+			if (ParseExact (s, dfi.GetAllDateTimePatternsInternal (), dfi, styles, out result, false, ref longYear, setExceptionOnError, ref exception))
+				return true;
 
+			if (!setExceptionOnError)
+				return false;
+			
 #if NET_2_0
-			// .NET does not throw an ArgumentOutOfRangeException, but .NET 1.1 does.
-			throw new FormatException (formatExceptionMessage);
+			// .NET 2.x does not throw an ArgumentOutOfRangeException, but .NET 1.1 does.
+			exception = new FormatException (formatExceptionMessage);
 #else
-			if (longYear) {
-				throw new ArgumentOutOfRangeException ("year",
-					argumentYearRangeExceptionMessage);
-			}
-
-			throw new FormatException (formatExceptionMessage);
+			if (longYear)
+				exception = new ArgumentOutOfRangeException ("year", "Valid values are between 1 and 9999, inclusive.");
+			else 
+				exception = new FormatException (formatExceptionMessage);
 #endif
+			return false;
 		}
 
 		public static DateTime ParseExact (string s, string format, IFormatProvider provider)
@@ -963,39 +1030,38 @@ namespace System
 			return ParseExact (s, format, provider, DateTimeStyles.None);
 		}
 
-		private static bool IsDayBeforeMonth (DateTimeFormatInfo dfi)
-		{
-			int dayIndex = dfi.MonthDayPattern.IndexOf('d');
-			int monthIndex = dfi.MonthDayPattern.IndexOf('M');
-			if (dayIndex == -1 || monthIndex == -1)
-				throw new FormatException (Locale.GetText("Order of month and date is not defined by {0}", dfi.MonthDayPattern));
-
-			return dayIndex < monthIndex;
-		}
-
-		private static string[] YearMonthDayFormats (DateTimeFormatInfo dfi)
+		private static string[] YearMonthDayFormats (DateTimeFormatInfo dfi, bool setExceptionOnError, ref Exception exc)
 		{
 			int dayIndex = dfi.ShortDatePattern.IndexOf('d');
 			int monthIndex = dfi.ShortDatePattern.IndexOf('M');
 			int yearIndex = dfi.ShortDatePattern.IndexOf('y');
-			if (dayIndex == -1 || monthIndex == -1 || yearIndex == -1)
-				throw new FormatException (Locale.GetText("Order of year, month and date is not defined by {0}", dfi.ShortDatePattern));
+			if (dayIndex == -1 || monthIndex == -1 || yearIndex == -1){
+				if (setExceptionOnError)
+					exc = new FormatException (Locale.GetText("Order of year, month and date is not defined by {0}", dfi.ShortDatePattern));
+				return null;
+			}
 
 			if (yearIndex < monthIndex)
 				if (monthIndex < dayIndex)
 					return ParseYearMonthDayFormats;
 				else if (yearIndex < dayIndex)
 					return ParseYearDayMonthFormats;
-				else
+				else {
 					// The year cannot be between the date and the month
-					throw new FormatException (Locale.GetText("Order of date, year and month defined by {0} is not supported", dfi.ShortDatePattern));
+					if (setExceptionOnError)
+						exc = new FormatException (Locale.GetText("Order of date, year and month defined by {0} is not supported", dfi.ShortDatePattern));
+					return null;
+				}
 			else if (dayIndex < monthIndex)
 				return ParseDayMonthYearFormats;
 			else if (dayIndex < yearIndex)
 				return ParseMonthDayYearFormats;
-			else
+			else {
 				// The year cannot be between the month and the date
-				throw new FormatException (Locale.GetText("Order of month, year and date defined by {0} is not supported", dfi.ShortDatePattern));
+				if (setExceptionOnError)
+					exc = new FormatException (Locale.GetText("Order of month, year and date defined by {0} is not supported", dfi.ShortDatePattern));
+				return null;
+			}
 		}
 
 		private static int _ParseNumber (string s, int valuePos,
@@ -1161,6 +1227,7 @@ namespace System
 					      string secondPart,
 					      bool exact,
 					      out DateTime result,
+					      out DateTimeOffset dto,
 					      DateTimeFormatInfo dfi,
 					      DateTimeStyles style,
 					      bool firstPartIsDate,
@@ -1170,6 +1237,7 @@ namespace System
 			bool useutc = false;
 			bool use_invariant = false;
 			bool sloppy_parsing = false;
+			dto = new DateTimeOffset (0, TimeSpan.Zero);
 #if !NET_2_0
 			bool afterTimePart = firstPartIsDate && secondPart == "";
 #endif
@@ -1186,6 +1254,9 @@ namespace System
 			if (format == null)
 				return false;
 
+			if (s == null)
+				return false;
+
 			if ((style & DateTimeStyles.AllowLeadingWhite) != 0) {
 				format = format.TrimStart (null);
 
@@ -1523,7 +1594,7 @@ namespace System
 					else if (num == 1)
 						tzoffset = _ParseNumber (s, valuePos, 1, 2, true, sloppy_parsing, out num_parsed);
 					else {
-						tzoffset = _ParseNumber (s, valuePos, 1, 2, true, sloppy_parsing, out num_parsed);
+						tzoffset = _ParseNumber (s, valuePos, 1, 2, true, /*sloppy_parsing*/true, out num_parsed);
 						valuePos += num_parsed;
 						if (num_parsed < 0)
 							return false;
@@ -1654,6 +1725,11 @@ namespace System
 			}
 
 #if NET_2_0
+			if (pos + 1 < len && chars [pos] == '.' && chars [pos + 1] == 'F') {
+				pos++;
+				while (pos < len && chars [pos] == 'F') // '.FFF....' can be mapped to nothing. See bug #444103
+					pos++;
+			}
 			while (pos < len && chars [pos] == 'K') // 'K' can be mapped to nothing
 				pos++;
 #endif
@@ -1729,20 +1805,29 @@ namespace System
 			if (dayofweek != -1 && dayofweek != (int) result.DayOfWeek)
 				return false;
 
-			TimeSpan utcoffset;
-
-			bool adjustToUniversal = (style & DateTimeStyles.AdjustToUniversal) != 0;
-			
-			if (tzsign != -1) {
+			if (tzsign == -1) {
+				if (result != DateTime.MinValue) {
+					try {
+						dto = new DateTimeOffset (result);
+					} catch { } // We handle this error in DateTimeOffset.Parse
+				}
+			} else {
 				if (tzoffmin == -1)
 					tzoffmin = 0;
 				if (tzoffset == -1)
 					tzoffset = 0;
-				if (tzsign == 1)
+				if (tzsign == 1) {
 					tzoffset = -tzoffset;
-
-				utcoffset = new TimeSpan (tzoffset, tzoffmin, 0);
-				long newticks = (result.ticks - utcoffset).Ticks;
+					tzoffmin = -tzoffmin;
+				}
+				try {
+					dto = new DateTimeOffset (result, new TimeSpan (tzoffset, tzoffmin, 0));
+				} catch {} // We handle this error in DateTimeOffset.Parse
+			}
+			bool adjustToUniversal = (style & DateTimeStyles.AdjustToUniversal) != 0;
+			
+			if (tzsign != -1) {
+				long newticks = (result.ticks - dto.Offset).Ticks;
 				if (newticks < 0)
 					newticks += TimeSpan.TicksPerDay;
 				result = new DateTime (false, new TimeSpan (newticks));
@@ -1802,8 +1887,9 @@ namespace System
 
 			DateTime result;
 			bool longYear = false;
-			if (!ParseExact (s, formats, dfi, style, out result, true, ref longYear))
-				throw new FormatException ();
+			Exception e = null;
+			if (!ParseExact (s, formats, dfi, style, out result, true, ref longYear, true, ref e))
+				throw e;
 			return result;
 		}		
 
@@ -1822,24 +1908,30 @@ namespace System
 
 		public static bool TryParse (string s, out DateTime result)
 		{
-			try {
-				result = Parse (s);
-			} catch {
-				result = MinValue;
-				return false;
+			if (s != null){
+				try {
+					Exception exception = null;
+					DateTimeOffset dto;
+
+					return CoreParse (s, null, DateTimeStyles.AllowWhiteSpaces, out result, out dto, false, ref exception);
+				} catch { }
 			}
-			return true;
+			result = MinValue;
+			return false;
 		}
 		
 		public static bool TryParse (string s, IFormatProvider provider, DateTimeStyles styles, out DateTime result)
 		{
-			try {
-				result = Parse (s, provider, styles);
-			} catch {
-				result = MinValue;
-				return false;
-			}
-			return true;
+			if (s != null){
+				try {
+					Exception exception = null;
+					DateTimeOffset dto;
+					
+					return CoreParse (s, provider, styles, out result, out dto, false, ref exception);
+				} catch {}
+			} 
+			result = MinValue;
+			return false;
 		}
 		
 		public static bool TryParseExact (string s, string format,
@@ -1848,7 +1940,6 @@ namespace System
 						  out DateTime result)
 		{
 			string[] formats;
-
 			formats = new string [1];
 			formats[0] = format;
 
@@ -1860,16 +1951,23 @@ namespace System
 						  DateTimeStyles style,
 						  out DateTime result)
 		{
-			DateTimeFormatInfo dfi = DateTimeFormatInfo.GetInstance (provider);
+			try {
+				DateTimeFormatInfo dfi = DateTimeFormatInfo.GetInstance (provider);
 
-			bool longYear = false;
-			return ParseExact (s, formats, dfi, style, out result, true, ref longYear);
+				bool longYear = false;
+				Exception e = null;
+				return ParseExact (s, formats, dfi, style, out result, true, ref longYear, false, ref e);
+			} catch {
+				result = MinValue;
+				return false;
+			}
 		}
 #endif
 
 		private static bool ParseExact (string s, string [] formats,
-			DateTimeFormatInfo dfi, DateTimeStyles style, out DateTime ret,
-			bool exact, ref bool longYear)
+						DateTimeFormatInfo dfi, DateTimeStyles style, out DateTime ret,
+						bool exact, ref bool longYear,
+						bool setExceptionOnError, ref Exception exception)
 		{
 			int i;
 			bool incompleteFormat = false;
@@ -1878,13 +1976,17 @@ namespace System
 				DateTime result;
 				string format = formats[i];
 				if (format == null || format == String.Empty)
-					throw new FormatException ("Invalid Format String");
+					break;
 
-				if (_DoParse (s, formats[i], null, exact, out result, dfi, style, false, ref incompleteFormat, ref longYear)) {
+				DateTimeOffset dto;
+				if (_DoParse (s, formats[i], null, exact, out result, out dto, dfi, style, false, ref incompleteFormat, ref longYear)) {
 					ret = result;
 					return true;
 				}
 			}
+
+			if (setExceptionOnError)
+				exception = new FormatException ("Invalid format string");
 			ret = DateTime.MinValue;
 			return false;
 		}
@@ -2125,44 +2227,72 @@ namespace System
 			throw new InvalidCastException();
 		}
 
+#if ONLY_1_1
+#pragma warning disable 3019
+		[CLSCompliant (false)]
+#endif
 		SByte IConvertible.ToSByte(IFormatProvider provider)
 		{
 			throw new InvalidCastException();
 		}
+#if ONLY_1_1
+#pragma warning restore 3019
+#endif
 
 		Single IConvertible.ToSingle(IFormatProvider provider)
 		{
 			throw new InvalidCastException();
 		}
 
-		object IConvertible.ToType (Type type, IFormatProvider provider)
+		object IConvertible.ToType (Type targetType, IFormatProvider provider)
 		{
-			if (type == null)
-				throw new ArgumentNullException ("type");
+			if (targetType == null)
+				throw new ArgumentNullException ("targetType");
 
-			if (type == typeof (DateTime))
+			if (targetType == typeof (DateTime))
 				return this;
-			else if (type == typeof (String))
+			else if (targetType == typeof (String))
 				return this.ToString (provider);
-			else if (type == typeof (Object))
+			else if (targetType == typeof (Object))
 				return this;
 			else
 				throw new InvalidCastException();
 		}
 
+#if ONLY_1_1
+#pragma warning disable 3019
+		[CLSCompliant (false)]
+#endif
 		UInt16 IConvertible.ToUInt16(IFormatProvider provider)
 		{
 			throw new InvalidCastException();
 		}
-		
+#if ONLY_1_1
+#pragma warning restore 3019
+#endif
+
+#if ONLY_1_1
+#pragma warning disable 3019
+		[CLSCompliant (false)]
+#endif
 		UInt32 IConvertible.ToUInt32(IFormatProvider provider)
 		{
 			throw new InvalidCastException();
 		}
+#if ONLY_1_1
+#pragma warning restore 3019
+#endif
 
+#if ONLY_1_1
+#pragma warning disable 3019
+		[CLSCompliant (false)]
+#endif
 		UInt64 IConvertible.ToUInt64(IFormatProvider provider)
 		{
 			throw new InvalidCastException();
 		}
+#if ONLY_1_1
+#pragma warning restore 3019
+#endif
 	}
 }
