@@ -185,12 +185,12 @@ namespace System
 				return false;
 			}
 
-			NumberFormatInfo nfi;
+			NumberFormatInfo nfi = null;
 			if (provider != null) {
 				Type typeNFI = typeof (NumberFormatInfo);
 				nfi = (NumberFormatInfo) provider.GetFormat (typeNFI);
 			}
-			else
+			if (nfi == null)
 				nfi = Thread.CurrentThread.CurrentCulture.NumberFormat;
 
 			if (!Int32.CheckStyle (style, tryParse, ref exc))
@@ -538,9 +538,12 @@ namespace System
 			return System.Convert.ToSingle (m_value);
 		}
 
-		object IConvertible.ToType (Type type, IFormatProvider provider)
+		object IConvertible.ToType (Type targetType, IFormatProvider provider)
 		{
-			return System.Convert.ToType (m_value, type, provider);
+			if (targetType == null)
+				throw new ArgumentNullException ("targetType");
+			
+			return System.Convert.ToType (m_value, targetType, provider, false);
 		}
 
 #if ONLY_1_1
