@@ -5,14 +5,16 @@
  *   Dietmar Maurer (dietmar@ximian.com)
  *   Gonzalo Paniagua Javier (gonzalo@ximian.com)
  *
- * (C) 2001-2003 Ximian, Inc.
- * (c) 2004,2005 Novell, Inc. (http://www.novell.com)
+ * Copyright 2001-2003 Ximian, Inc (http://www.ximian.com)
+ * Copyright 2004-2009 Novell, Inc (http://www.novell.com)
  */
 
 #include <config.h>
 #include <glib.h>
 
-#define THREADS_PER_CPU	5 /* 20 + THREADS_PER_CPU * number of CPUs */
+#define THREADS_PER_CPU	10 /* 20 + THREADS_PER_CPU * number of CPUs */
+#define THREAD_EXIT_TIMEOUT 1000
+#define INITIAL_QUEUE_LENGTH 128
 
 #include <mono/metadata/domain-internals.h>
 #include <mono/metadata/tabledefs.h>
@@ -28,6 +30,7 @@
 #include <mono/io-layer/io-layer.h>
 #include <mono/metadata/gc-internal.h>
 #include <mono/utils/mono-time.h>
+#include <mono/utils/mono-proclib.h>
 #include <errno.h>
 #ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
@@ -38,13 +41,17 @@
 #include <unistd.h>
 #endif
 #include <string.h>
-
+#ifdef HAVE_SYS_SOCKET_H
+#include <sys/socket.h>
+#endif
 #include <mono/utils/mono-poll.h>
 #ifdef HAVE_EPOLL
 #include <sys/epoll.h>
 #endif
 
+#ifndef DISABLE_SOCKETS
 #include "mono/io-layer/socket-wrappers.h"
+#endif
 
 #include "threadpool.h"
 
@@ -69,9 +76,8 @@ static int busy_io_worker_threads;
 
 /* mono_thread_pool_init called */
 static int tp_inited;
-
-/* we use this to store a reference to the AsyncResult to avoid GC */
-static MonoGHashTable *ares_htable = NULL;
+/* started idle threads */
+static int tp_idle_started;
 
 static CRITICAL_SECTION ares_lock;
 static CRITICAL_SECTION io_queue_lock;
@@ -119,6 +125,7 @@ typedef struct {
 static void async_invoke_thread (gpointer data);
 static void append_job (CRITICAL_SECTION *cs, TPQueue *list, MonoObject *ar);
 static void start_thread_or_queue (MonoAsyncResult *ares);
+static void start_tpthread (MonoAsyncResult *ares);
 static void mono_async_invoke (MonoAsyncResult *ares);
 static MonoObject* dequeue_job (CRITICAL_SECTION *cs, TPQueue *list);
 static void free_queue (TPQueue *list);
@@ -130,6 +137,14 @@ static MonoClass *async_call_klass;
 static MonoClass *socket_async_call_klass;
 static MonoClass *process_async_call_klass;
 
+/* Hooks */
+static MonoThreadPoolFunc tp_start_func;
+static MonoThreadPoolFunc tp_finish_func;
+static gpointer tp_hooks_user_data;
+static MonoThreadPoolItemFunc tp_item_begin_func;
+static MonoThreadPoolItemFunc tp_item_end_func;
+static gpointer tp_item_user_data;
+
 #define INIT_POLLFD(a, b, c) {(a)->fd = b; (a)->events = c; (a)->revents = 0;}
 enum {
 	AIO_OP_FIRST,
@@ -145,6 +160,9 @@ enum {
 	AIO_OP_LAST
 };
 
+#ifdef DISABLE_SOCKETS
+#define socket_io_cleanup(x)
+#else
 static void
 socket_io_cleanup (SocketIOData *data)
 {
@@ -225,14 +243,46 @@ get_events_from_list (MonoMList *list)
 				(SOCKET)(gssize)x->handle, x->buffer, x->offset, x->size,\
 				 x->socket_flags, &x->error);
 
+#endif /* !DISABLE_SOCKETS */
+
+static void
+unregister_job (MonoAsyncResult *obj)
+{
+}
+
+static void
+threadpool_jobs_inc (MonoObject *obj)
+{
+	if (obj)
+		InterlockedIncrement (&obj->vtable->domain->threadpool_jobs);
+}
+
+static gboolean
+threadpool_jobs_dec (MonoObject *obj)
+{
+	MonoDomain *domain = obj->vtable->domain;
+	int remaining_jobs = InterlockedDecrement (&domain->threadpool_jobs);
+	if (remaining_jobs == 0 && domain->cleanup_semaphore) {
+		ReleaseSemaphore (domain->cleanup_semaphore, 1, NULL);
+		return TRUE;
+	}
+	return FALSE;
+}
+
+#ifndef DISABLE_SOCKETS
 static void
 async_invoke_io_thread (gpointer data)
 {
 	MonoDomain *domain;
 	MonoThread *thread;
-	thread = mono_thread_current ();
-	ves_icall_System_Threading_Thread_SetState (thread, ThreadState_Background);
+	const gchar *version;
+	int workers_io, min_io;
 
+	thread = mono_thread_current ();
+ 	if (tp_start_func)
+ 		tp_start_func (tp_hooks_user_data);
+
+	version = mono_get_runtime_info ()->framework_version;
 	for (;;) {
 		MonoSocketAsyncResult *state;
 		MonoAsyncResult *ar;
@@ -253,27 +303,49 @@ async_invoke_io_thread (gpointer data)
 			/* worker threads invokes methods in different domains,
 			 * so we need to set the right domain here */
 			domain = ((MonoObject *)ar)->vtable->domain;
-			mono_thread_push_appdomain_ref (domain);
-			if (mono_domain_set (domain, FALSE)) {
-				ASyncCall *ac;
 
-				mono_async_invoke (ar);
-				ac = (ASyncCall *) ar->object_data;
-				/*
-				if (ac->msg->exc != NULL)
-					mono_unhandled_exception (ac->msg->exc);
-				*/
-				mono_domain_set (mono_get_root_domain (), TRUE);
+			g_assert (domain);
+
+			if (domain->state == MONO_APPDOMAIN_UNLOADED || domain->state == MONO_APPDOMAIN_UNLOADING) {
+				threadpool_jobs_dec ((MonoObject *)ar);
+				unregister_job (ar);
+				data = NULL;
+			} else {
+				mono_thread_push_appdomain_ref (domain);
+				if (threadpool_jobs_dec ((MonoObject *)ar)) {
+					unregister_job (ar);
+					data = NULL;
+					mono_thread_pop_appdomain_ref ();
+					continue;
+				}
+				if (mono_domain_set (domain, FALSE)) {
+					/* ASyncCall *ac; */
+
+					if (tp_item_begin_func)
+						tp_item_begin_func (tp_item_user_data);
+					mono_async_invoke (ar);
+					if (tp_item_end_func)
+						tp_item_end_func (tp_item_user_data);
+					/*
+					ac = (ASyncCall *) ar->object_data;
+					if (ac->msg->exc != NULL)
+						mono_unhandled_exception (ac->msg->exc);
+					*/
+					mono_domain_set (mono_get_root_domain (), TRUE);
+				}
+				mono_thread_pop_appdomain_ref ();
+				InterlockedDecrement (&busy_io_worker_threads);
+				/* If the callee changes the background status, set it back to TRUE */
+				if (*version != '1' && !mono_thread_test_state (thread , ThreadState_Background))
+					ves_icall_System_Threading_Thread_SetState (thread, ThreadState_Background);
 			}
-			mono_thread_pop_appdomain_ref ();
-			InterlockedDecrement (&busy_io_worker_threads);
 		}
 
 		data = dequeue_job (&io_queue_lock, &async_io_queue);
 	
 		if (!data) {
 			guint32 wr;
-			int timeout = 10000;
+			int timeout = THREAD_EXIT_TIMEOUT;
 			guint32 start_time = mono_msec_ticks ();
 			
 			do {
@@ -283,20 +355,31 @@ async_invoke_io_thread (gpointer data)
 			
 				timeout -= mono_msec_ticks () - start_time;
 			
-				if (wr != WAIT_TIMEOUT)
+				if (wr != WAIT_TIMEOUT && wr != WAIT_IO_COMPLETION)
 					data = dequeue_job (&io_queue_lock, &async_io_queue);
 			}
 			while (!data && timeout > 0);
 		}
 
 		if (!data) {
-			if (InterlockedDecrement (&io_worker_threads) < 2) {
-				/* If we have pending items, keep the thread alive */
-				if (InterlockedCompareExchange (&pending_io_items, 0, 0) != 0) {
-					InterlockedIncrement (&io_worker_threads);
-					continue;
-				}
+			workers_io = (int) InterlockedCompareExchange (&io_worker_threads, 0, -1); 
+			min_io = (int) InterlockedCompareExchange (&mono_io_min_worker_threads, 0, -1); 
+	
+			while (!data && workers_io <= min_io) {
+				WaitForSingleObjectEx (io_job_added, INFINITE, TRUE);
+				if (THREAD_WANTS_A_BREAK (thread))
+					mono_thread_interruption_checkpoint ();
+			
+				data = dequeue_job (&io_queue_lock, &async_io_queue);
+				workers_io = (int) InterlockedCompareExchange (&io_worker_threads, 0, -1); 
+				min_io = (int) InterlockedCompareExchange (&mono_io_min_worker_threads, 0, -1); 
 			}
+		}
+	
+		if (!data) {
+			InterlockedDecrement (&io_worker_threads);
+ 			if (tp_finish_func)
+ 				tp_finish_func (tp_hooks_user_data);
 			return;
 		}
 		
@@ -310,7 +393,6 @@ static void
 start_io_thread_or_queue (MonoSocketAsyncResult *ares)
 {
 	int busy, worker;
-	MonoDomain *domain;
 
 	busy = (int) InterlockedCompareExchange (&busy_io_worker_threads, 0, -1);
 	worker = (int) InterlockedCompareExchange (&io_worker_threads, 0, -1); 
@@ -318,7 +400,7 @@ start_io_thread_or_queue (MonoSocketAsyncResult *ares)
 	    worker < mono_io_max_worker_threads) {
 		InterlockedIncrement (&busy_io_worker_threads);
 		InterlockedIncrement (&io_worker_threads);
-		domain = ((ares) ? ((MonoObject *) ares)->vtable->domain : mono_domain_get ());
+		threadpool_jobs_inc ((MonoObject *)ares);
 		mono_thread_create_internal (mono_get_root_domain (), async_invoke_io_thread, ares, TRUE);
 	} else {
 		append_job (&io_queue_lock, &async_io_queue, (MonoObject*)ares);
@@ -391,7 +473,6 @@ socket_io_poll_main (gpointer p)
 	MonoThread *thread;
 
 	thread = mono_thread_current ();
-	ves_icall_System_Threading_Thread_SetState (thread, ThreadState_Background);
 
 	allocated = INITIAL_POLLFD_SIZE;
 	pfds = g_new0 (mono_pollfd, allocated);
@@ -532,8 +613,6 @@ socket_io_epoll_main (gpointer p)
 	data = p;
 	epollfd = data->epollfd;
 	thread = mono_thread_current ();
-	thread->threadpool_thread = TRUE;
-	ves_icall_System_Threading_Thread_SetState (thread, ThreadState_Background);
 	events = g_new0 (struct epoll_event, nevents);
 
 	while (1) {
@@ -744,10 +823,6 @@ socket_io_init (SocketIOData *data)
 	g_assert (data->pipe [0] != INVALID_SOCKET);
 	closesocket (srv);
 #endif
-	mono_io_max_worker_threads = mono_max_worker_threads / 2;
-	if (mono_io_max_worker_threads < 10)
-		mono_io_max_worker_threads = 10;
-
 	data->sock_to_state = mono_g_hash_table_new_type (g_direct_hash, g_direct_equal, MONO_HASH_VALUE_GC);
 
 	if (data->epoll_disabled) {
@@ -756,7 +831,6 @@ socket_io_init (SocketIOData *data)
 	}
 	io_job_added = CreateSemaphore (NULL, 0, 0x7fffffff, NULL);
 	g_assert (io_job_added != NULL);
-	InitializeCriticalSection (&io_queue_lock);
 	if (data->epoll_disabled) {
 		mono_thread_create_internal (mono_get_root_domain (), socket_io_poll_main, data, TRUE);
 	}
@@ -776,8 +850,9 @@ socket_io_add_poll (MonoSocketAsyncResult *state)
 	char msg [1];
 	MonoMList *list;
 	SocketIOData *data = &socket_io_data;
+	int w;
 
-#if defined(PLATFORM_MACOSX) || defined(PLATFORM_BSD6) || defined(PLATFORM_WIN32) || defined(PLATFORM_SOLARIS)
+#if defined(PLATFORM_MACOSX) || defined(PLATFORM_BSD) || defined(PLATFORM_WIN32) || defined(PLATFORM_SOLARIS)
 	/* select() for connect() does not work well on the Mac. Bug #75436. */
 	/* Bug #77637 for the BSD 6 case */
 	/* Bug #78888 for the Windows case */
@@ -805,7 +880,8 @@ socket_io_add_poll (MonoSocketAsyncResult *state)
 	LeaveCriticalSection (&data->io_lock);
 	*msg = (char) state->operation;
 #ifndef PLATFORM_WIN32
-	write (data->pipe [1], msg, 1);
+	w = write (data->pipe [1], msg, 1);
+	w = w;
 #else
 	send ((SOCKET) data->pipe [1], msg, 1, 0);
 #endif
@@ -916,6 +992,7 @@ socket_io_filter (MonoObject *target, MonoObject *state)
 
 	return TRUE;
 }
+#endif /* !DISABLE_SOCKETS */
 
 static void
 mono_async_invoke (MonoAsyncResult *ares)
@@ -928,8 +1005,8 @@ mono_async_invoke (MonoAsyncResult *ares)
 	if (ares->execution_context) {
 		/* use captured ExecutionContext (if available) */
 		thread = mono_thread_current ();
-		MONO_OBJECT_SETREF (ares, original_context, thread->execution_context);
-		MONO_OBJECT_SETREF (thread, execution_context, ares->execution_context);
+		MONO_OBJECT_SETREF (ares, original_context, mono_thread_get_execution_context ());
+		mono_thread_set_execution_context (ares->execution_context);
 	} else {
 		ares->original_context = NULL;
 	}
@@ -955,7 +1032,7 @@ mono_async_invoke (MonoAsyncResult *ares)
 
 	/* restore original thread execution context if flow isn't suppressed, i.e. non null */
 	if (ares->original_context) {
-		MONO_OBJECT_SETREF (thread, execution_context, ares->original_context);
+		mono_thread_set_execution_context (ares->original_context);
 		ares->original_context = NULL;
 	}
 
@@ -966,36 +1043,70 @@ mono_async_invoke (MonoAsyncResult *ares)
 		SetEvent ((gpointer)(gsize)ac->wait_event);
 	}
 	mono_monitor_exit ((MonoObject *) ares);
+}
 
-	EnterCriticalSection (&ares_lock);
-	mono_g_hash_table_remove (ares_htable, ares);
-	LeaveCriticalSection (&ares_lock);
+static void
+start_idle_threads (MonoAsyncResult *data)
+{
+	int needed;
+	int existing;
+
+	needed = (int) InterlockedCompareExchange (&mono_min_worker_threads, 0, -1); 
+	do {
+		existing = (int) InterlockedCompareExchange (&mono_worker_threads, 0, -1); 
+		if ((needed - existing) > 0) {
+			start_tpthread (data);
+			if (data) 
+				threadpool_jobs_dec ((MonoObject*)data);
+			data = NULL;
+			SleepEx (500, TRUE);
+		}
+	} while ((needed - existing) > 0);
+
+	/* If we don't start any thread here, make sure 'data' is processed. */
+	if (data != NULL) {
+		start_thread_or_queue (data);
+		threadpool_jobs_dec ((MonoObject*)data);
+	}
+}
+
+static void
+start_tpthread (MonoAsyncResult *data)
+{
+	InterlockedIncrement (&mono_worker_threads);
+	InterlockedIncrement (&busy_worker_threads);
+	threadpool_jobs_inc ((MonoObject *)data);
+	mono_thread_create_internal (mono_get_root_domain (), async_invoke_thread, data, TRUE);
 }
 
 void
 mono_thread_pool_init ()
 {
-	SYSTEM_INFO info;
 	int threads_per_cpu = THREADS_PER_CPU;
+	int cpu_count;
 
 	if ((int) InterlockedCompareExchange (&tp_inited, 1, 0) == 1)
 		return;
 
-	MONO_GC_REGISTER_ROOT (ares_htable);
 	MONO_GC_REGISTER_ROOT (socket_io_data.sock_to_state);
 	InitializeCriticalSection (&socket_io_data.io_lock);
 	InitializeCriticalSection (&ares_lock);
-	ares_htable = mono_g_hash_table_new_type (NULL, NULL, MONO_HASH_KEY_VALUE_GC);
+	InitializeCriticalSection (&io_queue_lock);
 	job_added = CreateSemaphore (NULL, 0, 0x7fffffff, NULL);
 	g_assert (job_added != NULL);
-	GetSystemInfo (&info);
 	if (g_getenv ("MONO_THREADS_PER_CPU") != NULL) {
 		threads_per_cpu = atoi (g_getenv ("MONO_THREADS_PER_CPU"));
 		if (threads_per_cpu <= 0)
 			threads_per_cpu = THREADS_PER_CPU;
 	}
 
-	mono_max_worker_threads = 20 + threads_per_cpu * info.dwNumberOfProcessors;
+	cpu_count = mono_cpu_count ();
+	mono_max_worker_threads = 20 + threads_per_cpu * cpu_count;
+	mono_min_worker_threads = cpu_count; /* 1 idle thread per cpu */
+	mono_io_max_worker_threads = mono_max_worker_threads / 2;
+	if (mono_io_max_worker_threads < 16)
+		mono_io_max_worker_threads = 16;
+	mono_io_min_worker_threads = cpu_count;
 
 	async_call_klass = mono_class_from_name (mono_defaults.corlib, "System", "MonoAsyncCall");
 	g_assert (async_call_klass);
@@ -1022,14 +1133,19 @@ mono_thread_pool_add (MonoObject *target, MonoMethodMessage *msg, MonoDelegate *
 	MONO_OBJECT_SETREF (ares, async_delegate, target);
 
 	EnterCriticalSection (&ares_lock);
-	mono_g_hash_table_insert (ares_htable, ares, ares);
+	if (domain->state == MONO_APPDOMAIN_UNLOADED || domain->state == MONO_APPDOMAIN_UNLOADING) {
+		LeaveCriticalSection (&ares_lock);
+		return ares;
+	}
 	LeaveCriticalSection (&ares_lock);
 
+#ifndef DISABLE_SOCKETS
 	if (socket_io_filter (target, state)) {
 		socket_io_add (ares, (MonoSocketAsyncResult *) state);
 		return ares;
 	}
-
+#endif
+	
 	start_thread_or_queue (ares);
 	return ares;
 }
@@ -1039,13 +1155,17 @@ start_thread_or_queue (MonoAsyncResult *ares)
 {
 	int busy, worker;
 
+	if ((int) InterlockedCompareExchange (&tp_idle_started, 1, 0) == 0) {
+		threadpool_jobs_inc ((MonoObject*)ares);
+		mono_thread_create_internal (mono_get_root_domain (), start_idle_threads, ares, TRUE);
+		return;
+	}
+
 	busy = (int) InterlockedCompareExchange (&busy_worker_threads, 0, -1);
 	worker = (int) InterlockedCompareExchange (&mono_worker_threads, 0, -1); 
 	if (worker <= ++busy &&
 	    worker < mono_max_worker_threads) {
-		InterlockedIncrement (&mono_worker_threads);
-		InterlockedIncrement (&busy_worker_threads);
-		mono_thread_create_internal (mono_get_root_domain (), async_invoke_thread, ares, TRUE);
+		start_tpthread (ares);
 	} else {
 		append_job (&mono_delegate_section, &async_call_queue, (MonoObject*)ares);
 		ReleaseSemaphore (job_added, 1, NULL);
@@ -1109,10 +1229,27 @@ mono_thread_pool_cleanup (void)
 	socket_io_cleanup (&socket_io_data);
 }
 
+static void
+null_array (MonoArray *a, int first, int last)
+{
+	/* We must null the old array because it might
+	   contain cross-appdomain references, which
+	   will crash the GC when the domains are
+	   unloaded. */
+	memset (mono_array_addr (a, MonoObject*, first), 0, sizeof (MonoObject*) * (last - first));
+}
+
 static void
 append_job (CRITICAL_SECTION *cs, TPQueue *list, MonoObject *ar)
 {
+	threadpool_jobs_inc (ar); 
+
 	EnterCriticalSection (cs);
+	if (ar->vtable->domain->state == MONO_APPDOMAIN_UNLOADING ||
+			ar->vtable->domain->state == MONO_APPDOMAIN_UNLOADED) {
+		LeaveCriticalSection (cs);
+		return;
+	}
 	if (list->array && (list->next_elem < mono_array_length (list->array))) {
 		mono_array_setref (list->array, list->next_elem, ar);
 		list->next_elem++;
@@ -1121,16 +1258,19 @@ append_job (CRITICAL_SECTION *cs, TPQueue *list, MonoObject *ar)
 	}
 	if (!list->array) {
 		MONO_GC_REGISTER_ROOT (list->array);
-		list->array = mono_array_new (mono_get_root_domain (), mono_defaults.object_class, 16);
+		list->array = mono_array_new_cached (mono_get_root_domain (), mono_defaults.object_class, INITIAL_QUEUE_LENGTH);
 	} else {
 		int count = list->next_elem - list->first_elem;
 		/* slide the array or create a larger one if it's full */
 		if (list->first_elem) {
 			mono_array_memcpy_refs (list->array, 0, list->array, list->first_elem, count);
+			null_array (list->array, count, list->next_elem);
 		} else {
-			MonoArray *newa = mono_array_new (mono_get_root_domain (), mono_defaults.object_class, mono_array_length (list->array) * 2);
+			MonoArray *olda = list->array;
+			MonoArray *newa = mono_array_new_cached (mono_get_root_domain (), mono_defaults.object_class, mono_array_length (list->array) * 2);
 			mono_array_memcpy_refs (newa, 0, list->array, list->first_elem, count);
 			list->array = newa;
+			null_array (olda, list->first_elem, list->next_elem);
 		}
 		list->first_elem = 0;
 		list->next_elem = count;
@@ -1140,6 +1280,85 @@ append_job (CRITICAL_SECTION *cs, TPQueue *list, MonoObject *ar)
 	LeaveCriticalSection (cs);
 }
 
+
+static void
+clear_queue (CRITICAL_SECTION *cs, TPQueue *list, MonoDomain *domain)
+{
+	int i, count = 0;
+	EnterCriticalSection (cs);
+	/*remove*/
+	for (i = list->first_elem; i < list->next_elem; ++i) {
+		MonoObject *obj = mono_array_get (list->array, MonoObject*, i);
+		if (obj->vtable->domain == domain) {
+			threadpool_jobs_dec (obj);
+			unregister_job ((MonoAsyncResult*)obj);
+
+			mono_array_set (list->array, MonoObject*, i, NULL);
+			InterlockedDecrement (&domain->threadpool_jobs);
+			++count;
+		}
+	}
+	/*compact*/
+	if (count) {
+		int idx = 0;
+		for (i = list->first_elem; i < list->next_elem; ++i) {
+			MonoObject *obj = mono_array_get (list->array, MonoObject*, i);
+			if (obj)
+				mono_array_set (list->array, MonoObject*, idx++, obj);
+		}
+		list->first_elem = 0;
+		list->next_elem = count;
+	}
+	LeaveCriticalSection (cs);
+}
+
+/*
+ * Clean up the threadpool of all domain jobs.
+ * Can only be called as part of the domain unloading process as
+ * it will wait for all jobs to be visible to the interruption code. 
+ */
+gboolean
+mono_thread_pool_remove_domain_jobs (MonoDomain *domain, int timeout)
+{
+	HANDLE sem_handle;
+	int result = TRUE;
+	guint32 start_time = 0;
+
+	g_assert (domain->state == MONO_APPDOMAIN_UNLOADING);
+
+	clear_queue (&mono_delegate_section, &async_call_queue, domain);
+	clear_queue (&io_queue_lock, &async_io_queue, domain);
+
+	/*
+	 * There might be some threads out that could be about to execute stuff from the given domain.
+	 * We avoid that by setting up a semaphore to be pulsed by the thread that reaches zero.
+	 */
+	sem_handle = CreateSemaphore (NULL, 0, 1, NULL);
+	
+	domain->cleanup_semaphore = sem_handle;
+	/*
+	 * The memory barrier here is required to have global ordering between assigning to cleanup_semaphone
+	 * and reading threadpool_jobs.
+	 * Otherwise this thread could read a stale version of threadpool_jobs and wait forever.
+	 */
+	mono_memory_write_barrier ();
+
+	if (domain->threadpool_jobs && timeout != -1)
+		start_time = mono_msec_ticks ();
+	while (domain->threadpool_jobs) {
+		WaitForSingleObject (sem_handle, timeout);
+		if (timeout != -1 && (mono_msec_ticks () - start_time) > timeout) {
+			result = FALSE;
+			break;
+		}
+	}
+
+	domain->cleanup_semaphore = NULL;
+	CloseHandle (sem_handle);
+	return result;
+}
+
+
 static MonoObject*
 dequeue_job (CRITICAL_SECTION *cs, TPQueue *list)
 {
@@ -1152,13 +1371,16 @@ dequeue_job (CRITICAL_SECTION *cs, TPQueue *list)
 		return NULL;
 	}
 	ar = mono_array_get (list->array, MonoObject*, list->first_elem);
+	mono_array_setref (list->array, list->first_elem, NULL);
 	list->first_elem++;
 	count = list->next_elem - list->first_elem;
 	/* reduce the size of the array if it's mostly empty */
-	if (mono_array_length (list->array) > 16 && count < (mono_array_length (list->array) / 3)) {
-		MonoArray *newa = mono_array_new (mono_get_root_domain (), mono_defaults.object_class, mono_array_length (list->array) / 2);
+	if (mono_array_length (list->array) > INITIAL_QUEUE_LENGTH && count < (mono_array_length (list->array) / 3)) {
+		MonoArray *olda = list->array;
+		MonoArray *newa = mono_array_new_cached (mono_get_root_domain (), mono_defaults.object_class, mono_array_length (list->array) / 2);
 		mono_array_memcpy_refs (newa, 0, list->array, list->first_elem, count);
 		list->array = newa;
+		null_array (olda, list->first_elem, list->next_elem);
 		list->first_elem = 0;
 		list->next_elem = count;
 	}
@@ -1170,6 +1392,8 @@ dequeue_job (CRITICAL_SECTION *cs, TPQueue *list)
 static void
 free_queue (TPQueue *list)
 {
+	if (list->array)
+		null_array (list->array, list->first_elem, list->next_elem);
 	list->array = NULL;
 	list->first_elem = list->next_elem = 0;
 }
@@ -1180,11 +1404,13 @@ async_invoke_thread (gpointer data)
 	MonoDomain *domain;
 	MonoThread *thread;
 	int workers, min;
+	const gchar *version;
  
 	thread = mono_thread_current ();
-	thread->threadpool_thread = TRUE;
-	ves_icall_System_Threading_Thread_SetState (thread, ThreadState_Background);
+ 	if (tp_start_func)
+ 		tp_start_func (tp_hooks_user_data);
 
+	version = mono_get_runtime_info ()->framework_version;
 	for (;;) {
 		MonoAsyncResult *ar;
 
@@ -1193,27 +1419,50 @@ async_invoke_thread (gpointer data)
 			/* worker threads invokes methods in different domains,
 			 * so we need to set the right domain here */
 			domain = ((MonoObject *)ar)->vtable->domain;
-			mono_thread_push_appdomain_ref (domain);
-			if (mono_domain_set (domain, FALSE)) {
-				ASyncCall *ac;
 
-				mono_async_invoke (ar);
-				ac = (ASyncCall *) ar->object_data;
-				/*
-				if (ac->msg->exc != NULL)
-					mono_unhandled_exception (ac->msg->exc);
-				*/
-				mono_domain_set (mono_get_root_domain (), TRUE);
+			g_assert (domain);
+
+			if (domain->state == MONO_APPDOMAIN_UNLOADED || domain->state == MONO_APPDOMAIN_UNLOADING) {
+				threadpool_jobs_dec ((MonoObject *)ar);
+				unregister_job (ar);
+				data = NULL;
+			} else {
+				mono_thread_push_appdomain_ref (domain);
+				if (threadpool_jobs_dec ((MonoObject *)ar)) {
+					unregister_job (ar);
+					data = NULL;
+					mono_thread_pop_appdomain_ref ();
+					continue;
+				}
+
+				if (mono_domain_set (domain, FALSE)) {
+					/* ASyncCall *ac; */
+
+					if (tp_item_begin_func)
+						tp_item_begin_func (tp_item_user_data);
+					mono_async_invoke (ar);
+					if (tp_item_end_func)
+						tp_item_end_func (tp_item_user_data);
+					/*
+					ac = (ASyncCall *) ar->object_data;
+					if (ac->msg->exc != NULL)
+						mono_unhandled_exception (ac->msg->exc);
+					*/
+					mono_domain_set (mono_get_root_domain (), TRUE);
+				}
+				mono_thread_pop_appdomain_ref ();
+				InterlockedDecrement (&busy_worker_threads);
+				/* If the callee changes the background status, set it back to TRUE */
+				if (*version != '1' && !mono_thread_test_state (thread , ThreadState_Background))
+					ves_icall_System_Threading_Thread_SetState (thread, ThreadState_Background);
 			}
-			mono_thread_pop_appdomain_ref ();
-			InterlockedDecrement (&busy_worker_threads);
 		}
 
 		data = dequeue_job (&mono_delegate_section, &async_call_queue);
 
 		if (!data) {
 			guint32 wr;
-			int timeout = 10000;
+			int timeout = THREAD_EXIT_TIMEOUT;
 			guint32 start_time = mono_msec_ticks ();
 			
 			do {
@@ -1223,7 +1472,7 @@ async_invoke_thread (gpointer data)
 			
 				timeout -= mono_msec_ticks () - start_time;
 			
-				if (wr != WAIT_TIMEOUT)
+				if (wr != WAIT_TIMEOUT && wr != WAIT_IO_COMPLETION)
 					data = dequeue_job (&mono_delegate_section, &async_call_queue);
 			}
 			while (!data && timeout > 0);
@@ -1246,6 +1495,8 @@ async_invoke_thread (gpointer data)
 	
 		if (!data) {
 			InterlockedDecrement (&mono_worker_threads);
+ 			if (tp_finish_func)
+ 				tp_finish_func (tp_hooks_user_data);
 			return;
 		}
 		
@@ -1294,32 +1545,81 @@ ves_icall_System_Threading_ThreadPool_GetMinThreads (gint *workerThreads, gint *
 MonoBoolean
 ves_icall_System_Threading_ThreadPool_SetMinThreads (gint workerThreads, gint completionPortThreads)
 {
+	int max_threads;
+	int max_io_threads;
+
 	MONO_ARCH_SAVE_REGS;
 
-	if (workerThreads < 0 || workerThreads > mono_max_worker_threads)
+	max_threads = InterlockedCompareExchange (&mono_max_worker_threads, -1, -1);
+	if (workerThreads <= 0 || workerThreads > max_threads)
 		return FALSE;
 
-	if (completionPortThreads < 0 || completionPortThreads > mono_io_max_worker_threads)
+	max_io_threads = InterlockedCompareExchange (&mono_io_max_worker_threads, -1, -1);
+	if (completionPortThreads <= 0 || completionPortThreads > max_io_threads)
 		return FALSE;
 
 	InterlockedExchange (&mono_min_worker_threads, workerThreads);
 	InterlockedExchange (&mono_io_min_worker_threads, completionPortThreads);
-	/* FIXME: should actually start the idle threads if needed */
+	mono_thread_create_internal (mono_get_root_domain (), start_idle_threads, NULL, TRUE);
 	return TRUE;
 }
 
 MonoBoolean
 ves_icall_System_Threading_ThreadPool_SetMaxThreads (gint workerThreads, gint completionPortThreads)
 {
+	gint32 min_threads;
+	gint32 min_io_threads;
+	gint32 cpu_count;
+
 	MONO_ARCH_SAVE_REGS;
 
-	if (workerThreads < mono_max_worker_threads)
+	cpu_count = mono_cpu_count ();
+	min_threads = InterlockedCompareExchange (&mono_min_worker_threads, -1, -1);
+	if (workerThreads < min_threads || workerThreads < cpu_count)
 		return FALSE;
 
-	if (completionPortThreads < mono_io_max_worker_threads)
+	/* We don't really have the concept of completion ports. Do we care here? */
+	min_io_threads = InterlockedCompareExchange (&mono_io_min_worker_threads, -1, -1);
+	if (completionPortThreads < min_io_threads || completionPortThreads < cpu_count)
 		return FALSE;
 
 	InterlockedExchange (&mono_max_worker_threads, workerThreads);
 	InterlockedExchange (&mono_io_max_worker_threads, completionPortThreads);
 	return TRUE;
 }
+
+/**
+ * mono_install_threadpool_thread_hooks
+ * @start_func: the function to be called right after a new threadpool thread is created. Can be NULL.
+ * @finish_func: the function to be called right before a thredpool thread is exiting. Can be NULL.
+ * @user_data: argument passed to @start_func and @finish_func.
+ *
+ * @start_fun will be called right after a threadpool thread is created and @finish_func right before a threadpool thread exits.
+ * The calls will be made from the thread itself.
+ */
+void
+mono_install_threadpool_thread_hooks (MonoThreadPoolFunc start_func, MonoThreadPoolFunc finish_func, gpointer user_data)
+{
+	tp_start_func = start_func;
+	tp_finish_func = finish_func;
+	tp_hooks_user_data = user_data;
+}
+
+/**
+ * mono_install_threadpool_item_hooks
+ * @begin_func: the function to be called before a threadpool work item processing starts.
+ * @end_func: the function to be called after a threadpool work item is finished.
+ * @user_data: argument passed to @begin_func and @end_func.
+ *
+ * The calls will be made from the thread itself and from the same AppDomain
+ * where the work item was executed.
+ *
+ */
+void
+mono_install_threadpool_item_hooks (MonoThreadPoolItemFunc begin_func, MonoThreadPoolItemFunc end_func, gpointer user_data)
+{
+	tp_item_begin_func = begin_func;
+	tp_item_end_func = end_func;
+	tp_item_user_data = user_data;
+}
+
