@@ -22,7 +22,6 @@
 guint8* mono_trampoline_code [MONO_TRAMPOLINE_NUM];
 
 static GHashTable *class_init_hash_addr = NULL;
-static GHashTable *delegate_trampoline_hash_addr = NULL;
 static GHashTable *rgctx_lazy_fetch_trampoline_hash = NULL;
 static GHashTable *rgctx_lazy_fetch_trampoline_hash_addr = NULL;
 
@@ -31,18 +30,135 @@ static GHashTable *rgctx_lazy_fetch_trampoline_hash_addr = NULL;
 static CRITICAL_SECTION trampolines_mutex;
 
 static gpointer
-get_unbox_trampoline (MonoGenericSharingContext *gsctx, MonoMethod *m, gpointer addr)
+get_unbox_trampoline (MonoGenericSharingContext *gsctx, MonoMethod *m, gpointer addr, gboolean need_rgctx_tramp)
 {
-	if (mono_aot_only)
-		return mono_aot_get_unbox_trampoline (m);
-	else
+	if (mono_aot_only) {
+		if (need_rgctx_tramp)
+			/* 
+			 * The unbox trampolines call the method directly, so need to add
+			 * an rgctx tramp before them.
+			 */
+			return mono_create_static_rgctx_trampoline (m, mono_aot_get_unbox_trampoline (m));
+		else
+			return mono_aot_get_unbox_trampoline (m);
+	} else {
 		return mono_arch_get_unbox_trampoline (gsctx, m, addr);
+	}
+}
+
+#ifdef MONO_ARCH_HAVE_STATIC_RGCTX_TRAMPOLINE
+
+typedef struct {
+	MonoMethod *m;
+	gpointer addr;
+} RgctxTrampInfo;
+
+static gint
+rgctx_tramp_info_equal (gconstpointer ka, gconstpointer kb)
+{
+	const RgctxTrampInfo *i1 = ka;
+	const RgctxTrampInfo *i2 = kb;
+
+	if (i1->m == i2->m && i1->addr == i2->addr)
+		return 1;
+	else
+		return 0;
+}
+
+static guint
+rgctx_tramp_info_hash (gconstpointer data)
+{
+	const RgctxTrampInfo *info = data;
+
+	return GPOINTER_TO_UINT (info->m) ^ GPOINTER_TO_UINT (info->addr);
+}
+
+/*
+ * mono_create_static_rgctx_trampoline:
+ *
+ *   Return a static rgctx trampoline for M which branches to ADDR which should
+ * point to the compiled code of M.
+ *
+ *   Static rgctx trampolines are used when a shared generic method which doesn't
+ * have a this argument is called indirectly, ie. from code which can't pass in
+ * the rgctx argument. The trampoline sets the rgctx argument and jumps to the
+ * methods code. These trampolines are similar to the unbox trampolines, they
+ * perform the same task as the static rgctx wrappers, but they are smaller/faster,
+ * and can be made to work with full AOT.
+ */
+gpointer
+mono_create_static_rgctx_trampoline (MonoMethod *m, gpointer addr)
+{
+	gpointer ctx;
+	gpointer res;
+	MonoDomain *domain;
+	RgctxTrampInfo tmp_info;
+	RgctxTrampInfo *info;
+
+	if (mini_method_get_context (m)->method_inst)
+		ctx = mono_method_lookup_rgctx (mono_class_vtable (mono_domain_get (), m->klass), mini_method_get_context (m)->method_inst);
+	else
+		ctx = mono_class_vtable (mono_domain_get (), m->klass);
+
+	domain = mono_domain_get ();
+
+	/* 
+	 * In the AOT case, addr might point to either the method, or to an unbox trampoline,
+	 * so make the hash keyed on the m+addr pair.
+	 */
+	mono_domain_lock (domain);
+	if (!domain_jit_info (domain)->static_rgctx_trampoline_hash)
+		domain_jit_info (domain)->static_rgctx_trampoline_hash = g_hash_table_new (rgctx_tramp_info_hash, rgctx_tramp_info_equal);
+	tmp_info.m = m;
+	tmp_info.addr = addr;
+	res = g_hash_table_lookup (domain_jit_info (domain)->static_rgctx_trampoline_hash,
+							   &tmp_info);
+	mono_domain_unlock (domain);
+	if (res)
+		return res;
+
+	if (mono_aot_only)
+		res = mono_aot_get_static_rgctx_trampoline (ctx, addr);
+	else
+		res = mono_arch_get_static_rgctx_trampoline (m, ctx, addr);
+
+	mono_domain_lock (domain);
+	/* Duplicates inserted while we didn't hold the lock are OK */
+	info = mono_domain_alloc (domain, sizeof (RgctxTrampInfo));
+	info->m = m;
+	info->addr = addr;
+	g_hash_table_insert (domain_jit_info (domain)->static_rgctx_trampoline_hash, info, res);
+	mono_domain_unlock (domain);
+
+	return res;
+}
+#else
+gpointer
+mono_create_static_rgctx_trampoline (MonoMethod *m, gpointer addr)
+{
+	/* 
+	 * This shouldn't happen as all arches which support generic sharing support
+	 * static rgctx trampolines as well.
+	 */
+	g_assert_not_reached ();
+}
+#endif
+
+gpointer*
+mono_get_vcall_slot_addr (guint8* code, mgreg_t *regs)
+{
+	gpointer vt;
+	int displacement;
+	vt = mono_arch_get_vcall_slot (code, regs, &displacement);
+	if (!vt)
+		return NULL;
+	return (gpointer*)((char*)vt + displacement);
 }
 
 #ifdef MONO_ARCH_HAVE_IMT
 
 static gpointer*
-mono_convert_imt_slot_to_vtable_slot (gpointer* slot, gpointer *regs, guint8 *code, MonoMethod *method, MonoMethod **impl_method)
+mono_convert_imt_slot_to_vtable_slot (gpointer* slot, mgreg_t *regs, guint8 *code, MonoMethod *method, MonoMethod **impl_method, gboolean *need_rgctx_tramp)
 {
 	MonoGenericSharingContext *gsctx = mono_get_generic_context_from_code (code);
 	MonoObject *this_argument = mono_arch_find_this_argument (regs, method, gsctx);
@@ -68,17 +184,46 @@ mono_convert_imt_slot_to_vtable_slot (gpointer* slot, gpointer *regs, guint8 *co
 		}
 		mono_vtable_build_imt_slot (vt, mono_method_get_imt_slot (imt_method));
 
-		if (impl_method)
-			*impl_method = mono_class_get_vtable_entry (vt->klass, interface_offset + imt_method->slot);
+		if (impl_method) {
+			MonoMethod *impl;
+
+			if (imt_method->is_inflated && ((MonoMethodInflated*)imt_method)->context.method_inst) {
+				MonoGenericContext context = { NULL, NULL };
+
+				/* 
+				 * Generic virtual method, imt_method contains the inflated interface 
+				 * method, need to get the inflated impl method.
+				 */
+				/* imt_method->slot might not be set */
+				impl = mono_class_get_vtable_entry (vt->klass, interface_offset + mono_method_get_declaring_generic_method (imt_method)->slot);
+
+				if (impl->klass->generic_class)
+					context.class_inst = impl->klass->generic_class->context.class_inst;
+				context.method_inst = ((MonoMethodInflated*)imt_method)->context.method_inst;
+				impl = mono_class_inflate_generic_method (impl, &context);
+			} else {
+				impl = mono_class_get_vtable_entry (vt->klass, interface_offset + mono_method_get_vtable_slot (imt_method));
+			}
+
+			if (mono_method_needs_static_rgctx_invoke (impl, FALSE))
+				*need_rgctx_tramp = TRUE;
+
+			*impl_method = impl;
 #if DEBUG_IMT
 		printf ("mono_convert_imt_slot_to_vtable_slot: method = %s.%s.%s, imt_method = %s.%s.%s\n",
 				method->klass->name_space, method->klass->name, method->name, 
 				imt_method->klass->name_space, imt_method->klass->name, imt_method->name);
 #endif
+		}
 		g_assert (imt_slot < MONO_IMT_SIZE);
 		if (vt->imt_collisions_bitmap & (1 << imt_slot)) {
-			int vtable_offset = interface_offset + mono_method_get_vtable_index (imt_method);
-			gpointer *vtable_slot = & (vt->vtable [vtable_offset]);
+			int slot = mono_method_get_vtable_index (imt_method);
+			int vtable_offset;
+			gpointer *vtable_slot;
+
+			g_assert (slot != -1);
+			vtable_offset = interface_offset + slot;
+			vtable_slot = & (vt->vtable [vtable_offset]);
 #if DEBUG_IMT
 			printf ("mono_convert_imt_slot_to_vtable_slot: slot %p[%d] is in the IMT, and colliding becomes %p[%d] (interface_offset = %d, method->slot = %d)\n", slot, imt_slot, vtable_slot, vtable_offset, interface_offset, imt_method->slot);
 #endif
@@ -99,22 +244,30 @@ mono_convert_imt_slot_to_vtable_slot (gpointer* slot, gpointer *regs, guint8 *co
  *   This trampoline handles calls from JITted code.
  */
 gpointer
-mono_magic_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8* tramp)
+mono_magic_trampoline (mgreg_t *regs, guint8 *code, gpointer arg, guint8* tramp)
 {
-	gpointer addr;
+	gpointer addr, compiled_method;
 	gpointer *vtable_slot;
 	gboolean generic_shared = FALSE;
+	MonoMethod *m;
 	MonoMethod *declaring = NULL;
-	MonoGenericInst *generic_virtual_method_inst = NULL;
+	MonoMethod *generic_virtual = NULL;
 	int context_used;
+	gboolean proxy = FALSE;
+	gboolean need_rgctx_tramp = FALSE;
+
+	m = arg;
 
-#if MONO_ARCH_COMMON_VTABLE_TRAMPOLINE
 	if (m == MONO_FAKE_VTABLE_METHOD) {
 		int displacement;
-		MonoVTable *vt = mono_arch_get_vcall_slot (code, (gpointer*)regs, &displacement);
+		MonoVTable *vt = mono_arch_get_vcall_slot (code, regs, &displacement);
 		if (!vt) {
 			int i;
+			MonoJitInfo *ji;
 
+			ji = mono_jit_info_table_find (mono_domain_get (), (char*)code);
+			if (ji)
+				printf ("Caller: %s\n", mono_method_full_name (ji->method, TRUE));
 			/* Print some debug info */
 			for (i = 0; i < 32; ++i)
 				printf ("0x%x ", code [-32 + i]);
@@ -128,8 +281,10 @@ mono_magic_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8* tramp)
 
 			/* Avoid loading metadata or creating a generic vtable if possible */
 			addr = mono_aot_get_method_from_vt_slot (mono_domain_get (), vt, displacement);
+			if (addr)
+				addr = mono_create_ftnptr (mono_domain_get (), addr);
 			if (addr && !vt->klass->valuetype) {
-				vtable_slot = mono_arch_get_vcall_slot_addr (code, (gpointer*)regs);
+				vtable_slot = mono_get_vcall_slot_addr (code, regs);
 				if (mono_aot_is_got_entry (code, (guint8*)vtable_slot) || mono_domain_owns_vtable_slot (mono_domain_get (), vtable_slot)) {
 					*vtable_slot = mono_get_addr_from_ftnptr (addr);
 				}
@@ -137,8 +292,10 @@ mono_magic_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8* tramp)
 				return addr;
 			}
 
-			mono_class_setup_vtable (vt->klass);
-			m = vt->klass->vtable [displacement];
+			m = mono_class_get_vtable_entry (vt->klass, displacement);
+			if (mono_method_needs_static_rgctx_invoke (m, FALSE))
+				need_rgctx_tramp = TRUE;
+
 			/*g_print ("%s with disp %d: %s at %p\n", vt->klass->name, displacement, m->name, code);*/
 		} else {
 			/* We got here from an interface method: redirect to IMT handling */
@@ -146,23 +303,43 @@ mono_magic_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8* tramp)
 			/*g_print ("vtable with disp %d at %p\n", displacement, code);*/
 		}
 	}
-#endif
+
 	/* this is the IMT trampoline */
 #ifdef MONO_ARCH_HAVE_IMT
 	if (m == MONO_FAKE_IMT_METHOD) {
 		MonoMethod *impl_method;
+		MonoGenericSharingContext *gsctx;
+		MonoObject *this_arg;
+
 		/* we get the interface method because mono_convert_imt_slot_to_vtable_slot ()
 		 * needs the signature to be able to find the this argument
 		 */
-		m = mono_arch_find_imt_method ((gpointer*)regs, code);
-		vtable_slot = mono_arch_get_vcall_slot_addr (code, (gpointer*)regs);
+		m = mono_arch_find_imt_method (regs, code);
+		vtable_slot = mono_get_vcall_slot_addr (code, regs);
 		g_assert (vtable_slot);
-		vtable_slot = mono_convert_imt_slot_to_vtable_slot (vtable_slot, (gpointer*)regs, code, m, &impl_method);
-		/* mono_convert_imt_slot_to_vtable_slot () also gives us the method that is supposed
-		 * to be called, so we compile it and go ahead as usual.
-		 */
-		/*g_print ("imt found method %p (%s) at %p\n", impl_method, impl_method->name, code);*/
-		m = impl_method;
+
+		gsctx = mono_get_generic_context_from_code (code);
+		this_arg = mono_arch_find_this_argument (regs, m, gsctx);
+
+		if (this_arg->vtable->klass == mono_defaults.transparent_proxy_class) {
+			/* Use the slow path for now */
+			proxy = TRUE;
+		    m = mono_object_get_virtual_method (this_arg, m);
+		} else {
+			vtable_slot = mono_convert_imt_slot_to_vtable_slot (vtable_slot, regs, code, m, &impl_method, &need_rgctx_tramp);
+			/* mono_convert_imt_slot_to_vtable_slot () also gives us the method that is supposed
+			 * to be called, so we compile it and go ahead as usual.
+			 */
+			/*g_print ("imt found method %p (%s) at %p\n", impl_method, impl_method->name, code);*/
+			if (m->is_inflated && ((MonoMethodInflated*)m)->context.method_inst) {
+				/* Generic virtual method */
+				generic_virtual = m;
+				m = impl_method;
+				need_rgctx_tramp = TRUE;
+			} else {
+				m = impl_method;
+			}
+		}
 	}
 #endif
 
@@ -181,13 +358,16 @@ mono_magic_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8* tramp)
 			g_assert (!m->klass->generic_container);
 
 #ifdef MONO_ARCH_HAVE_IMT
-		generic_virtual_method_inst = (MonoGenericInst*)mono_arch_find_imt_method ((gpointer*)regs, code);
+		generic_virtual = mono_arch_find_imt_method (regs, code);
 #endif
-		context.method_inst = generic_virtual_method_inst;
+		if (generic_virtual) {
+			g_assert (generic_virtual->is_inflated);
+			context.method_inst = ((MonoMethodInflated*)generic_virtual)->context.method_inst;
+		}
 
 		m = mono_class_inflate_generic_method (declaring, &context);
 		/* FIXME: only do this if the method is sharable */
-		m = mono_marshal_get_static_rgctx_invoke (m);
+		need_rgctx_tramp = TRUE;
 	} else if ((context_used = mono_method_check_context_used (m))) {
 		MonoClass *klass = NULL;
 		MonoMethod *actual_method = NULL;
@@ -201,7 +381,7 @@ mono_magic_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8* tramp)
 
 		if (m->is_inflated && mono_method_get_context (m)->method_inst) {
 #ifdef MONO_ARCH_RGCTX_REG
-			MonoMethodRuntimeGenericContext *mrgctx = (MonoMethodRuntimeGenericContext*)mono_arch_find_static_call_vtable ((gpointer*)regs, code);
+			MonoMethodRuntimeGenericContext *mrgctx = (MonoMethodRuntimeGenericContext*)mono_arch_find_static_call_vtable (regs, code);
 
 			klass = mrgctx->class_vtable->klass;
 			method_inst = mrgctx->method_inst;
@@ -210,7 +390,7 @@ mono_magic_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8* tramp)
 #endif
 		} else if ((m->flags & METHOD_ATTRIBUTE_STATIC) || m->klass->valuetype) {
 #ifdef MONO_ARCH_RGCTX_REG
-			MonoVTable *vtable = mono_arch_find_static_call_vtable ((gpointer*)regs, code);
+			MonoVTable *vtable = mono_arch_find_static_call_vtable (regs, code);
 
 			klass = vtable->klass;
 #else
@@ -218,11 +398,11 @@ mono_magic_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8* tramp)
 #endif
 		} else {
 #ifdef MONO_ARCH_HAVE_IMT
-			MonoObject *this_argument = mono_arch_find_this_argument ((gpointer*)regs, m,
+			MonoObject *this_argument = mono_arch_find_this_argument (regs, m,
 				mono_get_generic_context_from_code (code));
 
 			vt = this_argument->vtable;
-			vtable_slot = mono_arch_get_vcall_slot_addr (code, (gpointer*)regs);
+			vtable_slot = mono_get_vcall_slot_addr (code, regs);
 
 			g_assert (this_argument->vtable->klass->inited);
 			//mono_class_init (this_argument->vtable->klass);
@@ -266,6 +446,7 @@ mono_magic_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8* tramp)
 		}
 
 		g_assert (klass);
+		g_assert (actual_method);
 		g_assert (actual_method->klass == klass);
 
 		if (actual_method->is_inflated)
@@ -289,17 +470,31 @@ mono_magic_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8* tramp)
 			m = mono_marshal_get_synchronized_wrapper (m);
 	}
 
-	addr = mono_compile_method (m);
+	/* Calls made through delegates on platforms without delegate trampolines */
+	if (!code && mono_method_needs_static_rgctx_invoke (m, FALSE))
+		need_rgctx_tramp = TRUE;
+
+	addr = compiled_method = mono_compile_method (m);
 	g_assert (addr);
 
 	mono_debugger_trampoline_compiled (code, m, addr);
 
-	if (generic_virtual_method_inst) {
-		vtable_slot = mono_arch_get_vcall_slot_addr (code, (gpointer*)regs);
+	if (need_rgctx_tramp)
+		addr = mono_create_static_rgctx_trampoline (m, addr);
+
+	if (generic_virtual) {
+		int displacement;
+ 		MonoVTable *vt = mono_arch_get_vcall_slot (code, regs, &displacement);
+
+		vtable_slot = mono_get_vcall_slot_addr (code, regs);
 		g_assert (vtable_slot);
 
-		mono_method_add_generic_virtual_invocation (mono_domain_get (), vtable_slot,
-			generic_virtual_method_inst, addr);
+		if (vt->klass->valuetype)
+			addr = get_unbox_trampoline (mono_get_generic_context_from_code (code), m, addr, need_rgctx_tramp);
+
+		mono_method_add_generic_virtual_invocation (mono_domain_get (), 
+													vt, vtable_slot,
+													generic_virtual, addr);
 
 		return addr;
 	}
@@ -332,17 +527,16 @@ mono_magic_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8* tramp)
 		return addr;
 	}
 
-	vtable_slot = mono_arch_get_vcall_slot_addr (code, (gpointer*)regs);
+	vtable_slot = mono_get_vcall_slot_addr (code, regs);
 
 	if (vtable_slot) {
 		if (m->klass->valuetype)
-			addr = get_unbox_trampoline (mono_get_generic_context_from_code (code), m, addr);
-
+			addr = get_unbox_trampoline (mono_get_generic_context_from_code (code), m, addr, need_rgctx_tramp);
 		g_assert (*vtable_slot);
 
-		if (mono_aot_is_got_entry (code, (guint8*)vtable_slot) || mono_domain_owns_vtable_slot (mono_domain_get (), vtable_slot)) {
+		if (!proxy && (mono_aot_is_got_entry (code, (guint8*)vtable_slot) || mono_domain_owns_vtable_slot (mono_domain_get (), vtable_slot))) {
 #ifdef MONO_ARCH_HAVE_IMT
-			vtable_slot = mono_convert_imt_slot_to_vtable_slot (vtable_slot, (gpointer*)regs, code, m, NULL);
+			vtable_slot = mono_convert_imt_slot_to_vtable_slot (vtable_slot, regs, code, m, NULL, &need_rgctx_tramp);
 #endif
 			*vtable_slot = mono_get_addr_from_ftnptr (addr);
 		}
@@ -351,7 +545,7 @@ mono_magic_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8* tramp)
 		guint8 *plt_entry = mono_aot_get_plt_entry (code);
 
 		if (plt_entry) {
-			mono_arch_patch_plt_entry (plt_entry, addr);
+			mono_aot_patch_plt_entry (plt_entry, NULL, regs, addr);
 		} else if (!generic_shared || (m->iflags & METHOD_IMPL_ATTRIBUTE_INTERNAL_CALL) ||
 			mono_domain_lookup_shared_generic (mono_domain_get (), declaring)) {
 			if (generic_shared) {
@@ -367,7 +561,7 @@ mono_magic_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8* tramp)
 				MonoJitInfo *ji = 
 					mono_jit_info_table_find (mono_domain_get (), (char*)code);
 				MonoJitInfo *target_ji = 
-					mono_jit_info_table_find (mono_domain_get (), mono_get_addr_from_ftnptr (addr));
+					mono_jit_info_table_find (mono_domain_get (), mono_get_addr_from_ftnptr (compiled_method));
 
 				if (mono_method_same_domain (ji, target_ji))
 					mono_arch_patch_callsite (ji->code_start, code, addr);
@@ -377,12 +571,81 @@ mono_magic_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8* tramp)
 
 	return addr;
 }
+ 
+#ifdef ENABLE_LLVM
+/*
+ * mono_llvm_vcall_trampoline:
+ *
+ *   This trampoline handles virtual calls when using LLVM.
+ */
+static gpointer
+mono_llvm_vcall_trampoline (mgreg_t *regs, guint8 *code, MonoMethod *m, guint8 *tramp)
+{
+	MonoObject *this;
+	gpointer addr;
+	MonoVTable *vt;
+	gpointer *vtable_slot;
+	gboolean proxy = FALSE;
+	gboolean need_rgctx_tramp = FALSE;
+
+	/* 
+	 * We have the method which is called, we need to obtain the vtable slot without
+	 * disassembly which is impossible with LLVM.
+	 * So we use the this argument.
+	 */
+	this = mono_arch_get_this_arg_from_call (NULL, mono_method_signature (m), regs, code);
+	g_assert (this);
+
+	vt = this->vtable;
+
+	g_assert (!m->is_generic);
+
+	/* This is a simplified version of mono_magic_trampoline () */
+	/* FIXME: Avoid code duplication */
+
+	if (m->iflags & METHOD_IMPL_ATTRIBUTE_SYNCHRONIZED) {
+		MonoJitInfo *ji;
+
+		if (code)
+			ji = mono_jit_info_table_find (mono_domain_get (), (char*)code);
+		else
+			ji = NULL;
+
+		/* Avoid recursion */
+		if (!(ji && ji->method->wrapper_type == MONO_WRAPPER_SYNCHRONIZED))
+			m = mono_marshal_get_synchronized_wrapper (m);
+	}
+
+	addr = mono_compile_method (m);
+	g_assert (addr);
+
+	if (mono_method_needs_static_rgctx_invoke (m, 0))
+		addr = mono_create_static_rgctx_trampoline (m, addr);
+
+	if (m->klass->valuetype)
+		addr = get_unbox_trampoline (mono_get_generic_context_from_code (code), m, addr, need_rgctx_tramp);
+
+	vtable_slot = &(vt->vtable [mono_method_get_vtable_slot (m)]);
+	g_assert (*vtable_slot);
+
+	if (!proxy && (mono_aot_is_got_entry (code, (guint8*)vtable_slot) || mono_domain_owns_vtable_slot (mono_domain_get (), vtable_slot))) {
+#ifdef MONO_ARCH_HAVE_IMT
+		vtable_slot = mono_convert_imt_slot_to_vtable_slot (vtable_slot, regs, code, m, NULL, &need_rgctx_tramp);
+#endif
+		*vtable_slot = mono_get_addr_from_ftnptr (addr);
+	  }
+
+	mono_debugger_trampoline_compiled (code, m, addr);
+
+	return addr;
+}
+#endif
 
 gpointer
-mono_generic_virtual_remoting_trampoline (gssize *regs, guint8 *code, MonoMethod *m, guint8 *tramp)
+mono_generic_virtual_remoting_trampoline (mgreg_t *regs, guint8 *code, MonoMethod *m, guint8 *tramp)
 {
 	MonoGenericContext context = { NULL, NULL };
-	MonoMethod *declaring;
+	MonoMethod *imt_method, *declaring;
 	gpointer addr;
 
 	g_assert (m->is_generic);
@@ -398,7 +661,9 @@ mono_generic_virtual_remoting_trampoline (gssize *regs, guint8 *code, MonoMethod
 		g_assert (!m->klass->generic_container);
 
 #ifdef MONO_ARCH_HAVE_IMT
-	context.method_inst = (MonoGenericInst*)mono_arch_find_imt_method ((gpointer*)regs, code);
+	imt_method = mono_arch_find_imt_method (regs, code);
+	if (imt_method->is_inflated)
+		context.method_inst = ((MonoMethodInflated*)imt_method)->context.method_inst;
 #endif
 	m = mono_class_inflate_generic_method (declaring, &context);
 	m = mono_marshal_get_remoting_invoke_with_check (m);
@@ -417,9 +682,9 @@ mono_generic_virtual_remoting_trampoline (gssize *regs, guint8 *code, MonoMethod
  *   This trampoline handles calls made from AOT code. We try to bypass the 
  * normal JIT compilation logic to avoid loading the metadata for the method.
  */
-#ifdef MONO_ARCH_HAVE_CREATE_TRAMPOLINE_FROM_TOKEN
+#ifdef MONO_ARCH_AOT_SUPPORTED
 gpointer
-mono_aot_trampoline (gssize *regs, guint8 *code, guint8 *token_info, 
+mono_aot_trampoline (mgreg_t *regs, guint8 *code, guint8 *token_info, 
 					 guint8* tramp)
 {
 	MonoImage *image;
@@ -427,7 +692,6 @@ mono_aot_trampoline (gssize *regs, guint8 *code, guint8 *token_info,
 	MonoMethod *method = NULL;
 	gpointer addr;
 	gpointer *vtable_slot;
-	gboolean is_got_entry;
 	guint8 *plt_entry;
 
 	image = *(gpointer*)(gpointer)token_info;
@@ -443,31 +707,16 @@ mono_aot_trampoline (gssize *regs, guint8 *code, guint8 *token_info,
 		return mono_magic_trampoline (regs, code, method, tramp);
 	}
 
-	vtable_slot = mono_arch_get_vcall_slot_addr (code, (gpointer*)regs);
+	addr = mono_create_ftnptr (mono_domain_get (), addr);
+
+	vtable_slot = mono_get_vcall_slot_addr (code, regs);
 	g_assert (!vtable_slot);
 
 	/* This is a normal call through a PLT entry */
 	plt_entry = mono_aot_get_plt_entry (code);
 	g_assert (plt_entry);
 
-	mono_arch_patch_plt_entry (plt_entry, addr);
-
-	is_got_entry = FALSE;
-
-	/*
-	 * Since AOT code is only used in the root domain, 
-	 * mono_domain_get () != mono_get_root_domain () means the calling method
-	 * is AppDomain:InvokeInDomain, so this is the same check as in 
-	 * mono_method_same_domain () but without loading the metadata for the method.
-	 */
-	if ((is_got_entry && (mono_domain_get () == mono_get_root_domain ())) || mono_domain_owns_vtable_slot (mono_domain_get (), vtable_slot)) {
-#ifdef MONO_ARCH_HAVE_IMT
-		if (!method)
-			method = mono_get_method (image, token, NULL);
-		vtable_slot = mono_convert_imt_slot_to_vtable_slot (vtable_slot, (gpointer*)regs, code, method, NULL);
-#endif
-		*vtable_slot = addr;
-	}
+	mono_aot_patch_plt_entry (plt_entry, NULL, regs, addr);
 
 	return addr;
 }
@@ -478,7 +727,7 @@ mono_aot_trampoline (gssize *regs, guint8 *code, guint8 *token_info,
  *   This trampoline handles calls made from AOT code through the PLT table.
  */
 gpointer
-mono_aot_plt_trampoline (gssize *regs, guint8 *code, guint8 *aot_module, 
+mono_aot_plt_trampoline (mgreg_t *regs, guint8 *code, guint8 *aot_module, 
 						 guint8* tramp)
 {
 	guint32 plt_info_offset = mono_aot_get_plt_info_offset (regs, code);
@@ -494,14 +743,14 @@ mono_aot_plt_trampoline (gssize *regs, guint8 *code, guint8 *aot_module,
  * for the type, then patches the caller code so it is not called again.
  */
 void
-mono_class_init_trampoline (gssize *regs, guint8 *code, MonoVTable *vtable, guint8 *tramp)
+mono_class_init_trampoline (mgreg_t *regs, guint8 *code, MonoVTable *vtable, guint8 *tramp)
 {
 	guint8 *plt_entry = mono_aot_get_plt_entry (code);
 
 	mono_runtime_class_init (vtable);
 
 	if (plt_entry) {
-		mono_arch_nullify_plt_entry (plt_entry);
+		mono_arch_nullify_plt_entry (plt_entry, regs);
 	} else {
 		mono_arch_nullify_class_init_trampoline (code, regs);
 	}
@@ -514,19 +763,20 @@ mono_class_init_trampoline (gssize *regs, guint8 *code, MonoVTable *vtable, guin
  * for the type.
  */
 void
-mono_generic_class_init_trampoline (gssize *regs, guint8 *code, MonoVTable *vtable, guint8 *tramp)
+mono_generic_class_init_trampoline (mgreg_t *regs, guint8 *code, MonoVTable *vtable, guint8 *tramp)
 {
 	mono_runtime_class_init (vtable);
 }
 
 static gpointer
-mono_rgctx_lazy_fetch_trampoline (gssize *regs, guint8 *code, gpointer data, guint8 *tramp)
+mono_rgctx_lazy_fetch_trampoline (mgreg_t *regs, guint8 *code, gpointer data, guint8 *tramp)
 {
 #ifdef MONO_ARCH_VTABLE_REG
 	static gboolean inited = FALSE;
 	static int num_lookups = 0;
 	guint32 slot = GPOINTER_TO_UINT (data);
-	gpointer arg = (gpointer)(gssize)regs [MONO_ARCH_VTABLE_REG];
+	mgreg_t *r = (mgreg_t*)regs;
+	gpointer arg = (gpointer)(gssize)r [MONO_ARCH_VTABLE_REG];
 	guint32 index = MONO_RGCTX_SLOT_INDEX (slot);
 	gboolean mrgctx = MONO_RGCTX_SLOT_IS_MRGCTX (slot);
 
@@ -547,13 +797,13 @@ mono_rgctx_lazy_fetch_trampoline (gssize *regs, guint8 *code, gpointer data, gui
 }
 
 void
-mono_monitor_enter_trampoline (gssize *regs, guint8 *code, MonoObject *obj, guint8 *tramp)
+mono_monitor_enter_trampoline (mgreg_t *regs, guint8 *code, MonoObject *obj, guint8 *tramp)
 {
 	mono_monitor_enter (obj);
 }
 
 void
-mono_monitor_exit_trampoline (gssize *regs, guint8 *code, MonoObject *obj, guint8 *tramp)
+mono_monitor_exit_trampoline (mgreg_t *regs, guint8 *code, MonoObject *obj, guint8 *tramp)
 {
 	mono_monitor_exit (obj);
 }
@@ -567,7 +817,7 @@ mono_monitor_exit_trampoline (gssize *regs, guint8 *code, MonoObject *obj, guint
  * This is called once the first time a delegate is invoked, so it must be fast.
  */
 gpointer
-mono_delegate_trampoline (gssize *regs, guint8 *code, gpointer *tramp_data, guint8* tramp)
+mono_delegate_trampoline (mgreg_t *regs, guint8 *code, gpointer *tramp_data, guint8* tramp)
 {
 	MonoDomain *domain = mono_domain_get ();
 	MonoDelegate *delegate;
@@ -575,6 +825,7 @@ mono_delegate_trampoline (gssize *regs, guint8 *code, gpointer *tramp_data, guin
 	MonoMethod *m;
 	MonoMethod *method = NULL;
 	gboolean multicast, callvirt;
+	gboolean need_rgctx_tramp = FALSE;
 	MonoMethod *invoke = tramp_data [0];
 	guint8 *impl_this = tramp_data [1];
 	guint8 *impl_nothis = tramp_data [2];
@@ -597,8 +848,13 @@ mono_delegate_trampoline (gssize *regs, guint8 *code, gpointer *tramp_data, guin
 		 * (ctor_with_method () does this, but it doesn't store the wrapper back into
 		 * delegate->method).
 		 */
-		if (delegate->target && delegate->target->vtable->klass == mono_defaults.transparent_proxy_class)
-			method = mono_marshal_get_remoting_invoke (method);
+		if (delegate->target && delegate->target->vtable->klass == mono_defaults.transparent_proxy_class) {
+#ifndef DISABLE_COM
+			if (((MonoTransparentProxy *)delegate->target)->remote_class->proxy_class != mono_defaults.com_object_class && 
+			   !((MonoTransparentProxy *)delegate->target)->remote_class->proxy_class->is_com_object)
+#endif
+				method = mono_marshal_get_remoting_invoke (method);
+		}
 		else if (mono_method_signature (method)->hasthis && method->klass->valuetype)
 			method = mono_marshal_get_unbox_wrapper (method);
 	} else {
@@ -611,6 +867,9 @@ mono_delegate_trampoline (gssize *regs, guint8 *code, gpointer *tramp_data, guin
 	if (method && method->iflags & METHOD_IMPL_ATTRIBUTE_SYNCHRONIZED)
 		method = mono_marshal_get_synchronized_wrapper (method);
 
+	if (method && mono_method_needs_static_rgctx_invoke (method, FALSE))
+		need_rgctx_tramp = TRUE;
+
 	/* 
 	 * If the called address is a trampoline, replace it with the compiled method so
 	 * further calls don't have to go through the trampoline.
@@ -627,12 +886,19 @@ mono_delegate_trampoline (gssize *regs, guint8 *code, gpointer *tramp_data, guin
 		}
 	}
 
+	if (need_rgctx_tramp)
+		delegate->method_ptr = mono_create_static_rgctx_trampoline (method, delegate->method_ptr);
+
 	multicast = ((MonoMulticastDelegate*)delegate)->prev != NULL;
 	if (!multicast && !callvirt) {
-		code = delegate->target ? impl_this : impl_nothis;
+		if (method && (method->flags & METHOD_ATTRIBUTE_STATIC) && mono_method_signature (method)->param_count == mono_method_signature (invoke)->param_count + 1)
+			/* Closed static delegate */
+			code = impl_this;
+		else
+			code = delegate->target ? impl_this : impl_nothis;
 
 		if (code) {
-			delegate->invoke_impl = code;
+			delegate->invoke_impl = mono_get_addr_from_ftnptr (code);
 			return code;
 		}
 	}
@@ -643,12 +909,7 @@ mono_delegate_trampoline (gssize *regs, guint8 *code, gpointer *tramp_data, guin
 	delegate->invoke_impl = mono_get_addr_from_ftnptr (code);
 	mono_debugger_trampoline_compiled (NULL, m, delegate->invoke_impl);
 
-#ifdef __mono_ppc64__
-	// FIXME:
-	return mono_get_addr_from_ftnptr (code);
-#else
 	return code;
-#endif
 }
 
 #endif
@@ -690,6 +951,10 @@ mono_get_trampoline_func (MonoTrampolineType tramp_type)
 		return mono_monitor_enter_trampoline;
 	case MONO_TRAMPOLINE_MONITOR_EXIT:
 		return mono_monitor_exit_trampoline;
+#ifdef ENABLE_LLVM
+	case MONO_TRAMPOLINE_LLVM_VCALL:
+		return mono_llvm_vcall_trampoline;
+#endif
 	default:
 		g_assert_not_reached ();
 		return NULL;
@@ -720,6 +985,9 @@ mono_trampolines_init (void)
 	mono_trampoline_code [MONO_TRAMPOLINE_GENERIC_VIRTUAL_REMOTING] = mono_arch_create_trampoline_code (MONO_TRAMPOLINE_GENERIC_VIRTUAL_REMOTING);
 	mono_trampoline_code [MONO_TRAMPOLINE_MONITOR_ENTER] = mono_arch_create_trampoline_code (MONO_TRAMPOLINE_MONITOR_ENTER);
 	mono_trampoline_code [MONO_TRAMPOLINE_MONITOR_EXIT] = mono_arch_create_trampoline_code (MONO_TRAMPOLINE_MONITOR_EXIT);
+#ifdef ENABLE_LLVM
+	mono_trampoline_code [MONO_TRAMPOLINE_LLVM_VCALL] = mono_arch_create_trampoline_code (MONO_TRAMPOLINE_LLVM_VCALL);
+#endif
 }
 
 void
@@ -727,8 +995,6 @@ mono_trampolines_cleanup (void)
 {
 	if (class_init_hash_addr)
 		g_hash_table_destroy (class_init_hash_addr);
-	if (delegate_trampoline_hash_addr)
-		g_hash_table_destroy (delegate_trampoline_hash_addr);
 
 	DeleteCriticalSection (&trampolines_mutex);
 }
@@ -794,8 +1060,12 @@ mono_create_generic_class_init_trampoline (void)
 
 	mono_trampolines_lock ();
 
-	if (!code)
-		code = mono_arch_create_generic_class_init_trampoline ();
+	if (!code) {
+		if (mono_aot_only)
+			code = mono_aot_get_named_code ("generic_class_init_trampoline");
+		else
+			code = mono_arch_create_generic_class_init_trampoline ();
+	}
 
 	mono_trampolines_unlock ();
 
@@ -830,9 +1100,7 @@ mono_create_jump_trampoline (MonoDomain *domain, MonoMethod *method, gboolean ad
 	code = mono_create_specific_trampoline (method, MONO_TRAMPOLINE_JUMP, mono_domain_get (), &code_size);
 	g_assert (code_size);
 
-	mono_domain_lock (domain);
-	ji = mono_domain_alloc0 (domain, sizeof (MonoJitInfo));
-	mono_domain_unlock (domain);
+	ji = mono_domain_alloc0 (domain, MONO_SIZEOF_JIT_INFO);
 	ji->code_start = code;
 	ji->code_size = code_size;
 	ji->method = method;
@@ -887,7 +1155,6 @@ mono_create_jit_trampoline (MonoMethod *method)
 	return mono_create_jit_trampoline_in_domain (mono_domain_get (), method);
 }
 
-#ifdef MONO_ARCH_HAVE_CREATE_TRAMPOLINE_FROM_TOKEN
 gpointer
 mono_create_jit_trampoline_from_token (MonoImage *image, guint32 token)
 {
@@ -896,9 +1163,7 @@ mono_create_jit_trampoline_from_token (MonoImage *image, guint32 token)
 	MonoDomain *domain = mono_domain_get ();
 	guint8 *buf, *start;
 
-	mono_domain_lock (domain);
-	buf = start = mono_code_manager_reserve (domain->code_mp, 2 * sizeof (gpointer));
-	mono_domain_unlock (domain);
+	buf = start = mono_domain_code_reserve (domain, 2 * sizeof (gpointer));
 
 	*(gpointer*)(gpointer)buf = image;
 	buf += sizeof (gpointer);
@@ -910,7 +1175,6 @@ mono_create_jit_trampoline_from_token (MonoImage *image, guint32 token)
 
 	return tramp;
 }	
-#endif
 
 gpointer
 mono_create_delegate_trampoline (MonoClass *klass)
@@ -932,17 +1196,10 @@ mono_create_delegate_trampoline (MonoClass *klass)
 	invoke = mono_get_delegate_invoke (klass);
 	g_assert (invoke);
 
-	mono_domain_lock (domain );
 	tramp_data = mono_domain_alloc (domain, sizeof (gpointer) * 3);
-	mono_domain_unlock (domain);
 	tramp_data [0] = invoke;
-	if (mono_aot_only) {
-		tramp_data [1] = NULL;
-		tramp_data [2] = NULL;
-	} else {
-		tramp_data [1] = mono_arch_get_delegate_invoke_impl (mono_method_signature (invoke), TRUE);
-		tramp_data [2] = mono_arch_get_delegate_invoke_impl (mono_method_signature (invoke), FALSE);
-	}
+	tramp_data [1] = mono_arch_get_delegate_invoke_impl (mono_method_signature (invoke), TRUE);
+	tramp_data [2] = mono_arch_get_delegate_invoke_impl (mono_method_signature (invoke), FALSE);
 
 	ptr = mono_create_specific_trampoline (tramp_data, MONO_TRAMPOLINE_DELEGATE, mono_domain_get (), &code_size);
 	g_assert (code_size);
@@ -953,12 +1210,6 @@ mono_create_delegate_trampoline (MonoClass *klass)
 							  klass, ptr);
 	mono_domain_unlock (domain);
 
-	mono_trampolines_lock ();
-	if (!delegate_trampoline_hash_addr)
-		delegate_trampoline_hash_addr = g_hash_table_new (NULL, NULL);
-	g_hash_table_insert (delegate_trampoline_hash_addr, ptr, klass);
-	mono_trampolines_unlock ();
-
 	return ptr;
 #else
 	return NULL;
@@ -1058,6 +1309,61 @@ mono_create_monitor_exit_trampoline (void)
 #endif
 	return code;
 }
+ 
+#ifdef ENABLE_LLVM
+/*
+ * mono_create_llvm_vcall_trampoline:
+ *
+ *  LLVM emits code for virtual calls which mono_get_vcall_slot is unable to
+ * decode, i.e. only the final branch address is available:
+ * mov <offset>(%rax), %rax
+ * <random code inserted by instruction scheduling>
+ * call *%rax
+ *
+ * To work around this problem, we don't use the common vtable trampoline when
+ * llvm is enabled. Instead, we use one trampoline per method.
+ */
+gpointer
+mono_create_llvm_vcall_trampoline (MonoMethod *method)
+{
+	MonoDomain *domain;
+	gpointer res;
+
+	domain = mono_domain_get ();
+
+	mono_domain_lock (domain);
+	res = g_hash_table_lookup (domain_jit_info (domain)->llvm_vcall_trampoline_hash, method);
+	mono_domain_unlock (domain);
+	if (res)
+		return res;
+
+	res = mono_create_specific_trampoline (method, MONO_TRAMPOLINE_LLVM_VCALL, domain, NULL);
+
+	mono_domain_lock (domain);
+	g_hash_table_insert (domain_jit_info (domain)->llvm_vcall_trampoline_hash, method, res);
+	mono_domain_unlock (domain);
+
+	return res;
+}
+
+/*
+ * mono_create_llvm_imt_trampoline:
+ *
+ *   LLVM compiled code can't pass in the IMT argument, so we use this trampoline, which
+ * sets the IMT argument, then branches to the contents of the vtable slot given by
+ * vt_offset in the vtable which is obtained from the argument list.
+ */
+gpointer
+mono_create_llvm_imt_trampoline (MonoDomain *domain, MonoMethod *m, int vt_offset)
+{
+#ifdef MONO_ARCH_HAVE_LLVM_IMT_TRAMPOLINE
+	return mono_arch_get_llvm_imt_trampoline (domain, m, vt_offset);
+#else
+	g_assert_not_reached ();
+	return NULL;
+#endif
+}
+#endif
 
 MonoVTable*
 mono_find_class_init_trampoline_by_addr (gconstpointer addr)
@@ -1073,20 +1379,6 @@ mono_find_class_init_trampoline_by_addr (gconstpointer addr)
 	return res;
 }
 
-MonoClass*
-mono_find_delegate_trampoline_by_addr (gconstpointer addr)
-{
-	MonoClass *res;
-
-	mono_trampolines_lock ();
-	if (delegate_trampoline_hash_addr)
-		res = g_hash_table_lookup (delegate_trampoline_hash_addr, addr);
-	else
-		res = NULL;
-	mono_trampolines_unlock ();
-	return res;
-}
-
 guint32
 mono_find_rgctx_lazy_fetch_trampoline_by_addr (gconstpointer addr)
 {
