@@ -48,9 +48,7 @@ mono_arch_get_unbox_trampoline (MonoGenericSharingContext *gsctx, MonoMethod *m,
 	if (MONO_TYPE_ISSTRUCT (mono_method_signature (m)->ret))
 		this_pos = 8;
 	    
-	mono_domain_lock (domain);
-	start = code = mono_code_manager_reserve (domain->code_mp, 16);
-	mono_domain_unlock (domain);
+	start = code = mono_domain_code_reserve (domain, 16);
 
 	x86_alu_membase_imm (code, X86_ADD, X86_ESP, this_pos, sizeof (MonoObject));
 	x86_jump_code (code, addr);
@@ -59,6 +57,27 @@ mono_arch_get_unbox_trampoline (MonoGenericSharingContext *gsctx, MonoMethod *m,
 	return start;
 }
 
+gpointer
+mono_arch_get_static_rgctx_trampoline (MonoMethod *m, MonoMethodRuntimeGenericContext *mrgctx, gpointer addr)
+{
+	guint8 *code, *start;
+	int buf_len;
+
+	MonoDomain *domain = mono_domain_get ();
+
+	buf_len = 10;
+
+	start = code = mono_domain_code_reserve (domain, buf_len);
+
+	x86_mov_reg_imm (code, MONO_ARCH_RGCTX_REG, mrgctx);
+	x86_jump_code (code, addr);
+	g_assert ((code - start) <= buf_len);
+
+	mono_arch_flush_icache (start, code - start);
+
+	return start;
+}
+
 void
 mono_arch_patch_callsite (guint8 *method_start, guint8 *orig_code, guint8 *addr)
 {
@@ -100,7 +119,7 @@ mono_arch_patch_callsite (guint8 *method_start, guint8 *orig_code, guint8 *addr)
 }
 
 void
-mono_arch_patch_plt_entry (guint8 *code, guint8 *addr)
+mono_arch_patch_plt_entry (guint8 *code, gpointer *got, mgreg_t *regs, guint8 *addr)
 {
 	/* A PLT entry: jmp <DISP> */
 	g_assert (code [0] == 0xe9);
@@ -110,7 +129,7 @@ mono_arch_patch_plt_entry (guint8 *code, guint8 *addr)
 }
 
 void
-mono_arch_nullify_class_init_trampoline (guint8 *code, gssize *regs)
+mono_arch_nullify_class_init_trampoline (guint8 *code, mgreg_t *regs)
 {
 	guint8 buf [16];
 	gboolean can_write = mono_breakpoint_clean_code (NULL, code, 6, buf, sizeof (buf));
@@ -155,7 +174,7 @@ mono_arch_nullify_class_init_trampoline (guint8 *code, gssize *regs)
 		/* call *<OFFSET>(<REG>) -> Call made from AOT code */
 		gpointer *vtable_slot;
 
-		vtable_slot = mono_arch_get_vcall_slot_addr (code + 5, (gpointer*)regs);
+		vtable_slot = mono_get_vcall_slot_addr (code + 5, regs);
 		g_assert (vtable_slot);
 
 		*vtable_slot = nullified_class_init_trampoline;
@@ -167,7 +186,7 @@ mono_arch_nullify_class_init_trampoline (guint8 *code, gssize *regs)
 }
 
 void
-mono_arch_nullify_plt_entry (guint8 *code)
+mono_arch_nullify_plt_entry (guint8 *code, mgreg_t *regs)
 {
 	if (!mono_running_on_valgrind ()) {
 		guint32 ops;
@@ -314,9 +333,12 @@ mono_arch_create_trampoline_code (MonoTrampolineType tramp_type)
 
 	/* Check for thread interruption */
 	/* This is not perf critical code so no need to check the interrupt flag */
+	/* Align the stack on osx */
+	x86_alu_reg_imm (buf, X86_SUB, X86_ESP, 3 * 4);
 	x86_push_reg (buf, X86_EAX);
 	x86_call_code (buf, (guint8*)mono_thread_force_interruption_checkpoint);
 	x86_pop_reg (buf, X86_EAX);
+	x86_alu_reg_imm (buf, X86_ADD, X86_ESP, 3 * 4);
 
 	/* Restore LMF */
 
@@ -405,9 +427,7 @@ mono_arch_create_specific_trampoline (gpointer arg1, MonoTrampolineType tramp_ty
 	
 	tramp = mono_get_trampoline_code (tramp_type);
 
-	mono_domain_lock (domain);
-	code = buf = mono_code_manager_reserve_align (domain->code_mp, TRAMPOLINE_SIZE, 4);
-	mono_domain_unlock (domain);
+	code = buf = mono_domain_code_reserve_align (domain, TRAMPOLINE_SIZE, 4);
 
 	x86_push_imm (buf, arg1);
 	x86_jump_code (buf, tramp);
@@ -435,7 +455,7 @@ mono_arch_create_rgctx_lazy_fetch_trampoline (guint32 slot)
 	mrgctx = MONO_RGCTX_SLOT_IS_MRGCTX (slot);
 	index = MONO_RGCTX_SLOT_INDEX (slot);
 	if (mrgctx)
-		index += sizeof (MonoMethodRuntimeGenericContext) / sizeof (gpointer);
+		index += MONO_SIZEOF_METHOD_RUNTIME_GENERIC_CONTEXT / sizeof (gpointer);
 	for (depth = 0; ; ++depth) {
 		int size = mono_class_rgctx_get_array_size (depth, mrgctx);
 
@@ -465,7 +485,7 @@ mono_arch_create_rgctx_lazy_fetch_trampoline (guint32 slot)
 	for (i = 0; i < depth; ++i) {
 		/* load ptr to next array */
 		if (mrgctx && i == 0)
-			x86_mov_reg_membase (buf, X86_EAX, X86_EAX, sizeof (MonoMethodRuntimeGenericContext), 4);
+			x86_mov_reg_membase (buf, X86_EAX, X86_EAX, MONO_SIZEOF_METHOD_RUNTIME_GENERIC_CONTEXT, 4);
 		else
 			x86_mov_reg_membase (buf, X86_EAX, X86_EAX, 0, 4);
 		/* is the ptr null? */
@@ -681,7 +701,7 @@ mono_arch_create_monitor_exit_trampoline_full (guint32 *code_size, MonoJumpInfo
 {
 	guint8 *tramp = mono_get_trampoline_code (MONO_TRAMPOLINE_MONITOR_EXIT);
 	guint8 *code, *buf;
-	guint8 *jump_obj_null, *jump_have_waiters;
+	guint8 *jump_obj_null, *jump_have_waiters, *jump_sync_null, *jump_not_owned;
 	guint8 *jump_next;
 	int tramp_size;
 	int owner_offset, nest_offset, entry_count_offset;
@@ -712,28 +732,22 @@ mono_arch_create_monitor_exit_trampoline_full (guint32 *code_size, MonoJumpInfo
 		x86_mov_reg_membase (buf, X86_ECX, X86_EAX, G_STRUCT_OFFSET (MonoObject, synchronisation), 4);
 		/* is synchronization null? */
 		x86_test_reg_reg (buf, X86_ECX, X86_ECX);
-		/* if not, jump to next case */
-		jump_next = buf;
-		x86_branch8 (buf, X86_CC_NZ, -1, 1);
-		/* if yes, just return */
-		x86_ret (buf);
+		/* if yes, jump to actual trampoline */
+		jump_sync_null = buf;
+		x86_branch8 (buf, X86_CC_Z, -1, 1);
 
 		/* next case: synchronization is not null */
-		x86_patch (jump_next, buf);
 		/* load MonoThread* into EDX */
 		buf = mono_x86_emit_tls_get (buf, X86_EDX, mono_thread_get_tls_offset ());
 		/* load TID into EDX */
 		x86_mov_reg_membase (buf, X86_EDX, X86_EDX, G_STRUCT_OFFSET (MonoThread, tid), 4);
 		/* is synchronization->owner == TID */
 		x86_alu_membase_reg (buf, X86_CMP, X86_ECX, owner_offset, X86_EDX);
-		/* if yes, jump to next case */
-		jump_next = buf;
-		x86_branch8 (buf, X86_CC_Z, -1, 1);
-		/* if not, just return */
-		x86_ret (buf);
+		/* if no, jump to actual trampoline */
+		jump_not_owned = buf;
+		x86_branch8 (buf, X86_CC_NZ, -1, 1);
 
 		/* next case: synchronization->owner == TID */
-		x86_patch (jump_next, buf);
 		/* is synchronization->nest == 1 */
 		x86_alu_membase_imm (buf, X86_CMP, X86_ECX, nest_offset, 1);
 		/* if not, jump to next case */
@@ -757,6 +771,9 @@ mono_arch_create_monitor_exit_trampoline_full (guint32 *code_size, MonoJumpInfo
 		/* push obj and jump to the actual trampoline */
 		x86_patch (jump_obj_null, buf);
 		x86_patch (jump_have_waiters, buf);
+		x86_patch (jump_not_owned, buf);
+		x86_patch (jump_sync_null, buf);
+
 		x86_push_reg (buf, X86_EAX);
 		x86_jump_code (buf, tramp);
 	} else {
